package lcsb.mapviewer.web.app;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.module.SimpleModule;
import lcsb.mapviewer.common.MinervaConfigurationHolder;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.modelutils.map.ClassTreeNode;
import lcsb.mapviewer.modelutils.serializer.CalendarSerializer;
import lcsb.mapviewer.modelutils.serializer.ClassTreeNodeSerializer;
import lcsb.mapviewer.modelutils.serializer.ColorSerializer;
import lcsb.mapviewer.modelutils.serializer.Line2DSerializer;
import lcsb.mapviewer.modelutils.serializer.model.map.MinifiedBioEntitySerializer;
import lcsb.mapviewer.modelutils.serializer.model.map.reaction.ReactionNodeSerializer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.CacheControl;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ViewResolver;
import org.springframework.web.servlet.config.annotation.CorsRegistry;
import org.springframework.web.servlet.config.annotation.InterceptorRegistry;
import org.springframework.web.servlet.config.annotation.ResourceHandlerRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;
import org.springframework.web.servlet.view.InternalResourceViewResolver;

import java.awt.Color;
import java.awt.geom.Line2D;
import java.util.Calendar;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Configuration
@EnableScheduling
@ComponentScan(basePackages = {"lcsb.mapviewer.api"})
public class WebConfig implements WebMvcConfigurer {

  private final Logger logger = LogManager.getLogger();

  @Autowired
  private MinervaConfigurationHolder configurationHolder;

  @Override
  public void addCorsMappings(final CorsRegistry registry) {
    registry.addMapping("/**").allowedMethods("*");
  }

  @Override
  public void addResourceHandlers(final ResourceHandlerRegistry registry) {
    final String path = configurationHolder.getDataRootPath();
    final CacheControl cacheControl = CacheControl.maxAge(31, TimeUnit.DAYS);
    registry.addResourceHandler("/**")
        .addResourceLocations("/", "classpath:/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/minerva/**")
        .addResourceLocations("/", "classpath:/", "file:" + path + "/front/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/map_images/**")
        .addResourceLocations("file:" + path + "/map_images/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/minerva-big/**")
        .addResourceLocations("file:" + path + "/minerva-big/")
        .setCacheControl(cacheControl);
  }

  private final List<HandlerInterceptor> interceptors;

  @Autowired
  public WebConfig(final List<HandlerInterceptor> interceptors) {
    this.interceptors = interceptors;
  }

  @Override
  public void extendMessageConverters(final List<HttpMessageConverter<?>> converters) {
    for (final HttpMessageConverter<?> converter : converters) {
      if (converter instanceof MappingJackson2HttpMessageConverter) {
        final MappingJackson2HttpMessageConverter m = (MappingJackson2HttpMessageConverter) converter;
        final SimpleModule module = new SimpleModule();
        module.addSerializer(Calendar.class, new CalendarSerializer());
        module.addSerializer(Color.class, new ColorSerializer());
        module.addSerializer(Line2D.class, new Line2DSerializer());
        module.addSerializer(ClassTreeNode.class, new ClassTreeNodeSerializer());
        module.addSerializer(ReactionNode.class, new ReactionNodeSerializer());
        module.addSerializer(BioEntity.class, new MinifiedBioEntitySerializer());

        m.getObjectMapper().registerModule(module);
      }
    }
  }

  @Bean
  public ViewResolver viewResolver() {
    final InternalResourceViewResolver resolver = new InternalResourceViewResolver();
    resolver.setPrefix("/");
    resolver.setSuffix(".html");
    return resolver;
  }

  @Override
  public void addInterceptors(final InterceptorRegistry registry) {
    interceptors.forEach(registry::addInterceptor);
  }

  @Bean
  public ObjectMapper objectMapper() {
    final ObjectMapper mapper = new ObjectMapper();

    return mapper;
  }

}
