package lcsb.mapviewer.model.map;

import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Chemical;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Drug;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Phenotype;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Type of known miriam annotation.
 *
 * @author Piotr Gawron
 */
@SuppressWarnings("unchecked")
public enum MiriamType {

  _3DMET("3DMET",
      "http://www.3dmet.dna.affrc.go.jp/",
      new String[]{"urn:miriam:3dmet",
          "http://identifiers.org/3dmet/",
          "http://identifiers.org/3dmet",
          "https://identifiers.org/3dmet/",
          "https://identifiers.org/3dmet",
      },
      new Class<?>[]{},
      " MIR:00000066",
      new Class<?>[]{},
      "3dmet",
      "B01164"),

  ABS("ABS",
      "http://genome.crg.es/datasets/abs2005/",
      new String[]{"urn:miriam:abs",
          "http://identifiers.org/abs/",
          "http://identifiers.org/abs",
          "https://identifiers.org/abs/",
          "https://identifiers.org/abs",
      }, new Class<?>[]{}, "MIR:00000277",
      new Class<?>[]{},
      "abs",
      "A0014"),

  ACEVIEW_WORM("Aceview Worm",
      "https://www.ncbi.nlm.nih.gov/IEB/Research/Acembly/index.html?worm",
      new String[]{"urn:miriam:aceview.worm",
          "http://identifiers.org/aceview.worm/",
          "http://identifiers.org/aceview.worm",
          "https://identifiers.org/aceview.worm/",
          "https://identifiers.org/aceview.worm",
      }, new Class<?>[]{}, "MIR:00000282",
      new Class<?>[]{},
      "aceview.worm",
      "aap-1"),

  AFFYMETRIX_PROBESET("Affymetrix Probeset",
      "http://www.affymetrix.com/",
      new String[]{"urn:miriam:affy.probeset",
          "http://identifiers.org/affy.probeset/",
          "http://identifiers.org/affy.probeset",
          "https://identifiers.org/affy.probeset/",
          "https://identifiers.org/affy.probeset",
      }, new Class<?>[]{}, "MIR:00000394",
      new Class<?>[]{},
      "affy.probeset",
      "243002_at"),

  AFTOL("AFTOL",
      "http://aftol.org/data.php",
      new String[]{"urn:miriam:aftol.taxonomy",
          "http://identifiers.org/aftol.taxonomy/",
          "http://identifiers.org/aftol.taxonomy",
          "https://identifiers.org/aftol.taxonomy/",
          "https://identifiers.org/aftol.taxonomy",
      }, new Class<?>[]{}, "MIR:00000411",
      new Class<?>[]{},
      "aftol.taxonomy",
      "959"),

  ALLERGOME("Allergome",
      "http://www.allergome.org/",
      new String[]{"urn:miriam:allergome",
          "http://identifiers.org/allergome/",
          "http://identifiers.org/allergome",
          "https://identifiers.org/allergome/",
          "https://identifiers.org/allergome",
      }, new Class<?>[]{}, "MIR:00000334",
      new Class<?>[]{},
      "allergome",
      "1948"),

  AMOEBADB("AmoebaDB",
      "http://amoebadb.org/amoeba/",
      new String[]{"urn:miriam:amoebadb",
          "http://identifiers.org/amoebadb/",
          "http://identifiers.org/amoebadb",
          "https://identifiers.org/amoebadb/",
          "https://identifiers.org/amoebadb",
      }, new Class<?>[]{}, "MIR:00000148",
      new Class<?>[]{},
      "amoebadb",
      "EDI_244000"),

  ANATOMICAL_THERAPEUTIC_CHEMICAL("Anatomical Therapeutic Chemical",
      "http://www.whocc.no/atc_ddd_index/",
      new String[]{"urn:miriam:atc",
          "http://identifiers.org/atc/",
          "http://identifiers.org/atc",
          "https://identifiers.org/atc/",
          "https://identifiers.org/atc",
      }, new Class<?>[]{}, "MIR:00000088",
      new Class<?>[]{},
      "atc",
      "A10BA02"),

  ANATOMICAL_THERAPEUTIC_CHEMICAL_VETINARY("Anatomical Therapeutic Chemical Vetinary",
      "http://www.whocc.no/atcvet/atcvet_index/",
      new String[]{"urn:miriam:atcvet",
          "http://identifiers.org/atcvet/",
          "http://identifiers.org/atcvet",
          "https://identifiers.org/atcvet/",
          "https://identifiers.org/atcvet",
      }, new Class<?>[]{}, "MIR:00000267",
      new Class<?>[]{},
      "atcvet",
      "QJ51RV02"),

  ANIMAL_DIVERSITY_WEB("Animal Diversity Web",
      "https://animaldiversity.org/",
      new String[]{"urn:miriam:adw",
          "http://identifiers.org/adw/",
          "http://identifiers.org/adw",
          "https://identifiers.org/adw/",
          "https://identifiers.org/adw",
      }, new Class<?>[]{}, "MIR:00000492",
      new Class<?>[]{},
      "adw",
      "Lycalopex_vetulus"),

  ANIMAL_GENOME_CATTLE_QTL("Animal Genome Cattle QTL",
      "https://www.animalgenome.org/QTLdb",
      new String[]{"urn:miriam:cattleqtldb",
          "http://identifiers.org/cattleqtldb/",
          "http://identifiers.org/cattleqtldb",
          "https://identifiers.org/cattleqtldb/",
          "https://identifiers.org/cattleqtldb",
      }, new Class<?>[]{}, "MIR:00000504",
      new Class<?>[]{},
      "cattleqtldb",
      "4685"),

  ANIMAL_GENOME_CHICKEN_QTL("Animal Genome Chicken QTL",
      "https://www.animalgenome.org/QTLdb",
      new String[]{"urn:miriam:chickenqtldb",
          "http://identifiers.org/chickenqtldb/",
          "http://identifiers.org/chickenqtldb",
          "https://identifiers.org/chickenqtldb/",
          "https://identifiers.org/chickenqtldb",
      }, new Class<?>[]{}, "MIR:00000505",
      new Class<?>[]{},
      "chickenqtldb",
      "14362"),

  ANIMAL_GENOME_PIG_QTL("Animal Genome Pig QTL",
      "https://www.animalgenome.org/QTLdb",
      new String[]{"urn:miriam:pigqtldb",
          "http://identifiers.org/pigqtldb/",
          "http://identifiers.org/pigqtldb",
          "https://identifiers.org/pigqtldb/",
          "https://identifiers.org/pigqtldb",
      }, new Class<?>[]{}, "MIR:00000506",
      new Class<?>[]{},
      "pigqtldb",
      "14"),

  ANIMAL_GENOME_SHEEP_QTL("Animal Genome Sheep QTL",
      "https://www.animalgenome.org/QTLdb",
      new String[]{"urn:miriam:sheepqtldb",
          "http://identifiers.org/sheepqtldb/",
          "http://identifiers.org/sheepqtldb",
          "https://identifiers.org/sheepqtldb/",
          "https://identifiers.org/sheepqtldb",
      }, new Class<?>[]{}, "MIR:00000507",
      new Class<?>[]{},
      "sheepqtldb",
      "19803"),

  ANIMAL_TFDB_FAMILY("Animal TFDB Family",
      "http://www.bioguo.org/AnimalTFDB/family_index.php",
      new String[]{"urn:miriam:atfdb.family",
          "http://identifiers.org/atfdb.family/",
          "http://identifiers.org/atfdb.family",
          "https://identifiers.org/atfdb.family/",
          "https://identifiers.org/atfdb.family",
      }, new Class<?>[]{}, "MIR:00000316",
      new Class<?>[]{},
      "atfdb.family",
      "CUT"),

  ANTIBIOTIC_RESISTANCE_GENES_DATABASE("Antibiotic Resistance Genes Database",
      "http://ardb.cbcb.umd.edu/",
      new String[]{"urn:miriam:ardb",
          "http://identifiers.org/ardb/",
          "http://identifiers.org/ardb",
          "https://identifiers.org/ardb/",
          "https://identifiers.org/ardb",
      }, new Class<?>[]{}, "MIR:00000522",
      new Class<?>[]{},
      "ardb",
      "CAE46076"),

  ANTIBODY_REGISTRY("Antibody Registry",
      "http://antibodyregistry.org/",
      new String[]{"urn:miriam:antibodyregistry",
          "http://identifiers.org/antibodyregistry/",
          "http://identifiers.org/antibodyregistry",
          "https://identifiers.org/antibodyregistry/",
          "https://identifiers.org/antibodyregistry",
      }, new Class<?>[]{}, "MIR:00000516",
      new Class<?>[]{},
      "antibodyregistry",
      "493771"),

  ANTWEB("AntWeb",
      "http://www.antweb.org/",
      new String[]{"urn:miriam:antweb",
          "http://identifiers.org/antweb/",
          "http://identifiers.org/antweb",
          "https://identifiers.org/antweb/",
          "https://identifiers.org/antweb",
      }, new Class<?>[]{}, "MIR:00000146",
      new Class<?>[]{},
      "antweb",
      "casent0106247"),

  APD("APD",
      "http://aps.unmc.edu/AP/",
      new String[]{"urn:miriam:apd",
          "http://identifiers.org/apd/",
          "http://identifiers.org/apd",
          "https://identifiers.org/apd/",
          "https://identifiers.org/apd",
      }, new Class<?>[]{}, "MIR:00000278",
      new Class<?>[]{},
      "apd",
      "01001"),

  APHIDBASE_TRANSCRIPT("AphidBase Transcript",
      "http://www.aphidbase.com/aphidbase",
      new String[]{"urn:miriam:aphidbase.transcript",
          "http://identifiers.org/aphidbase.transcript/",
          "http://identifiers.org/aphidbase.transcript",
          "https://identifiers.org/aphidbase.transcript/",
          "https://identifiers.org/aphidbase.transcript",
      }, new Class<?>[]{}, "MIR:00000393",
      new Class<?>[]{},
      "aphidbase.transcript",
      "ACYPI000159"),

  ARACHNOSERVER("ArachnoServer",
      "http://www.arachnoserver.org/",
      new String[]{"urn:miriam:arachnoserver",
          "http://identifiers.org/arachnoserver/",
          "http://identifiers.org/arachnoserver",
          "https://identifiers.org/arachnoserver/",
          "https://identifiers.org/arachnoserver",
      }, new Class<?>[]{}, "MIR:00000193",
      new Class<?>[]{},
      "arachnoserver",
      "AS000060"),

  ARRAYEXPRESS("ArrayExpress",
      "https://www.ebi.ac.uk/arrayexpress/",
      new String[]{"urn:miriam:arrayexpress",
          "http://identifiers.org/arrayexpress/",
          "http://identifiers.org/arrayexpress",
          "https://identifiers.org/arrayexpress/",
          "https://identifiers.org/arrayexpress",
      }, new Class<?>[]{}, "MIR:00000036",
      new Class<?>[]{},
      "arrayexpress",
      "E-MEXP-1712"),

  ARRAYEXPRESS_PLATFORM("ArrayExpress Platform",
      "https://www.ebi.ac.uk/arrayexpress/",
      new String[]{"urn:miriam:arrayexpress.platform",
          "http://identifiers.org/arrayexpress.platform/",
          "http://identifiers.org/arrayexpress.platform",
          "https://identifiers.org/arrayexpress.platform/",
          "https://identifiers.org/arrayexpress.platform",
      }, new Class<?>[]{}, "MIR:00000294",
      new Class<?>[]{},
      "arrayexpress.platform",
      "A-GEOD-50"),

  ARXIV("arXiv",
      "https://arxiv.org/",
      new String[]{"urn:miriam:arxiv", "urn:oai:arXiv.org",
          "http://identifiers.org/arxiv/",
          "http://identifiers.org/arxiv",
          "https://identifiers.org/arxiv/",
          "https://identifiers.org/arxiv",
      }, new Class<?>[]{}, "MIR:00000035",
      new Class<?>[]{},
      "arxiv",
      "0807.4956v1"),

  ASAP("ASAP",
      "http://asap.ahabs.wisc.edu/asap/home.php",
      new String[]{"urn:miriam:asap",
          "http://identifiers.org/asap/",
          "http://identifiers.org/asap",
          "https://identifiers.org/asap/",
          "https://identifiers.org/asap",
      }, new Class<?>[]{}, "MIR:00000283",
      new Class<?>[]{},
      "asap",
      "ABE-0009634"),

  ASPGD_LOCUS("AspGD Locus",
      "http://www.aspgd.org/",
      new String[]{"urn:miriam:aspgd.locus",
          "http://identifiers.org/aspgd.locus/",
          "http://identifiers.org/aspgd.locus",
          "https://identifiers.org/aspgd.locus/",
          "https://identifiers.org/aspgd.locus",
      }, new Class<?>[]{}, "MIR:00000412",
      new Class<?>[]{},
      "aspgd.locus",
      "ASPL0000349247"),

  ASPGD_PROTEIN("AspGD Protein",
      "http://www.aspgd.org/",
      new String[]{"urn:miriam:aspgd.protein",
          "http://identifiers.org/aspgd.protein/",
          "http://identifiers.org/aspgd.protein",
          "https://identifiers.org/aspgd.protein/",
          "https://identifiers.org/aspgd.protein",
      }, new Class<?>[]{}, "MIR:00000413",
      new Class<?>[]{},
      "aspgd.protein",
      "ASPL0000052200"),

  ATCC("ATCC",
      "http://www.atcc.org/",
      new String[]{"urn:miriam:atcc",
          "http://identifiers.org/atcc/",
          "http://identifiers.org/atcc",
          "https://identifiers.org/atcc/",
          "https://identifiers.org/atcc",
      }, new Class<?>[]{}, "MIR:00000284",
      new Class<?>[]{},
      "atcc",
      "11303"),

  AUTDB("AutDB",
      "http://autism.mindspec.org/autdb/",
      new String[]{"urn:miriam:autdb",
          "http://identifiers.org/autdb/",
          "http://identifiers.org/autdb",
          "https://identifiers.org/autdb/",
          "https://identifiers.org/autdb",
      }, new Class<?>[]{}, "MIR:00000415",
      new Class<?>[]{},
      "autdb",
      "ADA"),

  BACMAP_BIOGRAPHY("BacMap Biography",
      "http://bacmap.wishartlab.com/",
      new String[]{"urn:miriam:bacmap.biog",
          "http://identifiers.org/bacmap.biog/",
          "http://identifiers.org/bacmap.biog",
          "https://identifiers.org/bacmap.biog/",
          "https://identifiers.org/bacmap.biog",
      }, new Class<?>[]{}, "MIR:00000361",
      new Class<?>[]{},
      "bacmap.biog",
      "1050"),

  BACMAP_MAP("BacMap Map",
      "http://bacmap.wishartlab.com/",
      new String[]{"urn:miriam:bacmap.map",
          "http://identifiers.org/bacmap.map/",
          "http://identifiers.org/bacmap.map",
          "https://identifiers.org/bacmap.map/",
          "https://identifiers.org/bacmap.map",
      }, new Class<?>[]{}, "MIR:00000416",
      new Class<?>[]{},
      "bacmap.map",
      "AP011135"),

  BDGP_EST("BDGP EST",
      "https://www.ncbi.nlm.nih.gov/dbEST/index.html",
      new String[]{"urn:miriam:bdgp.est",
          "http://identifiers.org/bdgp.est/",
          "http://identifiers.org/bdgp.est",
          "https://identifiers.org/bdgp.est/",
          "https://identifiers.org/bdgp.est",
      }, new Class<?>[]{}, "MIR:00000285",
      new Class<?>[]{},
      "bdgp.est",
      "EY223054.1"),

  BDGP_INSERTION_DB("BDGP insertion DB",
      "http://flypush.imgen.bcm.tmc.edu/pscreen/",
      new String[]{"urn:miriam:bdgp.insertion",
          "http://identifiers.org/bdgp.insertion/",
          "http://identifiers.org/bdgp.insertion",
          "https://identifiers.org/bdgp.insertion/",
          "https://identifiers.org/bdgp.insertion",
      }, new Class<?>[]{}, "MIR:00000156",
      new Class<?>[]{},
      "bdgp.insertion",
      "KG09531"),

  BEETLEBASE("BeetleBase",
      "http://beetlebase.org/",
      new String[]{"urn:miriam:beetlebase",
          "http://identifiers.org/beetlebase/",
          "http://identifiers.org/beetlebase",
          "https://identifiers.org/beetlebase/",
          "https://identifiers.org/beetlebase",
      }, new Class<?>[]{}, "MIR:00000157",
      new Class<?>[]{},
      "beetlebase",
      "TC010103"),
  BGEE_FAMILY("Bgee family",
      "http://bgee.unil.ch/bgee/bgee",
      new String[]{"urn:miriam:bgee.family",
          "http://identifiers.org/bgee.family/",
          "http://identifiers.org/bgee.family",
          "https://identifiers.org/bgee.family/",
          "https://identifiers.org/bgee.family",
      }, new Class<?>[]{}, "MIR:00000417",
      new Class<?>[]{},
      "bgee.family",
      "ENSFM00500000270089"),
  BGEE_GENE("Bgee gene",
      "https://bgee.org/",
      new String[]{"urn:miriam:bgee.gene",
          "http://identifiers.org/bgee.gene/",
          "http://identifiers.org/bgee.gene",
          "https://identifiers.org/bgee.gene/",
          "https://identifiers.org/bgee.gene",
      }, new Class<?>[]{}, "MIR:00000418",
      new Class<?>[]{},
      "bgee.gene",
      "FBgn0000015"),
  BGEE_ORGAN("Bgee organ",
      "http://bgee.unil.ch/bgee/bgee",
      new String[]{"urn:miriam:bgee.organ",
          "http://identifiers.org/bgee.organ/",
          "http://identifiers.org/bgee.organ",
          "https://identifiers.org/bgee.organ/",
          "https://identifiers.org/bgee.organ",
      }, new Class<?>[]{}, "MIR:00000420",
      new Class<?>[]{},
      "bgee.organ",
      "EHDAA:2185"),
  BGEE_STAGE("Bgee stage",
      "http://bgee.unil.ch/bgee/bgee",
      new String[]{"urn:miriam:bgee.stage",
          "http://identifiers.org/bgee.stage/",
          "http://identifiers.org/bgee.stage",
          "https://identifiers.org/bgee.stage/",
          "https://identifiers.org/bgee.stage",
      }, new Class<?>[]{}, "MIR:00000419",
      new Class<?>[]{},
      "bgee.stage",
      "HsapDO:0000004"),

  BiGG_COMPARTMENT("BiGG Compartment",
      "http://bigg.ucsd.edu/compartments/",
      new String[]{"urn:miriam:bigg.compartment",
          "http://identifiers.org/bigg.compartment/",
          "http://identifiers.org/bigg.compartment",
          "https://identifiers.org/bigg.compartment/",
          "https://identifiers.org/bigg.compartment",
      },
      new Class<?>[]{},
      "MIR:00000555",
      new Class<?>[]{},
      "bigg.compartment",
      "c"),

  BiGG_METABOLITE("BiGG Metabolite",
      "http://bigg.ucsd.edu/universal/metabolites",
      new String[]{"urn:miriam:bigg.metabolite",
          "http://identifiers.org/bigg.metabolite/",
          "http://identifiers.org/bigg.metabolite",
          "https://identifiers.org/bigg.metabolite/",
          "https://identifiers.org/bigg.metabolite",
      },
      new Class<?>[]{}, "MIR:00000556",
      new Class<?>[]{},
      "bigg.metabolite",
      "12dgr161"),

  BiGG_REACTIONS("BiGG Reaction",
      "http://bigg.ucsd.edu/universal/reactions",
      new String[]{"urn:miriam:bigg.reaction",
          "http://identifiers.org/bigg.reaction/",
          "http://identifiers.org/bigg.reaction",
          "https://identifiers.org/bigg.reaction/",
          "https://identifiers.org/bigg.reaction",
      },
      new Class<?>[]{}, "MIR:00000557",
      new Class<?>[]{},
      "bigg.reaction",
      "13GS"),

  BINDINGDB("BindingDB",
      "https://www.bindingdb.org",
      new String[]{"urn:miriam:bindingDB",
          "http://identifiers.org/bindingdb/",
          "http://identifiers.org/bindingdb",
          "https://identifiers.org/bindingdb/",
          "https://identifiers.org/bindingdb",
      }, new Class<?>[]{}, "MIR:00000264",
      new Class<?>[]{},
      "bindingdb",
      "e999"),

  BIOCARTA_PATHWAY("BioCarta Pathway",
      "https://www.biocarta.com/",
      new String[]{"urn:miriam:biocarta.pathway",
          "http://identifiers.org/biocarta.pathway/",
          "http://identifiers.org/biocarta.pathway",
          "https://identifiers.org/biocarta.pathway/",
          "https://identifiers.org/biocarta.pathway",
      }, new Class<?>[]{}, "MIR:00000421",
      new Class<?>[]{},
      "biocarta.pathway",
      "h_aktPathway"),
  BIOCATALOGUE("BioCatalogue",
      "https://www.biocatalogue.org/",
      new String[]{"urn:miriam:biocatalogue.service",
          "http://identifiers.org/biocatalogue.service/",
          "http://identifiers.org/biocatalogue.service",
          "https://identifiers.org/biocatalogue.service/",
          "https://identifiers.org/biocatalogue.service",
      }, new Class<?>[]{}, "MIR:00000140",
      new Class<?>[]{},
      "biocatalogue.service",
      "614"),
  BIOCYC("BioCyc",
      "http://biocyc.org",
      new String[]{"urn:miriam:biocyc",
          "http://identifiers.org/biocyc/",
          "http://identifiers.org/biocyc",
          "https://identifiers.org/biocyc/",
          "https://identifiers.org/biocyc",
      }, new Class<?>[]{}, "MIR:00000194",
      new Class<?>[]{},
      "biocyc",
      "ECOLI:CYT-D-UBIOX-CPLX"),

  BIOGRID("BioGRID",
      "http://thebiogrid.org/",
      new String[]{"urn:miriam:biogrid",
          "http://identifiers.org/biogrid/",
          "http://identifiers.org/biogrid",
          "https://identifiers.org/biogrid/",
          "https://identifiers.org/biogrid",
      }, new Class<?>[]{}, "MIR:00000058",
      new Class<?>[]{},
      "biogrid",
      "31623"),

  @NoMiriamUrn
  BIOKC("BioKC",
      "https://biokb.lcsb.uni.lu/",
      new String[]{
          "http://identifiers.org/biokc",
          "http://identifiers.org/biokc/",
          "https://identifiers.org/biokc",
          "https://identifiers.org/biokc/",
      }, new Class<?>[]{}, "MIR:00000995",
      new Class<?>[]{},
      "biokc",
      "bkc640"),

  BIOMODELS_DATABASE("BioModels Database",
      "https://www.ebi.ac.uk/biomodels/",
      new String[]{"urn:miriam:biomodels.db",
          "http://identifiers.org/biomodels.db/",
          "http://identifiers.org/biomodels.db",
          "https://identifiers.org/biomodels.db/",
          "https://identifiers.org/biomodels.db",
      },
      new Class<?>[]{}, "MIR:00000007",
      new Class<?>[]{},
      "biomodels.db",
      "BIOMD0000000048"),

  BIONUMBERS("BioNumbers",
      "https://bionumbers.hms.harvard.edu",
      new String[]{"urn:miriam:bionumbers",
          "http://identifiers.org/bionumbers/",
          "http://identifiers.org/bionumbers",
          "https://identifiers.org/bionumbers/",
          "https://identifiers.org/bionumbers",
      }, new Class<?>[]{}, "MIR:00000101",
      new Class<?>[]{},
      "bionumbers",
      "104674"),

  BIOPORTAL("BioPortal",
      "http://bioportal.bioontology.org/",
      new String[]{"urn:miriam:bioportal",
          "http://identifiers.org/bioportal/",
          "http://identifiers.org/bioportal",
          "https://identifiers.org/bioportal/",
          "https://identifiers.org/bioportal",
      }, new Class<?>[]{}, "MIR:00000187",
      new Class<?>[]{},
      "bioportal",
      "1046"),
  BIOPROJECT("BioProject",
      "http://trace.ddbj.nig.ac.jp/bioproject/",
      new String[]{"urn:miriam:bioproject",
          "http://identifiers.org/bioproject/",
          "http://identifiers.org/bioproject",
          "https://identifiers.org/bioproject/",
          "https://identifiers.org/bioproject",
      }, new Class<?>[]{}, "MIR:00000349",
      new Class<?>[]{},
      "bioproject",
      "PRJDB3"),
  BIOSAMPLE("BioSample",
      "https://www.ebi.ac.uk/biosamples/",
      new String[]{"urn:miriam:biosample",
          "http://identifiers.org/biosample/",
          "http://identifiers.org/biosample",
          "https://identifiers.org/biosample/",
          "https://identifiers.org/biosample",
      }, new Class<?>[]{}, "MIR:00000350",
      new Class<?>[]{},
      "biosample",
      "SAMEA2397676"),

  BIOSYSTEMS("BioSystems",
      "https://www.ncbi.nlm.nih.gov/biosystems/",
      new String[]{"urn:miriam:biosystems",
          "http://identifiers.org/biosystems/",
          "http://identifiers.org/biosystems",
          "https://identifiers.org/biosystems/",
          "https://identifiers.org/biosystems",
      }, new Class<?>[]{}, "MIR:00000097",
      new Class<?>[]{},
      "biosystems",
      "001"),

  BITTERDB_COMPOUND("BitterDB Compound",
      "http://bitterdb.agri.huji.ac.il/dbbitter.php",
      new String[]{"urn:miriam:bitterdb.cpd",
          "http://identifiers.org/bitterdb.cpd/",
          "http://identifiers.org/bitterdb.cpd",
          "https://identifiers.org/bitterdb.cpd/",
          "https://identifiers.org/bitterdb.cpd",
      }, new Class<?>[]{}, "MIR:00000348",
      new Class<?>[]{},
      "bitterdb.cpd",
      "46"),
  BITTERDB_RECEPTOR("BitterDB Receptor",
      "http://bitterdb.agri.huji.ac.il/dbbitter.php",
      new String[]{"urn:miriam:bitterdb.rec",
          "http://identifiers.org/bitterdb.rec/",
          "http://identifiers.org/bitterdb.rec",
          "https://identifiers.org/bitterdb.rec/",
          "https://identifiers.org/bitterdb.rec",
      }, new Class<?>[]{}, "MIR:00000347",
      new Class<?>[]{},
      "bitterdb.rec",
      "1"),
  BOLD_TAXONOMY("BOLD Taxonomy",
      "http://www.boldsystems.org/",
      new String[]{"urn:miriam:bold.taxonomy",
          "http://identifiers.org/bold.taxonomy/",
          "http://identifiers.org/bold.taxonomy",
          "https://identifiers.org/bold.taxonomy/",
          "https://identifiers.org/bold.taxonomy",
      }, new Class<?>[]{}, "MIR:00000158",
      new Class<?>[]{},
      "bold.taxonomy",
      "27267"),
  BROAD_FUNGAL_GENOME_INITIATIVE("Broad Fungal Genome Initiative",
      "https://www.broadinstitute.org/annotation/genome/magnaporthe_grisea/",
      new String[]{"urn:miriam:broad",
          "http://identifiers.org/broad/",
          "http://identifiers.org/broad",
          "https://identifiers.org/broad/",
          "https://identifiers.org/broad",
      }, new Class<?>[]{}, "MIR:00000438",
      new Class<?>[]{},
      "broad",
      "S7000002168151102"),

  /**
   * Brenda enzyme database: http://www.brenda-enzymes.org.
   */
  BRENDA("BRENDA",
      "http://www.brenda-enzymes.org",
      new String[]{"urn:miriam:brenda",
          "http://identifiers.org/brenda/",
          "http://identifiers.org/brenda",
          "https://identifiers.org/brenda/",
          "https://identifiers.org/brenda",
      },
      new Class<?>[]{}, "MIR:00000071",
      new Class<?>[]{},
      "brenda",
      "1.1.1.1"),

  BRENDA_TISSUE_ONTOLOGY("Brenda Tissue Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/bto",
      new String[]{"urn:miriam:bto", "urn:miriam:obo.bto",
          "http://identifiers.org/bto/",
          "http://identifiers.org/obo.bto/",
          "http://identifiers.org/BTO",
          "https://identifiers.org/bto/",
          "https://identifiers.org/obo.bto/",
          "https://identifiers.org/BTO",
      }, new Class<?>[]{}, "MIR:00000111",
      new Class<?>[]{},
      "",
      "BTO:0000146"),

  BUGBASE_EXPT("BugBase Expt",
      "http://bugs.sgul.ac.uk/E-BUGS",
      new String[]{"urn:miriam:bugbase.expt",
          "http://identifiers.org/bugbase.expt/",
          "http://identifiers.org/bugbase.expt",
          "https://identifiers.org/bugbase.expt/",
          "https://identifiers.org/bugbase.expt",
      }, new Class<?>[]{}, "MIR:00000404",
      new Class<?>[]{},
      "bugbase.expt",
      "288"),
  BUGBASE_PROTOCOL("BugBase Protocol",
      "http://bugs.sgul.ac.uk/E-BUGS",
      new String[]{"urn:miriam:bugbase.protocol",
          "http://identifiers.org/bugbase.protocol/",
          "http://identifiers.org/bugbase.protocol",
          "https://identifiers.org/bugbase.protocol/",
          "https://identifiers.org/bugbase.protocol",
      }, new Class<?>[]{}, "MIR:00000403",
      new Class<?>[]{},
      "bugbase.protocol",
      "67"),
  BYKDB("BYKdb",
      "https://bykdb.ibcp.fr/BYKdb/",
      new String[]{"urn:miriam:bykdb",
          "http://identifiers.org/bykdb/",
          "http://identifiers.org/bykdb",
          "https://identifiers.org/bykdb/",
          "https://identifiers.org/bykdb",
      }, new Class<?>[]{}, "MIR:00000253",
      new Class<?>[]{},
      "bykdb",
      "A0A009E7X8"),

  CANADIAN_DRUG_PRODUCT_DATABASE("Canadian Drug Product Database",
      "http://webprod3.hc-sc.gc.ca/dpd-bdpp/index-eng.jsp",
      new String[]{"urn:miriam:cdpd",
          "http://identifiers.org/cdpd/",
          "http://identifiers.org/cdpd",
          "https://identifiers.org/cdpd/",
          "https://identifiers.org/cdpd",
      }, new Class<?>[]{}, "MIR:00000272",
      new Class<?>[]{},
      "cdpd",
      "63250"),

  CANDIDA_GENOME_DATABASE("Candida Genome Database",
      "http://www.candidagenome.org/",
      new String[]{"urn:miriam:cgd",
          "http://identifiers.org/cgd/",
          "http://identifiers.org/cgd",
          "https://identifiers.org/cgd/",
          "https://identifiers.org/cgd",
      }, new Class<?>[]{}, "MIR:00000145",
      new Class<?>[]{},
      "cgd",
      "CAL0003079"),

  CAPS_DB("CAPS-DB",
      "http://www.bioinsilico.org/cgi-bin/CAPSDB/staticHTML/home",
      new String[]{"urn:miriam:caps",
          "http://identifiers.org/caps/",
          "http://identifiers.org/caps",
          "https://identifiers.org/caps/",
          "https://identifiers.org/caps",
      }, new Class<?>[]{}, "MIR:00000396",
      new Class<?>[]{},
      "caps",
      "434"),

  /**
   * Chemical Abstracts Service database: http://commonchemistry.org.
   */
  CAS("Chemical Abstracts Service",
      "http://commonchemistry.org",
      new String[]{"urn:miriam:cas",
          "http://identifiers.org/cas/",
          "http://identifiers.org/cas",
          "https://identifiers.org/cas/",
          "https://identifiers.org/cas",
      },
      new Class<?>[]{}, "MIR:00000237",
      new Class<?>[]{},
      "cas",
      "50-00-0"),

  CATH_DOMAIN("CATH domain",
      "http://www.cathdb.info/",
      new String[]{"urn:miriam:cath.domain",
          "http://identifiers.org/cath.domain/",
          "http://identifiers.org/cath.domain",
          "https://identifiers.org/cath.domain/",
          "https://identifiers.org/cath.domain",
      }, new Class<?>[]{}, "MIR:00000210",
      new Class<?>[]{},
      "cath.domain",
      "1cukA01"),
  CATH_SUPERFAMILY("CATH superfamily",
      "http://www.cathdb.info/",
      new String[]{"urn:miriam:cath.superfamily",
          "http://identifiers.org/cath.superfamily/",
          "http://identifiers.org/cath.superfamily",
          "https://identifiers.org/cath.superfamily/",
          "https://identifiers.org/cath.superfamily",
      }, new Class<?>[]{}, "MIR:00000209",
      new Class<?>[]{},
      "cath.superfamily",
      "1.10.10.200"),

  /**
   * The Carbohydrate-Active Enzyme (CAZy) database: http://www.cazy.org/.
   */
  CAZY("CAZy",
      "http://commonchemistry.org",
      new String[]{"urn:miriam:cazy",
          "http://identifiers.org/cazy/",
          "http://identifiers.org/cazy",
          "https://identifiers.org/cazy/",
          "https://identifiers.org/cazy",
      },
      new Class<?>[]{}, "MIR:00000195",
      new Class<?>[]{},
      "cazy",
      "GT10"),

  /**
   * Consensus CDS: http://identifiers.org/ccds/.
   */
  CCDS("Consensus CDS",
      "http://www.ncbi.nlm.nih.gov/CCDS/",
      new String[]{"urn:miriam:ccds",
          "http://identifiers.org/ccds/",
          "http://identifiers.org/ccds",
          "https://identifiers.org/ccds/",
          "https://identifiers.org/ccds",
      },
      new Class<?>[]{}, "MIR:00000375",
      new Class<?>[]{},
      "ccds",
      "CCDS13573.1"),

  CELL_CYCLE_ONTOLOGY("Cell Cycle Ontology",
      "https://www.ebi.ac.uk/ontology-lookup/browse.do?ontName=CCO",
      new String[]{"urn:miriam:cco",
          "http://identifiers.org/cco/",
          "http://identifiers.org/CCO",
          "https://identifiers.org/cco/",
          "https://identifiers.org/CCO",
      }, new Class<?>[]{}, "MIR:00000234",
      new Class<?>[]{},
      "",
      "CCO:0000003"),

  CELL_IMAGE_LIBRARY("Cell Image Library",
      "http://cellimagelibrary.org/",
      new String[]{"urn:miriam:cellimage",
          "http://identifiers.org/cellimage/",
          "http://identifiers.org/cellimage",
          "https://identifiers.org/cellimage/",
          "https://identifiers.org/cellimage",
      }, new Class<?>[]{}, "MIR:00000257",
      new Class<?>[]{},
      "cellimage",
      "24801"),

  CELL_SIGNALING_TECHNOLOGY_PATHWAYS("Cell Signaling Technology Pathways",
      "http://www.cellsignal.com/pathways/index.html",
      new String[]{"urn:miriam:cst",
          "http://identifiers.org/cst/",
          "http://identifiers.org/cst",
          "https://identifiers.org/cst/",
          "https://identifiers.org/cst",
      }, new Class<?>[]{}, "MIR:00000429",
      new Class<?>[]{},
      "cst",
      "Akt_PKB"),

  CELL_SIGNALING_TECHNOLOGY_ANTIBODY("Cell Signaling Technology Antibody",
      "http://www.cellsignal.com/catalog/index.html",
      new String[]{"urn:miriam:cst.ab",
          "http://identifiers.org/cst.ab/",
          "http://identifiers.org/cst.ab",
          "https://identifiers.org/cst.ab/",
          "https://identifiers.org/cst.ab",
      }, new Class<?>[]{}, "MIR:00000430",
      new Class<?>[]{},
      "cst.ab",
      "3305"),

  CELL_TYPE_ONTOLOGY("Cell Type Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/cl",
      new String[]{"urn:miriam:cl", "urn:miriam:obo.clo",
          "http://identifiers.org/cl/",
          "http://identifiers.org/CL",
          "https://identifiers.org/cl/",
          "https://identifiers.org/CL",
      }, new Class<?>[]{}, "MIR:00000110",
      new Class<?>[]{},
      "",
      "CL:0000232"),

  CGSC_STRAIN("CGSC Strain",
      "http://cgsc.biology.yale.edu/index.php",
      new String[]{"urn:miriam:cgsc",
          "http://identifiers.org/cgsc/",
          "http://identifiers.org/cgsc",
          "https://identifiers.org/cgsc/",
          "https://identifiers.org/cgsc",
      }, new Class<?>[]{}, "MIR:00000295",
      new Class<?>[]{},
      "cgsc",
      "74"),
  CHARPROT("CharProt",
      "http://www.jcvi.org/charprotdb",
      new String[]{"urn:miriam:charprot",
          "http://identifiers.org/charprot/",
          "http://identifiers.org/charprot",
          "https://identifiers.org/charprot/",
          "https://identifiers.org/charprot",
      }, new Class<?>[]{}, "MIR:00000341",
      new Class<?>[]{},
      "charprot",
      "CH_001923"),

  /**
   * Chebi database:
   * <a href = "http://www.ebi.ac.uk/chebi/">http://www.ebi.ac.uk/chebi/</a>.
   */
  CHEBI("Chebi",
      "http://www.ebi.ac.uk/chebi/",
      new String[]{"urn:miriam:obo.chebi", "urn:miriam:chebi",
          "http://identifiers.org/chebi/",
          "http://identifiers.org/obo.chebi/",
          "http://identifiers.org/CHEBI",
          "https://identifiers.org/chebi/",
          "https://identifiers.org/obo.chebi/",
          "https://identifiers.org/CHEBI",
      },
      new Class<?>[]{Chemical.class, Drug.class,}, "MIR:00000002",
      new Class<?>[]{Chemical.class},
      "",
      "CHEBI:36927"),

  CHEMDB("ChemDB",
      "http://cdb.ics.uci.edu/",
      new String[]{"urn:miriam:chemdb",
          "http://identifiers.org/chemdb/",
          "http://identifiers.org/chemdb",
          "https://identifiers.org/chemdb/",
          "https://identifiers.org/chemdb",
      }, new Class<?>[]{}, "MIR:00000279",
      new Class<?>[]{},
      "chemdb",
      "3966782"),

  CHEM_ID_PLUS("ChemIDplus",
      "https://chem.nlm.nih.gov/chemidplus/",
      new String[]{"urn:miriam:chemidplus",
          "http://identifiers.org/chemidplus/",
          "http://identifiers.org/chemidplus",
          "https://identifiers.org/chemidplus/",
          "https://identifiers.org/chemidplus",
      },
      new Class<?>[]{Chemical.class, Drug.class,}, "MIR:00000096",
      new Class<?>[]{Chemical.class},
      "chemidplus",
      "57-27-2"),

  /**
   * ChemSpider database:
   * <a href = "http://www.chemspider.com/">http://www.chemspider.com/</a>.
   */
  CHEMSPIDER("ChemSpider",
      "http://www.chemspider.com/",
      new String[]{"urn:miriam:chemspider",
          "http://identifiers.org/chemspider/",
          "http://identifiers.org/chemspider",
          "https://identifiers.org/chemspider/",
          "https://identifiers.org/chemspider",
      },
      new Class<?>[]{}, "MIR:00000138",
      new Class<?>[]{},
      "chemspider",
      "56586"),

  /**
   * Chembl database: https://www.ebi.ac.uk/chembldb/.
   */
  CHEMBL_COMPOUND("ChEMBL",
      "https://www.ebi.ac.uk/chembldb/",
      new String[]{"urn:miriam:chembl.compound",
          "http://identifiers.org/chembl.compound/",
          "http://identifiers.org/chembl.compound",
          "https://identifiers.org/chembl.compound/",
          "https://identifiers.org/chembl.compound",
      },
      new Class<?>[]{Drug.class}, "MIR:00000084",
      new Class<?>[]{},
      "chembl.compound",
      "CHEMBL308052"),

  /**
   * Target in chembl database: https://www.ebi.ac.uk/chembldb/.
   */
  CHEMBL_TARGET("ChEMBL target",
      "https://www.ebi.ac.uk/chembldb/",
      new String[]{"urn:miriam:chembl.target",
          "http://identifiers.org/chembl.target/",
          "http://identifiers.org/chembl.target",
          "https://identifiers.org/chembl.target/",
          "https://identifiers.org/chembl.target",
      },
      new Class<?>[]{Protein.class, Complex.class}, "MIR:00000085",
      new Class<?>[]{},
      "chembl.target",
      "CHEMBL3467"),

  CLDB("CLDB",
      "http://bioinformatics.hsanmartino.it/hypercldb/indexes.html",
      new String[]{"urn:miriam:cldb",
          "http://identifiers.org/cldb/",
          "http://identifiers.org/cldb",
          "https://identifiers.org/cldb/",
          "https://identifiers.org/cldb",
      }, new Class<?>[]{}, "MIR:00000390",
      new Class<?>[]{},
      "cldb",
      "cl3603"),

  CLINICAL_TRIALS_GOV("ClinicalTrials.gov",
      "https://clinicaltrials.gov/",
      new String[]{"urn:miriam:clinicaltrials",
          "http://identifiers.org/clinicaltrials/",
          "http://identifiers.org/clinicaltrials",
          "https://identifiers.org/clinicaltrials/",
          "https://identifiers.org/clinicaltrials",
      },
      new Class<?>[]{}, "MIR:00000137",
      new Class<?>[]{},
      "clinicaltrials",
      "NCT00222573"),

  CLINVAR_RECORD("ClinVar Record",
      "http://www.ncbi.nlm.nih.gov/clinvar/",
      new String[]{"urn:miriam:clinvar.record",
          "http://identifiers.org/clinvar.record/",
          "http://identifiers.org/clinvar.record",
          "https://identifiers.org/clinvar.record/",
          "https://identifiers.org/clinvar.record",
      }, new Class<?>[]{}, "MIR:00000534",
      new Class<?>[]{},
      "clinvar.record",
      "RCV000033555.3"),

  /**
   * Clusters of Orthologous Groups: https://www.ncbi.nlm.nih.gov/COG/.
   */
  @Deprecated
  COG("Clusters of Orthologous Groups",
      "https://www.ncbi.nlm.nih.gov/COG/",
      new String[]{"urn:miriam:cogs"},
      new Class<?>[]{Reaction.class}, "MIR:00000296",
      new Class<?>[]{},
      "cogs",
      "COG0001"),

  COMBINE_SPECIFICATIONS("COMBINE specifications",
      "https://co.mbine.org/standards/",
      new String[]{"urn:miriam:combine.specifications",
          "http://identifiers.org/combine.specifications/",
          "http://identifiers.org/combine.specifications",
          "https://identifiers.org/combine.specifications/",
          "https://identifiers.org/combine.specifications",
      }, new Class<?>[]{}, "MIR:00000258",
      new Class<?>[]{},
      "combine.specifications",
      "sbgn.er.level-1.version-1.2"),

  @NoMiriamUrn
  COMPLEX_PORTAL("Complex Portal",
      "https://www.ebi.ac.uk/complexportal",
      new String[]{
          "https://identifiers.org/complexportal",
          "https://identifiers.org/complexportal/",
          "http://identifiers.org/complexportal",
          "http://identifiers.org/complexportal/",
      },
      new Class<?>[]{}, "MIR:00000657",
      new Class<?>[]{},
      "complexportal",
      "CPX-2158"),

  COMPULYEAST("Compulyeast",
      "http://compluyeast2dpage.dacya.ucm.es/",
      new String[]{"urn:miriam:compulyeast",
          "http://identifiers.org/compulyeast/",
          "http://identifiers.org/compulyeast",
          "https://identifiers.org/compulyeast/",
          "https://identifiers.org/compulyeast",
      }, new Class<?>[]{}, "MIR:00000198",
      new Class<?>[]{},
      "compulyeast",
      "O08709"),

  CONSERVED_DOMAIN_DATABASE("Conserved Domain Database",
      "https://www.ncbi.nlm.nih.gov/sites/entrez?db=cdd",
      new String[]{"urn:miriam:cdd",
          "http://identifiers.org/cdd/",
          "http://identifiers.org/cdd",
          "https://identifiers.org/cdd/",
          "https://identifiers.org/cdd",
      }, new Class<?>[]{}, "MIR:00000119",
      new Class<?>[]{},
      "cdd",
      "cd00400"),

  CONOSERVER("Conoserver",
      "http://www.conoserver.org/",
      new String[]{"urn:miriam:conoserver",
          "http://identifiers.org/conoserver/",
          "http://identifiers.org/conoserver",
          "https://identifiers.org/conoserver/",
          "https://identifiers.org/conoserver",
      }, new Class<?>[]{}, "MIR:00000254",
      new Class<?>[]{},
      "conoserver",
      "2639"),
  CORIELL_CELL_REPOSITORIES("Coriell Cell Repositories",
      "http://ccr.coriell.org/",
      new String[]{"urn:miriam:coriell",
          "http://identifiers.org/coriell/",
          "http://identifiers.org/coriell",
          "https://identifiers.org/coriell/",
          "https://identifiers.org/coriell",
      }, new Class<?>[]{}, "MIR:00000439",
      new Class<?>[]{},
      "coriell",
      "GM17027"),
  CORUM("CORUM",
      "https://mips.helmholtz-muenchen.de/genre/proj/corum/",
      new String[]{"urn:miriam:corum",
          "http://identifiers.org/corum/",
          "http://identifiers.org/corum",
          "https://identifiers.org/corum/",
          "https://identifiers.org/corum",
      }, new Class<?>[]{}, "MIR:00000440",
      new Class<?>[]{},
      "corum",
      "100"),

  CPC("Cooperative Patent Classification",
      "https://worldwide.espacenet.com/classification",
      new String[]{"urn:miriam:cpc",
          "https://identifiers.org/cpc",
          "https://identifiers.org/cpc/",
          "http://identifiers.org/cpc",
          "http://identifiers.org/cpc/",
      },
      new Class<?>[]{}, "MIR:00000539",
      new Class<?>[]{},
      "cpc",
      "A01M1/026"),

  CRYPTODB("CryptoDB",
      "https://cryptodb.org/cryptodb/",
      new String[]{"urn:miriam:cryptodb",
          "http://identifiers.org/cryptodb/",
          "http://identifiers.org/cryptodb",
          "https://identifiers.org/cryptodb/",
          "https://identifiers.org/cryptodb",
      }, new Class<?>[]{}, "MIR:00000149",
      new Class<?>[]{},
      "cryptodb",
      "cgd7_230"),
  CSA("CSA",
      "https://www.ebi.ac.uk/thornton-srv/databases/CSA/",
      new String[]{"urn:miriam:csa",
          "http://identifiers.org/csa/",
          "http://identifiers.org/csa",
          "https://identifiers.org/csa/",
          "https://identifiers.org/csa",
      }, new Class<?>[]{}, "MIR:00000144",
      new Class<?>[]{},
      "csa",
      "1a05"),

  CTD_GENE("CTD Gene",
      "http://ctdbase.org/",
      new String[]{"urn:miriam:ctd.gene",
          "http://identifiers.org/ctd.gene/",
          "http://identifiers.org/ctd.gene",
          "https://identifiers.org/ctd.gene/",
          "https://identifiers.org/ctd.gene",
      }, new Class<?>[]{}, "MIR:00000100",
      new Class<?>[]{},
      "ctd.gene",
      "101"),

  CTD_DISEASE("CTD Disease",
      "http://ctdbase.org/",
      new String[]{"urn:miriam:ctd.disease",
          "http://identifiers.org/ctd.disease/",
          "http://identifiers.org/ctd.disease",
          "https://identifiers.org/ctd.disease/",
          "https://identifiers.org/ctd.disease",
      }, new Class<?>[]{}, "MIR:00000099",
      new Class<?>[]{},
      "ctd.disease",
      "D053716"),

  CUTDB("CutDB",
      "http://cutdb.burnham.org",
      new String[]{"urn:miriam:pmap.cutdb",
          "http://identifiers.org/pmap.cutdb/",
          "http://identifiers.org/pmap.cutdb",
          "https://identifiers.org/pmap.cutdb/",
          "https://identifiers.org/pmap.cutdb",
      }, new Class<?>[]{}, "MIR:00000225",
      new Class<?>[]{},
      "pmap.cutdb",
      "25782"),

  CUBE_DB("Cube db",
      "http://epsf.bmad.bii.a-star.edu.sg/cube/db/html/home.html",
      new String[]{"urn:miriam:cubedb",
          "http://identifiers.org/cubedb/",
          "http://identifiers.org/cubedb",
          "https://identifiers.org/cubedb/",
          "https://identifiers.org/cubedb",
      }, new Class<?>[]{}, "MIR:00000397",
      new Class<?>[]{},
      "cubedb",
      "AKR"),

  DAILYMED("DailyMed",
      "https://dailymed.nlm.nih.gov/dailymed/",
      new String[]{"urn:miriam:dailymed",
          "http://identifiers.org/dailymed/",
          "http://identifiers.org/dailymed",
          "https://identifiers.org/dailymed/",
          "https://identifiers.org/dailymed",
      }, new Class<?>[]{}, "MIR:00000434",
      new Class<?>[]{},
      "dailymed",
      "973a9333-fec7-46dd-8eb5-25738f06ee54"),
  DARC("DARC",
      "http://darcsite.genzentrum.lmu.de/darc/index.php",
      new String[]{"urn:miriam:darc",
          "http://identifiers.org/darc/",
          "http://identifiers.org/darc",
          "https://identifiers.org/darc/",
          "https://identifiers.org/darc",
      }, new Class<?>[]{}, "MIR:00000366",
      new Class<?>[]{},
      "darc",
      "1250"),

  DATABASE_OF_INTERACTING_PROTEINS("Database of Interacting Proteins",
      "https://dip.doe-mbi.ucla.edu/",
      new String[]{"urn:miriam:dip",
          "http://identifiers.org/dip/",
          "http://identifiers.org/dip",
          "https://identifiers.org/dip/",
          "https://identifiers.org/dip",
      }, new Class<?>[]{}, "MIR:00000044",
      new Class<?>[]{},
      "dip",
      "DIP-743N"),

  DATABASE_OF_QUANTITATIVE_CELLULAR_SIGNALING_MODEL("Database of Quantitative Cellular Signaling: Model",
      "http://doqcs.ncbs.res.in/",
      new String[]{"urn:miriam:doqcs.model",
          "http://identifiers.org/doqcs.model/",
          "http://identifiers.org/doqcs.model",
          "https://identifiers.org/doqcs.model/",
          "https://identifiers.org/doqcs.model",
      }, new Class<?>[]{}, "MIR:00000134",
      new Class<?>[]{},
      "doqcs.model",
      "57"),

  DATABASE_OF_QUANTITATIVE_CELLULAR_SIGNALING_PATHWAY("Database of Quantitative Cellular Signaling: Pathway",
      "http://doqcs.ncbs.res.in/",
      new String[]{"urn:miriam:doqcs.pathway",
          "http://identifiers.org/doqcs.pathway/",
          "http://identifiers.org/doqcs.pathway",
          "https://identifiers.org/doqcs.pathway/",
          "https://identifiers.org/doqcs.pathway",
      }, new Class<?>[]{}, "MIR:00000135",
      new Class<?>[]{},
      "doqcs.pathway",
      "131"),

  DATF("DATF",
      "http://datf.cbi.pku.edu.cn/",
      new String[]{"urn:miriam:datf",
          "http://identifiers.org/datf/",
          "http://identifiers.org/datf",
          "https://identifiers.org/datf/",
          "https://identifiers.org/datf",
      }, new Class<?>[]{}, "MIR:00000456",
      new Class<?>[]{},
      "datf",
      "AT1G01030.1"),
  DBD("DBD",
      "http://www.transcriptionfactor.org/",
      new String[]{"urn:miriam:dbd",
          "http://identifiers.org/dbd/",
          "http://identifiers.org/dbd",
          "https://identifiers.org/dbd/",
          "https://identifiers.org/dbd",
      }, new Class<?>[]{}, "MIR:00000455",
      new Class<?>[]{},
      "dbd",
      "0045310"),
  DBEST("dbEST",
      "https://www.ncbi.nlm.nih.gov/nucest",
      new String[]{"urn:miriam:dbest",
          "http://identifiers.org/dbest/",
          "http://identifiers.org/dbest",
          "https://identifiers.org/dbest/",
          "https://identifiers.org/dbest",
      }, new Class<?>[]{}, "MIR:00000159",
      new Class<?>[]{},
      "dbest",
      "BP100000"),
  DBG2_INTRONS("DBG2 Introns",
      "http://webapps2.ucalgary.ca/~groupii/",
      new String[]{"urn:miriam:dbg2introns",
          "http://identifiers.org/dbg2introns/",
          "http://identifiers.org/dbg2introns",
          "https://identifiers.org/dbg2introns/",
          "https://identifiers.org/dbg2introns",
      }, new Class<?>[]{}, "MIR:00000318",
      new Class<?>[]{},
      "dbg2introns",
      "Cu.me.I1"),
  DBPROBE("dbProbe",
      "https://www.ncbi.nlm.nih.gov/sites/entrez?db=probe",
      new String[]{"urn:miriam:dbprobe",
          "http://identifiers.org/dbprobe/",
          "http://identifiers.org/dbprobe",
          "https://identifiers.org/dbprobe/",
          "https://identifiers.org/dbprobe",
      }, new Class<?>[]{}, "MIR:00000160",
      new Class<?>[]{},
      "dbprobe",
      "1000000"),

  /**
   * Digital Object Identifier: http://www.doi.org/.
   */
  DB_SNP("dbSNP at NCBI",
      "https://www.ncbi.nlm.nih.gov/snp/",
      new String[]{"urn:miriam:dbsnp",
          "http://identifiers.org/dbsnp/",
          "http://identifiers.org/dbsnp",
          "https://identifiers.org/dbsnp/",
          "https://identifiers.org/dbsnp",
      },
      new Class<?>[]{}, "MIR:00000161",
      new Class<?>[]{},
      "dbsnp",
      "rs121909098"),

  DEGRADOME_DATABASE("Degradome Database",
      "http://degradome.uniovi.es/",
      new String[]{"urn:miriam:degradome",
          "http://identifiers.org/degradome/",
          "http://identifiers.org/degradome",
          "https://identifiers.org/degradome/",
          "https://identifiers.org/degradome",
      }, new Class<?>[]{}, "MIR:00000454",
      new Class<?>[]{},
      "degradome",
      "Ax1"),
  DEPOD("DEPOD",
      "http://www.depod.bioss.uni-freiburg.de",
      new String[]{"urn:miriam:depod",
          "http://identifiers.org/depod/",
          "http://identifiers.org/depod",
          "https://identifiers.org/depod/",
          "https://identifiers.org/depod",
      }, new Class<?>[]{}, "MIR:00000428",
      new Class<?>[]{},
      "depod",
      "PTPN1"),
  DICTYBASE_EST("Dictybase EST",
      "http://dictybase.org/",
      new String[]{"urn:miriam:dictybase.est",
          "http://identifiers.org/dictybase.est/",
          "http://identifiers.org/dictybase.est",
          "https://identifiers.org/dictybase.est/",
          "https://identifiers.org/dictybase.est",
      }, new Class<?>[]{}, "MIR:00000330",
      new Class<?>[]{},
      "dictybase.est",
      "DDB0016567"),
  DICTYBASE_GENE("Dictybase Gene",
      "http://dictybase.org/",
      new String[]{"urn:miriam:dictybase.gene",
          "http://identifiers.org/dictybase.gene/",
          "http://identifiers.org/dictybase.gene",
          "https://identifiers.org/dictybase.gene/",
          "https://identifiers.org/dictybase.gene",
      }, new Class<?>[]{}, "MIR:00000286",
      new Class<?>[]{},
      "dictybase.gene",
      "DDB_G0267522"),
  DISPROT("DisProt",
      "https://disprot.org/",
      new String[]{"urn:miriam:disprot",
          "http://identifiers.org/disprot/",
          "http://identifiers.org/disprot",
          "https://identifiers.org/disprot/",
          "https://identifiers.org/disprot",
      }, new Class<?>[]{}, "MIR:00000199",
      new Class<?>[]{},
      "disprot",
      "DP00003"),

  /**
   * Digital Object Identifier: http://www.doi.org/.
   */
  DOI("Digital Object Identifier",
      "http://www.doi.org/",
      new String[]{"urn:miriam:doi",
          "http://identifiers.org/doi/",
          "http://identifiers.org/doi",
          "https://identifiers.org/doi/",
          "https://identifiers.org/doi",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000019",
      new Class<?>[]{},
      "doi",
      "10.1038/nbt1156"),

  DOMMINO("DOMMINO",
      "http://dommino.org/",
      new String[]{"urn:miriam:dommino",
          "http://identifiers.org/dommino/",
          "http://identifiers.org/dommino",
          "https://identifiers.org/dommino/",
          "https://identifiers.org/dommino",
      }, new Class<?>[]{}, "MIR:00000373",
      new Class<?>[]{},
      "dommino",
      "2GC4"),
  DOOR("DOOR",
      "http://csbl.bmb.uga.edu/DOOR/operon.php",
      new String[]{"urn:miriam:door",
          "http://identifiers.org/door/",
          "http://identifiers.org/door",
          "https://identifiers.org/door/",
          "https://identifiers.org/door",
      }, new Class<?>[]{}, "MIR:00000453",
      new Class<?>[]{},
      "door",
      "1398574"),

  DPV("DPV",
      "http://www.dpvweb.net/",
      new String[]{"urn:miriam:dpv",
          "http://identifiers.org/dpv/",
          "http://identifiers.org/dpv",
          "https://identifiers.org/dpv/",
          "https://identifiers.org/dpv",
      }, new Class<?>[]{}, "MIR:00000280",
      new Class<?>[]{},
      "dpv",
      "229"),

  DRAGONDB_ALLELE("DragonDB Allele",
      "http://www.antirrhinum.net/",
      new String[]{"urn:miriam:dragondb.allele",
          "http://identifiers.org/dragondb.allele/",
          "http://identifiers.org/dragondb.allele",
          "https://identifiers.org/dragondb.allele/",
          "https://identifiers.org/dragondb.allele",
      }, new Class<?>[]{}, "MIR:00000300",
      new Class<?>[]{},
      "dragondb.allele",
      "cho"),
  DRAGONDB_DNA("DragonDB DNA",
      "http://www.antirrhinum.net/",
      new String[]{"urn:miriam:dragondb.dna",
          "http://identifiers.org/dragondb.dna/",
          "http://identifiers.org/dragondb.dna",
          "https://identifiers.org/dragondb.dna/",
          "https://identifiers.org/dragondb.dna",
      }, new Class<?>[]{}, "MIR:00000297",
      new Class<?>[]{},
      "dragondb.dna",
      "3hB06"),
  DRAGONDB_LOCUS("DragonDB Locus",
      "http://www.antirrhinum.net/",
      new String[]{"urn:miriam:dragondb.locus",
          "http://identifiers.org/dragondb.locus/",
          "http://identifiers.org/dragondb.locus",
          "https://identifiers.org/dragondb.locus/",
          "https://identifiers.org/dragondb.locus",
      }, new Class<?>[]{}, "MIR:00000299",
      new Class<?>[]{},
      "dragondb.locus",
      "DEF"),
  DRAGONDB_PROTEIN("DragonDB Protein",
      "http://www.antirrhinum.net/",
      new String[]{"urn:miriam:dragondb.protein",
          "http://identifiers.org/dragondb.protein/",
          "http://identifiers.org/dragondb.protein",
          "https://identifiers.org/dragondb.protein/",
          "https://identifiers.org/dragondb.protein",
      }, new Class<?>[]{}, "MIR:00000298",
      new Class<?>[]{},
      "dragondb.protein",
      "AMDEFA"),
  DRSC("DRSC",
      "http://flyrnai.org/",
      new String[]{"urn:miriam:drsc",
          "http://identifiers.org/drsc/",
          "http://identifiers.org/drsc",
          "https://identifiers.org/drsc/",
          "https://identifiers.org/drsc",
      }, new Class<?>[]{}, "MIR:00000367",
      new Class<?>[]{},
      "drsc",
      "DRSC05221"),

  /**
   * Drugbank database: http://www.drugbank.ca/.
   */
  DRUGBANK("DrugBank",
      "http://www.drugbank.ca/",
      new String[]{"urn:miriam:drugbank",
          "http://identifiers.org/drugbank/",
          "http://identifiers.org/drugbank",
          "https://identifiers.org/drugbank/",
          "https://identifiers.org/drugbank",
      },
      new Class<?>[]{Drug.class}, "MIR:00000102",
      new Class<?>[]{},
      "drugbank",
      "DB00001"),

  /**
   * Drugbank targets: http://www.drugbank.ca/targets.
   */
  DRUGBANK_TARGET_V4("DrugBank Target v4",
      "http://www.drugbank.ca/targets",
      new String[]{"urn:miriam:drugbankv4.target", "urn:miriam:drugbank.target",
          "http://identifiers.org/drugbankv4.target/",
          "http://identifiers.org/drugbankv4.target",
          "https://identifiers.org/drugbankv4.target/",
          "https://identifiers.org/drugbankv4.target",
      },
      new Class<?>[]{}, "MIR:00000528",
      new Class<?>[]{},
      "drugbankv4.target",
      "BE0000048"),

  /**
   * Enzyme Nomenclature: http://www.enzyme-database.org/.
   */
  EC("Enzyme Nomenclature",
      "http://www.enzyme-database.org/",
      new String[]{"urn:miriam:ec-code", "urn:lsid:ec-code.org",
          "http://identifiers.org/ec-code/",
          "http://identifiers.org/ec-code",
          "https://identifiers.org/ec-code/",
          "https://identifiers.org/ec-code",
      },
      new Class<?>[]{Protein.class, Complex.class}, "MIR:00000004",
      new Class<?>[]{},
      "ec-code",
      "1.1.1.2"),

  ECHOBASE("EchoBASE",
      "http://www.york.ac.uk/",
      new String[]{"urn:miriam:echobase",
          "http://identifiers.org/echobase/",
          "http://identifiers.org/echobase",
          "https://identifiers.org/echobase/",
          "https://identifiers.org/echobase",
      }, new Class<?>[]{}, "MIR:00000200",
      new Class<?>[]{},
      "echobase",
      "EB0170"),

  ECO("Evidence Code Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/eco",
      new String[]{"urn:miriam:obo.eco", "urn:miriam:eco",
          "http://identifiers.org/eco/",
          "http://identifiers.org/obo.eco/",
          "http://identifiers.org/ECO",
          "https://identifiers.org/eco/",
          "https://identifiers.org/obo.eco/",
          "https://identifiers.org/ECO",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000055",
      new Class<?>[]{},
      "",
      "ECO:0000006"),

  ECOGENE("EcoGene",
      "http://ecogene.org/",
      new String[]{"urn:miriam:ecogene",
          "http://identifiers.org/ecogene/",
          "http://identifiers.org/ecogene",
          "https://identifiers.org/ecogene/",
          "https://identifiers.org/ecogene",
      },
      new Class<?>[]{}, "MIR:00000163",
      new Class<?>[]{},
      "ecogene",
      "EG10173"),

  ECOLIWIKI("EcoliWiki",
      "http://ecoliwiki.net/colipedia/",
      new String[]{"urn:miriam:ecoliwiki",
          "http://identifiers.org/ecoliwiki/",
          "http://identifiers.org/ecoliwiki",
          "https://identifiers.org/ecoliwiki/",
          "https://identifiers.org/ecoliwiki",
      }, new Class<?>[]{}, "MIR:00000442",
      new Class<?>[]{},
      "ecoliwiki",
      "aaeA"),
  EDAM_ONTOLOGY("EDAM Ontology",
      "http://bioportal.bioontology.org/ontologies/EDAM",
      new String[]{"urn:miriam:edam",
          "http://identifiers.org/edam/",
          "http://identifiers.org/edam",
          "https://identifiers.org/edam/",
          "https://identifiers.org/edam",
      }, new Class<?>[]{}, "MIR:00000189",
      new Class<?>[]{},
      "edam",
      "data_1664"),

  EGGNOG("eggNOG",
      "http://eggnog.embl.de/version_3.0/",
      new String[]{"urn:miriam:eggnog",
          "http://identifiers.org/eggnog/",
          "http://identifiers.org/eggnog",
          "https://identifiers.org/eggnog/",
          "https://identifiers.org/eggnog",
      }, new Class<?>[]{}, "MIR:00000201",
      new Class<?>[]{},
      "eggnog",
      "veNOG12876"),

  ELM("ELM",
      "http://elm.eu.org/",
      new String[]{"urn:miriam:elm",
          "http://identifiers.org/elm/",
          "http://identifiers.org/elm",
          "https://identifiers.org/elm/",
          "https://identifiers.org/elm",
      }, new Class<?>[]{}, "MIR:00000250",
      new Class<?>[]{},
      "elm",
      "CLV_MEL_PAP_1"),
  ENA("ENA",
      "https://www.ncbi.nlm.nih.gov/Genbank/",
      new String[]{"urn:miriam:ena.embl",
          "http://identifiers.org/ena.embl/",
          "http://identifiers.org/ena.embl",
          "https://identifiers.org/ena.embl/",
          "https://identifiers.org/ena.embl",
      }, new Class<?>[]{}, "MIR:00000372",
      new Class<?>[]{},
      "ena.embl",
      "BN000065"),

  /**
   * Ensembl: www.ensembl.org.
   */
  ENSEMBL("Ensembl",
      "https://www.ensembl.org/",
      new String[]{"urn:miriam:ensembl",
          "http://identifiers.org/ensembl/",
          "http://identifiers.org/ensembl.gene/",
          "http://identifiers.org/ensembl",
          "https://identifiers.org/ensembl/",
          "https://identifiers.org/ensembl.gene/",
          "https://identifiers.org/ensembl",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000003",
      new Class<?>[]{},
      "ensembl",
      "ENSG00000139618"),

  ENSEMBL_BACTERIA("Ensembl Bacteria",
      "https://bacteria.ensembl.org/",
      new String[]{"urn:miriam:ensembl.bacteria",
          "http://identifiers.org/ensembl.bacteria/",
          "http://identifiers.org/ensembl.bacteria",
          "https://identifiers.org/ensembl.bacteria/",
          "https://identifiers.org/ensembl.bacteria",
      }, new Class<?>[]{}, "MIR:00000202",
      new Class<?>[]{},
      "ensembl.bacteria",
      "MU9_3181"),

  ENSEMBL_FUNGI("Ensembl Fungi",
      "https://fungi.ensembl.org/",
      new String[]{"urn:miriam:ensembl.fungi",
          "http://identifiers.org/ensembl.fungi/",
          "http://identifiers.org/ensembl.fungi",
          "https://identifiers.org/ensembl.fungi/",
          "https://identifiers.org/ensembl.fungi",
      }, new Class<?>[]{}, "MIR:00000206",
      new Class<?>[]{},
      "ensembl.fungi",
      "CADAFLAT00006211"),

  ENSEMBL_METAZOA("Ensembl Metazoa",
      "https://metazoa.ensembl.org/",
      new String[]{"urn:miriam:ensembl.metazoa",
          "http://identifiers.org/ensembl.metazoa/",
          "http://identifiers.org/ensembl.metazoa",
          "https://identifiers.org/ensembl.metazoa/",
          "https://identifiers.org/ensembl.metazoa",
      }, new Class<?>[]{}, "MIR:00000204",
      new Class<?>[]{},
      "ensembl.metazoa",
      "FBtr0084214"),

  /**
   * Ensembl Plants: http://plants.ensembl.org/.
   */
  ENSEMBL_PLANTS("Ensembl Plants",
      "http://plants.ensembl.org/",
      new String[]{"urn:miriam:ensembl.plant",
          "http://identifiers.org/ensembl.plant/",
          "http://identifiers.org/ensembl.plant",
          "https://identifiers.org/ensembl.plant/",
          "https://identifiers.org/ensembl.plant",
      },
      new Class<?>[]{}, "MIR:00000205",
      new Class<?>[]{},
      "ensembl.plant",
      "AT1G73965"),

  ENSEMBL_PROTISTS("Ensembl Protists",
      "https://protists.ensembl.org",
      new String[]{"urn:miriam:ensembl.protist",
          "http://identifiers.org/ensembl.protist/",
          "http://identifiers.org/ensembl.protist",
          "https://identifiers.org/ensembl.protist/",
          "https://identifiers.org/ensembl.protist",
      }, new Class<?>[]{}, "MIR:00000203",
      new Class<?>[]{},
      "ensembl.protist",
      "PF3D7_1328700"),

  /**
   * Entrez Gene: http://www.ncbi.nlm.nih.gov/gene.
   */
  ENTREZ("Entrez Gene",
      "http://www.ncbi.nlm.nih.gov/gene",
      new String[]{"urn:miriam:ncbigene", "urn:miriam:ncbi.gene",
          "urn:miriam:entrez.gene",
          "http://identifiers.org/ncbigene/",
          "http://identifiers.org/ncbigene",
          "https://identifiers.org/ncbigene/",
          "https://identifiers.org/ncbigene",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000069",
      new Class<?>[]{},
      "ncbigene",
      "100010"),

  @NoMiriamUrn
  ENVIPATH("enviPath",
      "https://envipath.org/",
      new String[]{
          "http://identifiers.org/envipath/",
          "http://identifiers.org/envipath",
          "https://identifiers.org/envipath/",
          "https://identifiers.org/envipath",
      },
      new Class<?>[]{}, "MIR:00000727",
      new Class<?>[]{},
      "envipath",
      "32de3cf4-e3e6-4168-956e-32fa5ddb0ce1/compound/b545cabc-8c9e-4b20-8848-efa015b481ea"),

  EPD("EPD",
      "http://epd.vital-it.ch/",
      new String[]{"urn:miriam:epd",
          "http://identifiers.org/epd/",
          "http://identifiers.org/epd",
          "https://identifiers.org/epd/",
          "https://identifiers.org/epd",
      }, new Class<?>[]{}, "MIR:00000408",
      new Class<?>[]{},
      "epd",
      "TA_H3"),

  EU_CLINICAL_TRIALS("EU Clinical Trials",
      "https://www.clinicaltrialsregister.eu/",
      new String[]{"urn:miriam:euclinicaltrials",
          "http://identifiers.org/euclinicaltrials/",
          "http://identifiers.org/euclinicaltrials",
          "https://identifiers.org/euclinicaltrials/",
          "https://identifiers.org/euclinicaltrials",
      },
      new Class<?>[]{}, "MIR:00000536",
      new Class<?>[]{},
      "euclinicaltrials",
      "2008-005144-16"),

  EXAC_GENE("ExAC Gene",
      "http://exac.broadinstitute.org/",
      new String[]{"urn:miriam:exac.gene",
          "http://identifiers.org/exac.gene/",
          "http://identifiers.org/exac.gene",
          "https://identifiers.org/exac.gene/",
          "https://identifiers.org/exac.gene",
      },
      new Class<?>[]{}, "MIR:00000548",
      new Class<?>[]{},
      "exac.gene",
      "ENSG00000169174"),

  EXAC_TRANSCRIPT("ExAC Transcript",
      "http://exac.broadinstitute.org/",
      new String[]{"urn:miriam:exac.transcript",
          "http://identifiers.org/exac.transcript/",
          "http://identifiers.org/exac.transcript",
          "https://identifiers.org/exac.transcript/",
          "https://identifiers.org/exac.transcript",
      },
      new Class<?>[]{}, "MIR:00000547",
      new Class<?>[]{},
      "exac.transcript",
      "ENST00000407236"),

  EXAC_VARIANT("ExAC Variant",
      "http://exac.broadinstitute.org/",
      new String[]{"urn:miriam:exac.variant",
          "http://identifiers.org/exac.variant/",
          "http://identifiers.org/exac.variant",
          "https://identifiers.org/exac.variant/",
          "https://identifiers.org/exac.variant",
      },
      new Class<?>[]{}, "MIR:00000541",
      new Class<?>[]{},
      "exac.variant",
      "22-46615880-T-C"),

  EXPERIMENTAL_FACTOR_ONTOLOGY("Experimental Factor Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/efo",
      new String[]{"urn:miriam:efo",
          "http://identifiers.org/efo/",
          "http://identifiers.org/efo",
          "https://identifiers.org/efo/",
          "https://identifiers.org/efo",
      }, new Class<?>[]{}, "MIR:00000391",
      new Class<?>[]{},
      "efo",
      "0004859"),
  EUROPEAN_GENOME_PHENOME_ARCHIVE_DATASET("European Genome-phenome Archive Dataset",
      "https://www.ebi.ac.uk/ega/dataset",
      new String[]{"urn:miriam:ega.dataset",
          "http://identifiers.org/ega.dataset/",
          "http://identifiers.org/ega.dataset",
          "https://identifiers.org/ega.dataset/",
          "https://identifiers.org/ega.dataset",
      }, new Class<?>[]{}, "MIR:00000512",
      new Class<?>[]{},
      "ega.dataset",
      "EGAD00000000001"),
  EUROPEAN_GENOME_PHENOME_ARCHIVE_STUDY("European Genome-phenome Archive Study",
      "https://www.ebi.ac.uk/ega/studies",
      new String[]{"urn:miriam:ega.study",
          "http://identifiers.org/ega.study/",
          "http://identifiers.org/ega.study",
          "https://identifiers.org/ega.study/",
          "https://identifiers.org/ega.study",
      }, new Class<?>[]{}, "MIR:00000511",
      new Class<?>[]{},
      "ega.study",
      "EGAS00000000001"),

  FMA("FMA",
      "https://www.ebi.ac.uk/ols/ontologies/fma/",
      new String[]{"urn:miriam:fma", "urn:miriam:obo.fma",
          "http://identifiers.org/fma/",
          "http://identifiers.org/FMA",
          "https://identifiers.org/fma/",
          "https://identifiers.org/FMA",
      }, new Class<?>[]{}, "MIR:00000067",
      new Class<?>[]{},
      "",
      "FMA:67112"),

  FOODB_COMPOUND("FooDB Compound",
      "http://foodb.ca/foods",
      new String[]{"urn:miriam:foodb.compound",
          "http://identifiers.org/foodb.compound/",
          "http://identifiers.org/foodb.compound",
          "https://identifiers.org/foodb.compound/",
          "https://identifiers.org/foodb.compound",
      }, new Class<?>[]{}, "MIR:00000530",
      new Class<?>[]{},
      "foodb.compound",
      "FDB002100"),

  F_SNP("F-SNP",
      "http://compbio.cs.queensu.ca/F-SNP/",
      new String[]{"urn:miriam:fsnp",
          "http://identifiers.org/fsnp/",
          "http://identifiers.org/fsnp",
          "https://identifiers.org/fsnp/",
          "https://identifiers.org/fsnp",
      }, new Class<?>[]{}, "MIR:00000496",
      new Class<?>[]{},
      "fsnp",
      "rs17852708"),

  FUNCBASE_FLY("FuncBase Fly",
      "http://func.mshri.on.ca/fly",
      new String[]{"urn:miriam:funcbase.fly",
          "http://identifiers.org/funcbase.fly/",
          "http://identifiers.org/funcbase.fly",
          "https://identifiers.org/funcbase.fly/",
          "https://identifiers.org/funcbase.fly",
      }, new Class<?>[]{}, "MIR:00000461",
      new Class<?>[]{},
      "funcbase.fly",
      "10194"),
  FUNCBASE_HUMAN("FuncBase Human",
      "http://func.mshri.on.ca/human/",
      new String[]{"urn:miriam:funcbase.human",
          "http://identifiers.org/funcbase.human/",
          "http://identifiers.org/funcbase.human",
          "https://identifiers.org/funcbase.human/",
          "https://identifiers.org/funcbase.human",
      }, new Class<?>[]{}, "MIR:00000462",
      new Class<?>[]{},
      "funcbase.human",
      "119514"),
  FUNCBASE_MOUSE("FuncBase Mouse",
      "http://func.mshri.on.ca/mouse/",
      new String[]{"urn:miriam:funcbase.mouse",
          "http://identifiers.org/funcbase.mouse/",
          "http://identifiers.org/funcbase.mouse",
          "https://identifiers.org/funcbase.mouse/",
          "https://identifiers.org/funcbase.mouse",
      }, new Class<?>[]{}, "MIR:00000463",
      new Class<?>[]{},
      "funcbase.mouse",
      "1351341"),
  FUNCBASE_YEAST("FuncBase Yeast",
      "http://func.mshri.on.ca/yeast",
      new String[]{"urn:miriam:funcbase.yeast",
          "http://identifiers.org/funcbase.yeast/",
          "http://identifiers.org/funcbase.yeast",
          "https://identifiers.org/funcbase.yeast/",
          "https://identifiers.org/funcbase.yeast",
      }, new Class<?>[]{}, "MIR:00000464",
      new Class<?>[]{},
      "funcbase.yeast",
      "2701"),

  FUNGIDB("FungiDB",
      "https://fungidb.org/fungidb",
      new String[]{"urn:miriam:fungidb",
          "http://identifiers.org/fungidb/",
          "http://identifiers.org/fungidb",
          "https://identifiers.org/fungidb/",
          "https://identifiers.org/fungidb",
      }, new Class<?>[]{}, "MIR:00000365",
      new Class<?>[]{},
      "fungidb",
      "CNBG_0001"),

  FLYBASE("FlyBase",
      "https://www.alliancegenome.org",
      new String[]{"urn:miriam:fb", "urn:miriam:flybase",
          "http://identifiers.org/fb/",
          "http://identifiers.org/fb",
          "https://identifiers.org/fb/",
          "https://identifiers.org/fb",
      },
      new Class<?>[]{}, "MIR:00000030",
      new Class<?>[]{},
      "fb",
      "FBgn0011293"),

  GABI("GABI",
      "http://www.gabipd.org/",
      new String[]{"urn:miriam:gabi",
          "http://identifiers.org/gabi/",
          "http://identifiers.org/gabi",
          "https://identifiers.org/gabi/",
          "https://identifiers.org/gabi",
      }, new Class<?>[]{}, "MIR:00000164",
      new Class<?>[]{},
      "gabi",
      "2679240"),
  GENATLAS("Genatlas",
      "http://genatlas.medecine.univ-paris5.fr/",
      new String[]{"urn:miriam:genatlas",
          "http://identifiers.org/genatlas/",
          "http://identifiers.org/genatlas",
          "https://identifiers.org/genatlas/",
          "https://identifiers.org/genatlas",
      }, new Class<?>[]{}, "MIR:00000208",
      new Class<?>[]{},
      "genatlas",
      "HBB"),
  GENECARDS("GeneCards",
      "http://www.genecards.org/",
      new String[]{"urn:miriam:genecards",
          "http://identifiers.org/genecards/",
          "http://identifiers.org/genecards",
          "https://identifiers.org/genecards/",
          "https://identifiers.org/genecards",
      }, new Class<?>[]{}, "MIR:00000323",
      new Class<?>[]{},
      "genecards",
      "ABL1"),

  GENE_DB("GeneDB",
      "https://www.genedb.org/",
      new String[]{"urn:miriam:genedb",
          "http://identifiers.org/genedb/",
          "http://identifiers.org/genedb",
          "https://identifiers.org/genedb/",
          "https://identifiers.org/genedb",
      },
      new Class<?>[]{}, "MIR:00000106",
      new Class<?>[]{},
      "genedb",
      "LinJ.20.0070"),

  GENEFARM("GeneFarm",
      "http://urgi.versailles.inra.fr/Genefarm/",
      new String[]{"urn:miriam:genefarm",
          "http://identifiers.org/genefarm/",
          "http://identifiers.org/genefarm",
          "https://identifiers.org/genefarm/",
          "https://identifiers.org/genefarm",
      }, new Class<?>[]{}, "MIR:00000211",
      new Class<?>[]{},
      "genefarm",
      "4892"),
  GENETREE("GeneTree",
      "http://www.ensembl.org/",
      new String[]{"urn:miriam:genetree",
          "http://identifiers.org/genetree/",
          "http://identifiers.org/genetree",
          "https://identifiers.org/genetree/",
          "https://identifiers.org/genetree",
      }, new Class<?>[]{}, "MIR:00000214",
      new Class<?>[]{},
      "genetree",
      "ENSGT00550000074763"),
  GENE_WIKI("Gene Wiki",
      "http://en.wikipedia.org/wiki/Gene_Wiki",
      new String[]{"urn:miriam:genewiki",
          "http://identifiers.org/genewiki/",
          "http://identifiers.org/genewiki",
          "https://identifiers.org/genewiki/",
          "https://identifiers.org/genewiki",
      }, new Class<?>[]{}, "MIR:00000487",
      new Class<?>[]{},
      "genewiki",
      "1017"),
  GENPEPT("GenPept",
      "https://www.ncbi.nlm.nih.gov/protein",
      new String[]{"urn:miriam:genpept",
          "http://identifiers.org/genpept/",
          "http://identifiers.org/genpept",
          "https://identifiers.org/genpept/",
          "https://identifiers.org/genpept",
      }, new Class<?>[]{}, "MIR:00000345",
      new Class<?>[]{},
      "genpept",
      "CAA71118.1"),

  GENOME_PROPERTIES("Genome Properties",
      "https://www.ebi.ac.uk/interpro/genomeproperties/",
      new String[]{"urn:miriam:genprop",
          "http://identifiers.org/genprop/",
          "http://identifiers.org/genprop",
          "https://identifiers.org/genprop/",
          "https://identifiers.org/genprop",
      }, new Class<?>[]{}, "MIR:00000443",
      new Class<?>[]{},
      "genprop",
      "GenProp0699"),

  @NoMiriamUrn
  GENOMIC_DATA_COMMONS_DATA_PORTAL("Genomic Data Commons Data Portal",
      "https://gdc.cancer.gov/",
      new String[]{
          "http://identifiers.org/gdc/",
          "http://identifiers.org/gdc",
          "https://identifiers.org/gdc/",
          "https://identifiers.org/gdc",
      }, new Class<?>[]{}, "MIR:00000604",
      new Class<?>[]{},
      "gdc",
      "ae8c77fe-e6c8-44d5-8265-4a38c637bbef"),

  GEO("GEO",
      "https://www.ncbi.nlm.nih.gov/geo/",
      new String[]{"urn:miriam:geo",
          "http://identifiers.org/geo/",
          "http://identifiers.org/geo",
          "https://identifiers.org/geo/",
          "https://identifiers.org/geo",
      }, new Class<?>[]{}, "MIR:00000054",
      new Class<?>[]{},
      "geo",
      "GDS1234"),

  GIARDIADB("GiardiaDB",
      "https://giardiadb.org/giardiadb/",
      new String[]{"urn:miriam:giardiadb",
          "http://identifiers.org/giardiadb/",
          "http://identifiers.org/giardiadb",
          "https://identifiers.org/giardiadb/",
          "https://identifiers.org/giardiadb",
      }, new Class<?>[]{}, "MIR:00000151",
      new Class<?>[]{},
      "giardiadb",
      "GL50803_102438"),
  GLIDA_GPCR("GLIDA GPCR",
      "http://pharminfo.pharm.kyoto-u.ac.jp/services/glida/",
      new String[]{"urn:miriam:glida.gpcr",
          "http://identifiers.org/glida.gpcr/",
          "http://identifiers.org/glida.gpcr",
          "https://identifiers.org/glida.gpcr/",
          "https://identifiers.org/glida.gpcr",
      }, new Class<?>[]{}, "MIR:00000493",
      new Class<?>[]{},
      "glida.gpcr",
      "ACM1_HUMAN"),
  GLIDA_LIGAND("GLIDA Ligand",
      "http://pharminfo.pharm.kyoto-u.ac.jp/services/glida/",
      new String[]{"urn:miriam:glida.ligand",
          "http://identifiers.org/glida.ligand/",
          "http://identifiers.org/glida.ligand",
          "https://identifiers.org/glida.ligand/",
          "https://identifiers.org/glida.ligand",
      }, new Class<?>[]{}, "MIR:00000494",
      new Class<?>[]{},
      "glida.ligand",
      "L000001"),
  GLYCOEPITOPE("GlycoEpitope",
      "https://www.glycoepitope.jp/epitopes/",
      new String[]{"urn:miriam:glycoepitope",
          "http://identifiers.org/glycoepitope/",
          "http://identifiers.org/glycoepitope",
          "https://identifiers.org/glycoepitope/",
          "https://identifiers.org/glycoepitope",
      }, new Class<?>[]{}, "MIR:00000478",
      new Class<?>[]{},
      "glycoepitope",
      "EP0311"),

  GLYCOMEDB("GlycomeDB",
      "https://glytoucan.org/",
      new String[]{"urn:miriam:glycomedb",
          "http://identifiers.org/glycomedb/",
          "http://identifiers.org/glycomedb",
          "https://identifiers.org/glycomedb/",
          "https://identifiers.org/glycomedb",
      }, new Class<?>[]{}, "MIR:00000114",
      new Class<?>[]{},
      "glycomedb",
      "G77500AY"),

  /**
   * Gene Ontology: http://amigo.geneontology.org/amigo.
   */
  GO("Gene Ontology",
      "http://amigo.geneontology.org/amigo",
      new String[]{"urn:miriam:obo.go", "urn:miriam:go",
          "http://identifiers.org/go/",
          "http://identifiers.org/obo.go/",
          "http://identifiers.org/GO",
          "https://identifiers.org/go/",
          "https://identifiers.org/obo.go/",
          "https://identifiers.org/GO",
      },
      new Class<?>[]{Phenotype.class, Compartment.class, Complex.class}, "MIR:00000022",
      new Class<?>[]{},
      "",
      "GO:0006915"),

  GO_REF("Gene Ontology Reference",
      "http://www.geneontology.org/cgi-bin/references.cgi",
      new String[]{"urn:miriam:go.ref",
          "http://identifiers.org/GO_REF/",
          "http://identifiers.org/GO_REF",
          "https://identifiers.org/GO_REF/",
          "https://identifiers.org/GO_REF",
      },
      new Class<?>[]{}, "MIR:00000450",
      new Class<?>[]{},
      "",
      "GO_REF:0000041"),

  GOA("GOA",
      "https://www.ebi.ac.uk/GOA/",
      new String[]{"urn:miriam:goa",
          "http://identifiers.org/goa/",
          "http://identifiers.org/goa",
          "https://identifiers.org/goa/",
          "https://identifiers.org/goa",
      }, new Class<?>[]{}, "MIR:00000196",
      new Class<?>[]{},
      "goa",
      "P12345"),

  GOLD_GENOME("GOLD genome",
      "http://www.genomesonline.org/cgi-bin/GOLD/index.cgi",
      new String[]{"urn:miriam:gold.genome",
          "http://identifiers.org/gold.genome/",
          "http://identifiers.org/gold.genome",
          "https://identifiers.org/gold.genome/",
          "https://identifiers.org/gold.genome",
      }, new Class<?>[]{}, "MIR:00000401",
      new Class<?>[]{},
      "gold.genome",
      "Gi07796"),
  GOLD_METADATA("GOLD metadata",
      "http://www.genomesonline.org/cgi-bin/GOLD/index.cgi",
      new String[]{"urn:miriam:gold.meta",
          "http://identifiers.org/gold.meta/",
          "http://identifiers.org/gold.meta",
          "https://identifiers.org/gold.meta/",
          "https://identifiers.org/gold.meta",
      }, new Class<?>[]{}, "MIR:00000402",
      new Class<?>[]{},
      "gold.meta",
      "Gm00047"),

  GOOGLE_PATENTS("Google Patents",
      "https://www.google.com/patents/",
      new String[]{"urn:miriam:google.patent",
          "http://identifiers.org/google.patent/",
          "http://identifiers.org/google.patent",
          "https://identifiers.org/google.patent/",
          "https://identifiers.org/google.patent",
      },
      new Class<?>[]{}, "MIR:00000537",
      new Class<?>[]{},
      "google.patent",
      "US4145692"),

  GOLM_METABOLOME_DATABASE("Golm Metabolome Database",
      "http://gmd.mpimp-golm.mpg.de/",
      new String[]{"urn:miriam:gmd",
          "http://identifiers.org/gmd/",
          "http://identifiers.org/gmd",
          "https://identifiers.org/gmd/",
          "https://identifiers.org/gmd",
      }, new Class<?>[]{}, "MIR:00000274",
      new Class<?>[]{},
      "gmd",
      "68513255-fc44-4041-bc4b-4fd2fae7541d"),
  GOLM_METABOLOME_DATABASE_ANALYTE("Golm Metabolome Database Analyte",
      "http://gmd.mpimp-golm.mpg.de/",
      new String[]{"urn:miriam:gmd.analyte",
          "http://identifiers.org/gmd.analyte/",
          "http://identifiers.org/gmd.analyte",
          "https://identifiers.org/gmd.analyte/",
          "https://identifiers.org/gmd.analyte",
      }, new Class<?>[]{}, "MIR:00000426",
      new Class<?>[]{},
      "gmd.analyte",
      "4f0fa9b6-514f-4ff4-98cc-0009bc08eb80"),
  GOLM_METABOLOME_DATABASE_GC_MS_SPECTRA("Golm Metabolome Database GC-MS spectra",
      "http://gmd.mpimp-golm.mpg.de/",
      new String[]{"urn:miriam:gmd.gcms",
          "http://identifiers.org/gmd.gcms/",
          "http://identifiers.org/gmd.gcms",
          "https://identifiers.org/gmd.gcms/",
          "https://identifiers.org/gmd.gcms",
      }, new Class<?>[]{}, "MIR:00000424",
      new Class<?>[]{},
      "gmd.gcms",
      "53d583d8-40c6-40e1-9296-23f821cd77a5"),
  GOLM_METABOLOME_DATABASE_PROFILE("Golm Metabolome Database Profile",
      "http://gmd.mpimp-golm.mpg.de/",
      new String[]{"urn:miriam:gmd.profile",
          "http://identifiers.org/gmd.profile/",
          "http://identifiers.org/gmd.profile",
          "https://identifiers.org/gmd.profile/",
          "https://identifiers.org/gmd.profile",
      }, new Class<?>[]{}, "MIR:00000423",
      new Class<?>[]{},
      "gmd.profile",
      "10b38aaf-b977-4950-85b8-f4775f66658d"),
  GOLM_METABOLOME_DATABASE_REFERENCE_SUBSTANCE("Golm Metabolome Database Reference Substance",
      "http://gmd.mpimp-golm.mpg.de/",
      new String[]{"urn:miriam:gmd.ref",
          "http://identifiers.org/gmd.ref/",
          "http://identifiers.org/gmd.ref",
          "https://identifiers.org/gmd.ref/",
          "https://identifiers.org/gmd.ref",
      }, new Class<?>[]{}, "MIR:00000425",
      new Class<?>[]{},
      "gmd.ref",
      "8cf84adb-b4db-4807-ac98-0004247c35df"),

  GPCRDB("GPCRDB",
      "http://www.gpcrdb.org/",
      new String[]{"urn:miriam:gpcrdb",
          "http://identifiers.org/gpcrdb/",
          "http://identifiers.org/gpcrdb",
          "https://identifiers.org/gpcrdb/",
          "https://identifiers.org/gpcrdb",
      }, new Class<?>[]{}, "MIR:00000212",
      new Class<?>[]{},
      "gpcrdb",
      "RL3R1_HUMAN"),
  GRAMENE_GENES("Gramene genes",
      "http://www.gramene.org/",
      new String[]{"urn:miriam:gramene.gene",
          "http://identifiers.org/gramene.gene/",
          "http://identifiers.org/gramene.gene",
          "https://identifiers.org/gramene.gene/",
          "https://identifiers.org/gramene.gene",
      }, new Class<?>[]{}, "MIR:00000182",
      new Class<?>[]{},
      "gramene.gene",
      "GR:0080039"),

  GRAMENE_PROTEIN("Gramene protein",
      "http://www.gramene.org/",
      new String[]{"urn:miriam:gramene.protein",
          "http://identifiers.org/gramene.protein/",
          "http://identifiers.org/gramene.protein",
          "https://identifiers.org/gramene.protein/",
          "https://identifiers.org/gramene.protein",
      }, new Class<?>[]{}, "MIR:00000181",
      new Class<?>[]{},
      "gramene.protein",
      "78073"),
  GRAMENE_QTL("Gramene QTL",
      "http://www.gramene.org/",
      new String[]{"urn:miriam:gramene.qtl",
          "http://identifiers.org/gramene.qtl/",
          "http://identifiers.org/gramene.qtl",
          "https://identifiers.org/gramene.qtl/",
          "https://identifiers.org/gramene.qtl",
      }, new Class<?>[]{}, "MIR:00000184",
      new Class<?>[]{},
      "gramene.qtl",
      "CQG5"),
  GRAMENE_TAXONOMY("Gramene Taxonomy",
      "http://www.gramene.org/",
      new String[]{"urn:miriam:gramene.taxonomy",
          "http://identifiers.org/gramene.taxonomy/",
          "http://identifiers.org/gramene.taxonomy",
          "https://identifiers.org/gramene.taxonomy/",
          "https://identifiers.org/gramene.taxonomy",
      }, new Class<?>[]{}, "MIR:00000183",
      new Class<?>[]{},
      "gramene.taxonomy",
      "GR_tax:013681"),
  GREENGENES("GreenGenes",
      "http://greengenes.lbl.gov/",
      new String[]{"urn:miriam:greengenes",
          "http://identifiers.org/greengenes/",
          "http://identifiers.org/greengenes",
          "https://identifiers.org/greengenes/",
          "https://identifiers.org/greengenes",
      }, new Class<?>[]{}, "MIR:00000165",
      new Class<?>[]{},
      "greengenes",
      "100000"),
  GRIN_PLANT_TAXONOMY("GRIN Plant Taxonomy",
      "http://www.ars-grin.gov/cgi-bin/npgs/html/index.pl?language=en",
      new String[]{"urn:miriam:grin.taxonomy",
          "http://identifiers.org/grin.taxonomy/",
          "http://identifiers.org/grin.taxonomy",
          "https://identifiers.org/grin.taxonomy/",
          "https://identifiers.org/grin.taxonomy",
      }, new Class<?>[]{}, "MIR:00000166",
      new Class<?>[]{},
      "grin.taxonomy",
      "19333"),
  GRSDB("GRSDB",
      "http://bioinformatics.ramapo.edu/GRSDB2/",
      new String[]{"urn:miriam:grsdb",
          "http://identifiers.org/grsdb/",
          "http://identifiers.org/grsdb",
          "https://identifiers.org/grsdb/",
          "https://identifiers.org/grsdb",
      }, new Class<?>[]{}, "MIR:00000495",
      new Class<?>[]{},
      "grsdb",
      "10142"),

  GWAS_CENRAL_MARKER("GWAS Central Marker",
      "https://www.gwascentral.org/markers/",
      new String[]{"urn:miriam:gwascentral.marker",
          "http://identifiers.org/gwascentral.marker/",
          "http://identifiers.org/gwascentral.marker",
          "https://identifiers.org/gwascentral.marker/",
          "https://identifiers.org/gwascentral.marker",
      },
      new Class<?>[]{}, "MIR:00000542",
      new Class<?>[]{},
      "gwascentral.marker",
      "HGVM15354"),

  GWAS_CENRAL_PHENOTYPE("GWAS Central Phenotype",
      "https://www.gwascentral.org/phenotypes",
      new String[]{"urn:miriam:gwascentral.phenotype",
          "http://identifiers.org/gwascentral.phenotype/",
          "http://identifiers.org/gwascentral.phenotype",
          "https://identifiers.org/gwascentral.phenotype/",
          "https://identifiers.org/gwascentral.phenotype",
      },
      new Class<?>[]{}, "MIR:00000543",
      new Class<?>[]{},
      "gwascentral.phenotype",
      "HGVPM623"),

  GWAS_CENRAL_STUDY("GWAS Central Study",
      "https://www.gwascentral.org/studies",
      new String[]{"urn:miriam:gwascentral.study",
          "http://identifiers.org/gwascentral.study/",
          "http://identifiers.org/gwascentral.study",
          "https://identifiers.org/gwascentral.study/",
          "https://identifiers.org/gwascentral.study",
      },
      new Class<?>[]{}, "MIR:00000540",
      new Class<?>[]{},
      "gwascentral.study",
      "HGVST1828"),

  GXA_EXPT("GXA Expt",
      "https://www.ebi.ac.uk/gxa/",
      new String[]{"urn:miriam:gxa.expt",
          "http://identifiers.org/gxa.expt/",
          "http://identifiers.org/gxa.expt",
          "https://identifiers.org/gxa.expt/",
          "https://identifiers.org/gxa.expt",
      }, new Class<?>[]{}, "MIR:00000379",
      new Class<?>[]{},
      "gxa.expt",
      "E-MTAB-2037"),
  GXA_GENE("GXA Gene",
      "https://www.ebi.ac.uk/gxa/",
      new String[]{"urn:miriam:gxa.gene",
          "http://identifiers.org/gxa.gene/",
          "http://identifiers.org/gxa.gene",
          "https://identifiers.org/gxa.gene/",
          "https://identifiers.org/gxa.gene",
      }, new Class<?>[]{}, "MIR:00000378",
      new Class<?>[]{},
      "gxa.gene",
      "AT4G01080"),
  HAMAP("HAMAP",
      "https://hamap.expasy.org/",
      new String[]{"urn:miriam:hamap",
          "http://identifiers.org/hamap/",
          "http://identifiers.org/hamap",
          "https://identifiers.org/hamap/",
          "https://identifiers.org/hamap",
      }, new Class<?>[]{}, "MIR:00000292",
      new Class<?>[]{},
      "hamap",
      "MF_01400"),

  HCVDB("HCVDB",
      "http://euhcvdb.ibcp.fr/euHCVdb/",
      new String[]{"urn:miriam:hcvdb",
          "http://identifiers.org/hcvdb/",
          "http://identifiers.org/hcvdb",
          "https://identifiers.org/hcvdb/",
          "https://identifiers.org/hcvdb",
      }, new Class<?>[]{}, "MIR:00000207",
      new Class<?>[]{},
      "hcvdb",
      "M58335"),

  HGMD("HGMD",
      "http://www.hgmd.cf.ac.uk/ac/index.php",
      new String[]{"urn:miriam:hgmd",
          "http://identifiers.org/hgmd/",
          "http://identifiers.org/hgmd",
          "https://identifiers.org/hgmd/",
          "https://identifiers.org/hgmd",
      }, new Class<?>[]{}, "MIR:00000392",
      new Class<?>[]{},
      "hgmd",
      "CALM1"),

  /**
   * HGNC: http://www.genenames.org.
   */
  HGNC("HGNC",
      "http://www.genenames.org",
      new String[]{"urn:miriam:hgnc",
          "http://identifiers.org/hgnc/",
          "http://identifiers.org/hgnc",
          "https://identifiers.org/hgnc/",
          "https://identifiers.org/hgnc",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000080",
      new Class<?>[]{Protein.class, Gene.class, Rna.class},
      "hgnc",
      "2674"),

  HGNC_FAMILY("HGNC Family",
      "https://www.genenames.org/",
      new String[]{"urn:miriam:hgnc.family",
          "http://identifiers.org/hgnc.family/",
          "http://identifiers.org/hgnc.family",
          "https://identifiers.org/hgnc.family/",
          "https://identifiers.org/hgnc.family",
      }, new Class<?>[]{}, "MIR:00000520",
      new Class<?>[]{},
      "hgnc.family",
      "PADI"),

  /**
   * HGNC symbol: http://www.genenames.org.
   */
  HGNC_SYMBOL("HGNC Symbol",
      "http://www.genenames.org",
      new String[]{"urn:miriam:hgnc.symbol",
          "http://identifiers.org/hgnc.symbol/",
          "http://identifiers.org/hgnc.symbol",
          "https://identifiers.org/hgnc.symbol/",
          "https://identifiers.org/hgnc.symbol",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000362",
      new Class<?>[]{Protein.class, Gene.class, Rna.class},
      "hgnc.symbol",
      "DAPK1"),

  H_INVDB_LOCUS("H-InvDb Locus",
      "http://h-invitational.jp/hinv/ahg-db/index.jsp",
      new String[]{"urn:miriam:hinv.locus",
          "http://identifiers.org/hinv.locus/",
          "http://identifiers.org/hinv.locus",
          "https://identifiers.org/hinv.locus/",
          "https://identifiers.org/hinv.locus",
      }, new Class<?>[]{}, "MIR:00000167",
      new Class<?>[]{},
      "hinv.locus",
      "HIX0004394"),
  H_INVDB_PROTEIN("H-InvDb Protein",
      "http://h-invitational.jp/hinv/ahg-db/index.jsp",
      new String[]{"urn:miriam:hinv.protein",
          "http://identifiers.org/hinv.protein/",
          "http://identifiers.org/hinv.protein",
          "https://identifiers.org/hinv.protein/",
          "https://identifiers.org/hinv.protein",
      }, new Class<?>[]{}, "MIR:00000169",
      new Class<?>[]{},
      "hinv.protein",
      "HIP000030660"),
  H_INVDB_TRANSCRIPT("H-InvDb Transcript",
      "http://h-invitational.jp/hinv/ahg-db/index.jsp",
      new String[]{"urn:miriam:hinv.transcript",
          "http://identifiers.org/hinv.transcript/",
          "http://identifiers.org/hinv.transcript",
          "https://identifiers.org/hinv.transcript/",
          "https://identifiers.org/hinv.transcript",
      }, new Class<?>[]{}, "MIR:00000168",
      new Class<?>[]{},
      "hinv.transcript",
      "HIT000195363"),

  /**
   * HMDB: http://www.hmdb.ca/.
   */
  HMDB("HMDB",
      "http://www.hmdb.ca/",
      new String[]{"urn:miriam:hmdb",
          "http://identifiers.org/hmdb/",
          "http://identifiers.org/hmdb",
          "https://identifiers.org/hmdb/",
          "https://identifiers.org/hmdb",
      },
      new Class<?>[]{Chemical.class, Drug.class,}, "MIR:00000051",
      new Class<?>[]{},
      "hmdb",
      "HMDB00001"),

  HOGENOM("HOGENOM",
      "http://pbil.univ-lyon1.fr/databases/hogenom/",
      new String[]{"urn:miriam:hogenom",
          "http://identifiers.org/hogenom/",
          "http://identifiers.org/hogenom",
          "https://identifiers.org/hogenom/",
          "https://identifiers.org/hogenom",
      }, new Class<?>[]{}, "MIR:00000213",
      new Class<?>[]{},
      "hogenom",
      "HBG284870"),
  HOMD_SEQUENCE_METAINFORMATION("HOMD Sequence Metainformation",
      "http://www.homd.org/index.php",
      new String[]{"urn:miriam:homd.seq",
          "http://identifiers.org/homd.seq/",
          "http://identifiers.org/homd.seq",
          "https://identifiers.org/homd.seq/",
          "https://identifiers.org/homd.seq",
      }, new Class<?>[]{}, "MIR:00000170",
      new Class<?>[]{},
      "homd.seq",
      "SEQF1003"),
  HOMD_TAXONOMY("HOMD Taxonomy",
      "http://www.homd.org/index.php",
      new String[]{"urn:miriam:homd.taxon",
          "http://identifiers.org/homd.taxon/",
          "http://identifiers.org/homd.taxon",
          "https://identifiers.org/homd.taxon/",
          "https://identifiers.org/homd.taxon",
      }, new Class<?>[]{}, "MIR:00000171",
      new Class<?>[]{},
      "homd.taxon",
      "811"),

  HOMEODOMAIN_RESEARCH("Homeodomain Research",
      "http://research.nhgri.nih.gov/apps/homeodomain/web/",
      new String[]{"urn:miriam:hdr",
          "http://identifiers.org/hdr/",
          "http://identifiers.org/hdr",
          "https://identifiers.org/hdr/",
          "https://identifiers.org/hdr",
      }, new Class<?>[]{}, "MIR:00000497",
      new Class<?>[]{},
      "hdr",
      "63"),

  HOMOLOGENE("HomoloGene",
      "https://www.ncbi.nlm.nih.gov/homologene/",
      new String[]{"urn:miriam:homologene",
          "http://identifiers.org/homologene/",
          "http://identifiers.org/homologene",
          "https://identifiers.org/homologene/",
          "https://identifiers.org/homologene",
      }, new Class<?>[]{}, "MIR:00000275",
      new Class<?>[]{},
      "homologene",
      "1002"),

  HOVERGEN("HOVERGEN",
      "http://pbil.univ-lyon1.fr/databases/hovergen.php",
      new String[]{"urn:miriam:hovergen",
          "http://identifiers.org/hovergen/",
          "http://identifiers.org/hovergen",
          "https://identifiers.org/hovergen/",
          "https://identifiers.org/hovergen",
      }, new Class<?>[]{}, "MIR:00000074",
      new Class<?>[]{},
      "hovergen",
      "HBG004341"),

  HPA("HPA",
      "http://www.proteinatlas.org/",
      new String[]{"urn:miriam:hpa",
          "http://identifiers.org/hpa/",
          "http://identifiers.org/hpa",
          "https://identifiers.org/hpa/",
          "https://identifiers.org/hpa",
      }, new Class<?>[]{}, "MIR:00000336",
      new Class<?>[]{},
      "hpa",
      "ENSG00000026508"),

  HPRD("HPRD",
      "http://www.hprd.org/",
      new String[]{"urn:miriam:hprd",
          "http://identifiers.org/hprd/",
          "http://identifiers.org/hprd",
          "https://identifiers.org/hprd/",
          "https://identifiers.org/hprd",
      }, new Class<?>[]{}, "MIR:00000377",
      new Class<?>[]{},
      "hprd",
      "00001"),

  HSSP("HSSP",
      "http://swift.cmbi.kun.nl/swift/hssp/",
      new String[]{"urn:miriam:hssp",
          "http://identifiers.org/hssp/",
          "http://identifiers.org/hssp",
          "https://identifiers.org/hssp/",
          "https://identifiers.org/hssp",
      }, new Class<?>[]{}, "MIR:00000215",
      new Class<?>[]{},
      "hssp",
      "102l"),

  HUGE("HUGE",
      "http://www.kazusa.or.jp/huge/",
      new String[]{"urn:miriam:huge",
          "http://identifiers.org/huge/",
          "http://identifiers.org/huge",
          "https://identifiers.org/huge/",
          "https://identifiers.org/huge",
      }, new Class<?>[]{}, "MIR:00000263",
      new Class<?>[]{},
      "huge",
      "KIAA0001"),

  HUMAN_DISEASE_ONTOLOGY("Human Disease Ontology",
      "http://bioportal.bioontology.org/ontologies/DOID",
      new String[]{"urn:miriam:obo.do", "urn:miriam:doid",
          "http://identifiers.org/doid/",
          "http://identifiers.org/DOID",
          "https://identifiers.org/doid/",
          "https://identifiers.org/DOID",
      }, new Class<?>[]{}, "MIR:00000233",
      new Class<?>[]{},
      "",
      "DOID:11337"),

  @NoMiriamUrn
  HUMAN_PHENOTYPE_ONTOLOGY("Human Phenotype Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/hp",
      new String[]{
          "http://identifiers.org/hp/",
          "http://identifiers.org/HP",
          "https://identifiers.org/hp/",
          "https://identifiers.org/HP",
      }, new Class<?>[]{}, "MIR:00000571",
      new Class<?>[]{},
      "",
      "HP:0000118"),

  HUMAN_PROTEOME_MAP_PEPTIDE("Human Proteome Map Peptide",
      "http://www.humanproteomemap.org/index.php",
      new String[]{"urn:miriam:hpm.peptide",
          "http://identifiers.org/hpm.peptide/",
          "http://identifiers.org/hpm.peptide",
          "https://identifiers.org/hpm.peptide/",
          "https://identifiers.org/hpm.peptide",
      }, new Class<?>[]{}, "MIR:00000527",
      new Class<?>[]{},
      "hpm.peptide",
      "9606117"),

  HUMAN_PROTEOME_MAP_PROTEIN("Human Proteome Map Protein",
      "http://www.humanproteomemap.org/index.php",
      new String[]{"urn:miriam:hpm.protein",
          "http://identifiers.org/hpm.protein/",
          "http://identifiers.org/hpm.protein",
          "https://identifiers.org/hpm.protein/",
          "https://identifiers.org/hpm.protein",
      }, new Class<?>[]{}, "MIR:00000526",
      new Class<?>[]{},
      "hpm.protein",
      "1968"),

  ICD("ICD",
      "http://www.who.int/classifications/icd/en/",
      new String[]{"urn:miriam:icd",
          "http://identifiers.org/icd/",
          "http://identifiers.org/icd",
          "https://identifiers.org/icd/",
          "https://identifiers.org/icd",
      }, new Class<?>[]{}, "MIR:00000009",
      new Class<?>[]{},
      "icd",
      "C34"),

  ICEBERG_ELEMENT("ICEberg element",
      "http://db-mml.sjtu.edu.cn/ICEberg/",
      new String[]{"urn:miriam:iceberg.element",
          "http://identifiers.org/iceberg.element/",
          "http://identifiers.org/iceberg.element",
          "https://identifiers.org/iceberg.element/",
          "https://identifiers.org/iceberg.element",
      }, new Class<?>[]{}, "MIR:00000469",
      new Class<?>[]{},
      "iceberg.element",
      "99"),

  ICEBERG_FAMILY("ICEberg family",
      "http://db-mml.sjtu.edu.cn/ICEberg/",
      new String[]{"urn:miriam:iceberg.family",
          "http://identifiers.org/iceberg.family/",
          "http://identifiers.org/iceberg.family",
          "https://identifiers.org/iceberg.family/",
          "https://identifiers.org/iceberg.family",
      }, new Class<?>[]{}, "MIR:00000470",
      new Class<?>[]{},
      "iceberg.family",
      "7"),

  IDEAL("IDEAL",
      "http://www.ideal.force.cs.is.nagoya-u.ac.jp/IDEAL/",
      new String[]{"urn:miriam:ideal",
          "http://identifiers.org/ideal/",
          "http://identifiers.org/ideal",
          "https://identifiers.org/ideal/",
          "https://identifiers.org/ideal",
      }, new Class<?>[]{}, "MIR:00000398",
      new Class<?>[]{},
      "ideal",
      "IID00001"),

  IDENTIFIERS_ORG_TERMS("Identifiers.org Terms",
      "https://identifiers.org/",
      new String[]{"urn:miriam:idot",
          "http://identifiers.org/idot/",
          "http://identifiers.org/idot",
          "https://identifiers.org/idot/",
          "https://identifiers.org/idot",
      }, new Class<?>[]{}, "MIR:00000519",
      new Class<?>[]{},
      "idot",
      "identifierPattern"),

  IMEX("IMEx",
      "https://www.imexconsortium.org/",
      new String[]{"urn:miriam:imex",
          "http://identifiers.org/imex/",
          "http://identifiers.org/imex",
          "https://identifiers.org/imex/",
          "https://identifiers.org/imex",
      }, new Class<?>[]{}, "MIR:00000122",
      new Class<?>[]{},
      "imex",
      "IM-19210-3"),

  IMGT_LIGM("IMGT LIGM",
      "http://genius.embnet.dkfz-heidelberg.de/",
      new String[]{"urn:miriam:imgt.ligm",
          "http://identifiers.org/imgt.ligm/",
          "http://identifiers.org/imgt.ligm",
          "https://identifiers.org/imgt.ligm/",
          "https://identifiers.org/imgt.ligm",
      }, new Class<?>[]{}, "MIR:00000287",
      new Class<?>[]{},
      "imgt.ligm",
      "M94112"),

  IMGT_HLA("IMGT HLA",
      "https://www.ebi.ac.uk/imgt/hla/allele.html",
      new String[]{"urn:miriam:imgt.hla",
          "http://identifiers.org/imgt.hla/",
          "http://identifiers.org/imgt.hla",
          "https://identifiers.org/imgt.hla/",
          "https://identifiers.org/imgt.hla",
      },
      new Class<?>[]{}, "MIR:00000331",
      new Class<?>[]{},
      "imgt.hla",
      "A*01:01:01:01"),
  INCHI("InChI",
      "http://www.chemspider.com/",
      new String[]{"urn:miriam:inchi",
          "http://identifiers.org/inchi/",
          "http://identifiers.org/inchi",
          "https://identifiers.org/inchi/",
          "https://identifiers.org/inchi",
      }, new Class<?>[]{}, "MIR:00000383",
      new Class<?>[]{},
      "inchi",
      "InChI=1S/C2H6O/c1-2-3/h3H,2H2,1H3"),

  INCHIKEY("InChIKey",
      "http://www.chemspider.com/",
      new String[]{"urn:miriam:inchikey",
          "http://identifiers.org/inchikey/",
          "http://identifiers.org/inchi_key/",
          "http://identifiers.org/inchikey",
          "https://identifiers.org/inchikey/",
          "https://identifiers.org/inchi_key/",
          "https://identifiers.org/inchikey",
      }, new Class<?>[]{}, "MIR:00000387",
      new Class<?>[]{},
      "inchikey",
      "RYYVLZVUVIJVGH-UHFFFAOYSA-N"),

  INTACT("IntAct",
      "https://www.ebi.ac.uk/intact/",
      new String[]{"urn:miriam:intact",
          "http://identifiers.org/intact/",
          "http://identifiers.org/intact",
          "https://identifiers.org/intact/",
          "https://identifiers.org/intact",
      },
      new Class<?>[]{}, "MIR:00000010",
      new Class<?>[]{},
      "intact",
      "EBI-2307691"),

  INTACT_MOLECULE("IntAct Molecule",
      "https://www.ebi.ac.uk/intact/",
      new String[]{"urn:miriam:intact.molecule",
          "http://identifiers.org/intact.molecule/",
          "http://identifiers.org/intact.molecule",
          "https://identifiers.org/intact.molecule/",
          "https://identifiers.org/intact.molecule",
      }, new Class<?>[]{}, "MIR:00000427",
      new Class<?>[]{},
      "intact.molecule",
      "EBI-366083"),

  INTEGRATED_MICROBIAL_GENOMES_GENE("Integrated Microbial Genomes Gene",
      "http://img.jgi.doe.gov/",
      new String[]{"urn:miriam:img.gene",
          "http://identifiers.org/img.gene/",
          "http://identifiers.org/img.gene",
          "https://identifiers.org/img.gene/",
          "https://identifiers.org/img.gene",
      }, new Class<?>[]{}, "MIR:00000176",
      new Class<?>[]{},
      "img.gene",
      "638309541"),
  INTEGRATED_MICROBIAL_GENOMES_TAXON("Integrated Microbial Genomes Taxon",
      "http://img.jgi.doe.gov/",
      new String[]{"urn:miriam:img.taxon",
          "http://identifiers.org/img.taxon/",
          "http://identifiers.org/img.taxon",
          "https://identifiers.org/img.taxon/",
          "https://identifiers.org/img.taxon",
      }, new Class<?>[]{}, "MIR:00000175",
      new Class<?>[]{},
      "img.taxon",
      "648028003"),

  /**
   * InterPro: http://www.ebi.ac.uk/interpro/.
   */
  INTERPRO("InterPro",
      "http://www.ebi.ac.uk/interpro/",
      new String[]{"urn:miriam:interpro",
          "http://identifiers.org/interpro/",
          "http://identifiers.org/interpro",
          "https://identifiers.org/interpro/",
          "https://identifiers.org/interpro",
      },
      new Class<?>[]{Protein.class, Complex.class}, "MIR:00000011",
      new Class<?>[]{},
      "interpro",
      "IPR000100"),

  IRD_SEGMENT_SEQUENCE("IRD Segment Sequence",
      "http://www.fludb.org/",
      new String[]{"urn:miriam:ird.segment",
          "http://identifiers.org/ird.segment/",
          "http://identifiers.org/ird.segment",
          "https://identifiers.org/ird.segment/",
          "https://identifiers.org/ird.segment",
      }, new Class<?>[]{}, "MIR:00000172",
      new Class<?>[]{},
      "ird.segment",
      "CY077097"),

  IREFWEB("iRefWeb",
      "http://wodaklab.org/iRefWeb/",
      new String[]{"urn:miriam:irefweb",
          "http://identifiers.org/irefweb/",
          "http://identifiers.org/irefweb",
          "https://identifiers.org/irefweb/",
          "https://identifiers.org/irefweb",
      }, new Class<?>[]{}, "MIR:00000123",
      new Class<?>[]{},
      "irefweb",
      "617102"),

  ISBN("ISBN",
      "http://isbndb.com/",
      new String[]{"urn:miriam:isbn",
          "http://identifiers.org/isbn/",
          "http://identifiers.org/isbn",
          "https://identifiers.org/isbn/",
          "https://identifiers.org/isbn",
      }, new Class<?>[]{}, "MIR:00000064",
      new Class<?>[]{},
      "isbn",
      "9781584885658"),

  ISFINDER("ISFinder",
      "http://www-is.biotoul.fr/i",
      new String[]{"urn:miriam:isfinder",
          "http://identifiers.org/isfinder/",
          "http://identifiers.org/isfinder",
          "https://identifiers.org/isfinder/",
          "https://identifiers.org/isfinder",
      }, new Class<?>[]{}, "MIR:00000173",
      new Class<?>[]{},
      "isfinder",
      "ISA1083-2"),
  ISSN("ISSN",
      "http://catalog.loc.gov/webvoy.htm",
      new String[]{"urn:miriam:issn",
          "http://identifiers.org/issn/",
          "http://identifiers.org/issn",
          "https://identifiers.org/issn/",
          "https://identifiers.org/issn",
      }, new Class<?>[]{}, "MIR:00000301",
      new Class<?>[]{},
      "issn",
      "0745-4570"),
  IUPHAR_FAMILY("IUPHAR family",
      "http://www.guidetopharmacology.org/",
      new String[]{"urn:miriam:iuphar.family",
          "http://identifiers.org/iuphar.family/",
          "http://identifiers.org/iuphar.family",
          "https://identifiers.org/iuphar.family/",
          "https://identifiers.org/iuphar.family",
      }, new Class<?>[]{}, "MIR:00000317",
      new Class<?>[]{},
      "iuphar.family",
      "78"),

  IUPHAR_LIGAND("IUPHAR ligand",
      "https://www.guidetopharmacology.org/GRAC/LigandListForward?database=all",
      new String[]{"urn:miriam:iuphar.ligand",
          "http://identifiers.org/iuphar.ligand/",
          "http://identifiers.org/iuphar.ligand",
          "https://identifiers.org/iuphar.ligand/",
          "https://identifiers.org/iuphar.ligand",
      },
      new Class<?>[]{}, "MIR:00000457",
      new Class<?>[]{},
      "iuphar.ligand",
      "1755"),

  IUPHAR_RECEPTOR("IUPHAR receptor",
      "http://www.guidetopharmacology.org/targets.jsp",
      new String[]{"urn:miriam:iuphar.receptor",
          "http://identifiers.org/iuphar.receptor/",
          "http://identifiers.org/iuphar.receptor",
          "https://identifiers.org/iuphar.receptor/",
          "https://identifiers.org/iuphar.receptor",
      }, new Class<?>[]{}, "MIR:00000281",
      new Class<?>[]{},
      "iuphar.receptor",
      "107"),

  JAPAN_COLLECTION_OF_MICROORGANISMS("Japan Collection of Microorganisms",
      "http://www.jcm.riken.go.jp/",
      new String[]{"urn:miriam:jcm",
          "http://identifiers.org/jcm/",
          "http://identifiers.org/jcm",
          "https://identifiers.org/jcm/",
          "https://identifiers.org/jcm",
      }, new Class<?>[]{}, "MIR:00000174",
      new Class<?>[]{},
      "jcm",
      "17254"),
  JAPAN_CHEMICAL_SUBSTANCE_DICTIONARY("Japan Chemical Substance Dictionary",
      "http://jglobal.jst.go.jp/en/",
      new String[]{"urn:miriam:jcsd",
          "http://identifiers.org/jcsd/",
          "http://identifiers.org/jcsd",
          "https://identifiers.org/jcsd/",
          "https://identifiers.org/jcsd",
      }, new Class<?>[]{}, "MIR:00000241",
      new Class<?>[]{},
      "jcsd",
      "J55.713G"),

  JAX_MICE("JAX Mice",
      "http://jaxmice.jax.org/",
      new String[]{"urn:miriam:jaxmice",
          "http://identifiers.org/jaxmice/",
          "http://identifiers.org/jaxmice",
          "https://identifiers.org/jaxmice/",
          "https://identifiers.org/jaxmice",
      }, new Class<?>[]{}, "MIR:00000337",
      new Class<?>[]{},
      "jaxmice",
      "005012"),

  JCGGDB("JCGGDB",
      "http://jcggdb.jp/index_en.html",
      new String[]{"urn:miriam:jcggdb",
          "http://identifiers.org/jcggdb/",
          "http://identifiers.org/jcggdb",
          "https://identifiers.org/jcggdb/",
          "https://identifiers.org/jcggdb",
      }, new Class<?>[]{}, "MIR:00000479",
      new Class<?>[]{},
      "jcggdb",
      "JCGG-STR008690"),

  JSTOR("JSTOR",
      "http://www.jstor.org/",
      new String[]{"urn:miriam:jstor",
          "http://identifiers.org/jstor/",
          "http://identifiers.org/jstor",
          "https://identifiers.org/jstor/",
          "https://identifiers.org/jstor",
      }, new Class<?>[]{}, "MIR:00000444",
      new Class<?>[]{},
      "jstor",
      "3075966"),

  JWS_ONLINE("JWS Online",
      "http://jjj.biochem.sun.ac.za/models/",
      new String[]{"urn:miriam:jws",
          "http://identifiers.org/jws/",
          "http://identifiers.org/jws",
          "https://identifiers.org/jws/",
          "https://identifiers.org/jws",
      }, new Class<?>[]{}, "MIR:00000130",
      new Class<?>[]{},
      "jws",
      "curien"),

  /**
   * KEGG Compound: http://www.genome.jp/kegg/ligand.html.
   */
  KEGG_COMPOUND("Kegg Compound",
      "http://www.genome.jp/kegg/ligand.html",
      new String[]{"urn:miriam:kegg.compound",
          "http://identifiers.org/kegg.compound/",
          "http://identifiers.org/kegg.compound",
          "https://identifiers.org/kegg.compound/",
          "https://identifiers.org/kegg.compound",
      },
      new Class<?>[]{Chemical.class}, "MIR:00000013",
      new Class<?>[]{},
      "kegg.compound",
      "C12345"),

  KEGG_DISEASE("KEGG Disease",
      "http://www.genome.jp/kegg/disease/",
      new String[]{"urn:miriam:kegg.disease",
          "http://identifiers.org/kegg.disease/",
          "http://identifiers.org/kegg.disease",
          "https://identifiers.org/kegg.disease/",
          "https://identifiers.org/kegg.disease",
      }, new Class<?>[]{}, "MIR:00000475",
      new Class<?>[]{},
      "kegg.disease",
      "H00076"),

  KEGG_DRUG("KEGG Drug",
      "https://www.genome.jp/kegg/drug/",
      new String[]{"urn:miriam:kegg.drug",
          "http://identifiers.org/kegg.drug/",
          "http://identifiers.org/kegg.drug",
          "https://identifiers.org/kegg.drug/",
          "https://identifiers.org/kegg.drug",
      },
      new Class<?>[]{}, "MIR:00000025",
      new Class<?>[]{},
      "kegg.drug",
      "D00123"),

  KEGG_ENVIRON("KEGG Environ",
      "http://www.genome.jp/kegg/drug/environ.html",
      new String[]{"urn:miriam:kegg.environ",
          "http://identifiers.org/kegg.environ/",
          "http://identifiers.org/kegg.environ",
          "https://identifiers.org/kegg.environ/",
          "https://identifiers.org/kegg.environ",
      }, new Class<?>[]{}, "MIR:00000389",
      new Class<?>[]{},
      "kegg.environ",
      "ev:E00032"),

  /**
   * KEGG Genes: http://www.genome.jp/kegg/genes.html.
   */
  KEGG_GENES("Kegg Genes",
      "http://www.genome.jp/kegg/genes.html",
      new String[]{"urn:miriam:kegg.genes",
          "http://identifiers.org/kegg.genes/",
          "http://identifiers.org/kegg.genes",
          "https://identifiers.org/kegg.genes/",
          "https://identifiers.org/kegg.genes",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000070",
      new Class<?>[]{},
      "kegg.genes",
      "syn:ssr3451"),

  KEGG_GENOME("KEGG Genome",
      "http://www.genome.jp/kegg/catalog/org_list.html",
      new String[]{"urn:miriam:kegg.genome",
          "http://identifiers.org/kegg.genome/",
          "http://identifiers.org/kegg.genome",
          "https://identifiers.org/kegg.genome/",
          "https://identifiers.org/kegg.genome",
      }, new Class<?>[]{}, "MIR:00000238",
      new Class<?>[]{},
      "kegg.genome",
      "eco"),

  KEGG_GLYCAN("Kegg Glycan",
      "https://www.genome.jp/kegg/glycan/",
      new String[]{"urn:miriam:kegg.glycan",
          "http://identifiers.org/kegg.glycan/",
          "http://identifiers.org/kegg.glycan",
          "https://identifiers.org/kegg.glycan/",
          "https://identifiers.org/kegg.glycan",
      },
      new Class<?>[]{}, "MIR:00000026",
      new Class<?>[]{},
      "kegg.glycan",
      "G00123"),

  KEGG_METAGENOME("KEGG Metagenome",
      "http://www.genome.jp/kegg/catalog/org_list3.html",
      new String[]{"urn:miriam:kegg.metagenome",
          "http://identifiers.org/kegg.metagenome/",
          "http://identifiers.org/kegg.metagenome",
          "https://identifiers.org/kegg.metagenome/",
          "https://identifiers.org/kegg.metagenome",
      }, new Class<?>[]{}, "MIR:00000239",
      new Class<?>[]{},
      "kegg.metagenome",
      "T30002"),

  KEGG_MODULE("KEGG Module",
      "http://www.kegg.jp/kegg/module.html",
      new String[]{"urn:miriam:kegg.module",
          "http://identifiers.org/kegg.module/",
          "http://identifiers.org/kegg.module",
          "https://identifiers.org/kegg.module/",
          "https://identifiers.org/kegg.module",
      }, new Class<?>[]{}, "MIR:00000474",
      new Class<?>[]{},
      "kegg.module",
      "M00002"),

  /**
   * KEGG Orthology: http://www.genome.jp/kegg/ko.html.
   */
  KEGG_ORTHOLOGY("KEGG Orthology",
      "http://www.genome.jp/kegg/ko.html",
      new String[]{"urn:miriam:kegg.orthology",
          "http://identifiers.org/kegg.orthology/",
          "http://identifiers.org/kegg.orthology",
          "https://identifiers.org/kegg.orthology/",
          "https://identifiers.org/kegg.orthology",
      },
      new Class<?>[]{}, "MIR:00000116",
      new Class<?>[]{},
      "kegg.orthology",
      "K00001"),

  /**
   * KEGG Pathway: http://www.genome.jp/kegg/pathway.html.
   */
  KEGG_PATHWAY("Kegg Pathway",
      "http://www.genome.jp/kegg/pathway.html",
      new String[]{"urn:miriam:kegg.pathway",
          "http://identifiers.org/kegg.pathway/",
          "http://identifiers.org/kegg.pathway",
          "https://identifiers.org/kegg.pathway/",
          "https://identifiers.org/kegg.pathway",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000012",
      new Class<?>[]{},
      "kegg.pathway",
      "hsa00620"),

  /**
   * KEGG Reaction: http://www.genome.jp/kegg/reaction/.
   */
  KEGG_REACTION("Kegg Reaction",
      "http://www.genome.jp/kegg/reaction/",
      new String[]{"urn:miriam:kegg.reaction",
          "http://identifiers.org/kegg.reaction/",
          "http://identifiers.org/kegg.reaction",
          "https://identifiers.org/kegg.reaction/",
          "https://identifiers.org/kegg.reaction",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000014",
      new Class<?>[]{},
      "kegg.reaction",
      "R00100"),

  KISAO("KiSAO",
      "http://bioportal.bioontology.org/ontologies/KISAO",
      new String[]{"urn:miriam:biomodels.kisao",
          "http://identifiers.org/biomodels.kisao/",
          "http://identifiers.org/biomodels.kisao",
          "https://identifiers.org/biomodels.kisao/",
          "https://identifiers.org/biomodels.kisao",
      }, new Class<?>[]{}, "MIR:00000108",
      new Class<?>[]{},
      "biomodels.kisao",
      "KISAO_0000057"),

  KNAPSACK("KnapSack",
      "http://kanaya.aist-nara.ac.jp/KNApSAcK/",
      new String[]{"urn:miriam:knapsack",
          "http://identifiers.org/knapsack/",
          "http://identifiers.org/knapsack",
          "https://identifiers.org/knapsack/",
          "https://identifiers.org/knapsack",
      },
      new Class<?>[]{}, "MIR:00000271",
      new Class<?>[]{},
      "knapsack",
      "C00000001"),

  LIGANDBOX("LigandBox",
      "http://www.mypresto5.com/ligandbox/cgi-bin/index.cgi?LANG=en",
      new String[]{"urn:miriam:ligandbox",
          "http://identifiers.org/ligandbox/",
          "http://identifiers.org/ligandbox",
          "https://identifiers.org/ligandbox/",
          "https://identifiers.org/ligandbox",
      }, new Class<?>[]{}, "MIR:00000477",
      new Class<?>[]{},
      "ligandbox",
      "D00001"),

  LIGAND_EXPO("Ligand Expo",
      "http://ligand-depot.rutgers.edu/index.html",
      new String[]{"urn:miriam:ligandexpo",
          "http://identifiers.org/ligandexpo/",
          "http://identifiers.org/ligandexpo",
          "https://identifiers.org/ligandexpo/",
          "https://identifiers.org/ligandexpo",
      }, new Class<?>[]{}, "MIR:00000062",
      new Class<?>[]{},
      "ligandexpo",
      "ABC"),

  LIGAND_GATED_ION_CHANNEL_DATABASE("Ligand-Gated Ion Channel database",
      "https://www.ebi.ac.uk/compneur-srv/LGICdb/LGICdb.php",
      new String[]{"urn:miriam:lgic",
          "http://identifiers.org/lgic/",
          "http://identifiers.org/lgic",
          "https://identifiers.org/lgic/",
          "https://identifiers.org/lgic",
      }, new Class<?>[]{}, "MIR:00000087",
      new Class<?>[]{},
      "lgic",
      "5HT3Arano"),

  LIPID_BANK("LipidBank",
      "http://lipidbank.jp/index.html",
      new String[]{"urn:miriam:lipidbank",
          "http://identifiers.org/lipidbank/",
          "http://identifiers.org/lipidbank",
          "https://identifiers.org/lipidbank/",
          "https://identifiers.org/lipidbank",
      },
      new Class<?>[]{}, "MIR:00000115",
      new Class<?>[]{},
      "lipidbank",
      "BBA0001"),

  LINCS_PROTEIN("LINCS Protein",
      "http://lincs.hms.harvard.edu/db/proteins/",
      new String[]{"urn:miriam:lincs.protein",
          "http://identifiers.org/lincs.protein/",
          "http://identifiers.org/lincs.protein",
          "https://identifiers.org/lincs.protein/",
          "https://identifiers.org/lincs.protein",
      },
      new Class<?>[]{}, "MIR:00000545",
      new Class<?>[]{},
      "lincs.protein",
      "200282"),

  LINCS_CELL("LINCS Cell",
      "http://lincsportal.ccs.miami.edu/cells/",
      new String[]{"urn:miriam:lincs.cell",
          "http://identifiers.org/lincs.cell/",
          "http://identifiers.org/lincs.cell",
          "https://identifiers.org/lincs.cell/",
          "https://identifiers.org/lincs.cell",
      },
      new Class<?>[]{}, "MIR:00000544",
      new Class<?>[]{},
      "lincs.cell",
      "LCL-2085"),

  LIPID_MAPS("LIPID MAPS",
      "http://www.lipidmaps.org",
      new String[]{"urn:miriam:lipidmaps",
          "http://identifiers.org/lipidmaps/",
          "http://identifiers.org/lipidmaps",
          "https://identifiers.org/lipidmaps/",
          "https://identifiers.org/lipidmaps",
      },
      new Class<?>[]{}, "MIR:00000052",
      new Class<?>[]{},
      "lipidmaps",
      "LMPR0102010012"),

  LOCUS_REFERENCE_GENOMIC("Locus Reference Genomic",
      "http://www.ensembl.org/",
      new String[]{"urn:miriam:lrg",
          "http://identifiers.org/lrg/",
          "http://identifiers.org/lrg",
          "https://identifiers.org/lrg/",
          "https://identifiers.org/lrg",
      }, new Class<?>[]{}, "MIR:00000376",
      new Class<?>[]{},
      "lrg",
      "LRG_1"),

  MACIE("MACiE",
      "https://www.ebi.ac.uk/thornton-srv/databases/MACiE/index.html",
      new String[]{"urn:miriam:macie",
          "http://identifiers.org/macie/",
          "http://identifiers.org/macie",
          "https://identifiers.org/macie/",
          "https://identifiers.org/macie",
      }, new Class<?>[]{}, "MIR:00000077",
      new Class<?>[]{},
      "macie",
      "M0001"),

  MAIZEGDB_LOCUS("MaizeGDB Locus",
      "http://www.maizegdb.org/",
      new String[]{"urn:miriam:maizegdb.locus",
          "http://identifiers.org/maizegdb.locus/",
          "http://identifiers.org/maizegdb.locus",
          "https://identifiers.org/maizegdb.locus/",
          "https://identifiers.org/maizegdb.locus",
      }, new Class<?>[]{}, "MIR:00000177",
      new Class<?>[]{},
      "maizegdb.locus",
      "25011"),

  MASSBANK("MassBank",
      "http://www.massbank.jp",
      new String[]{"urn:miriam:massbank",
          "http://identifiers.org/massbank/",
          "http://identifiers.org/massbank",
          "https://identifiers.org/massbank/",
          "https://identifiers.org/massbank",
      }, new Class<?>[]{}, "MIR:00000273",
      new Class<?>[]{},
      "massbank",
      "MSBNK-IPB_Halle-PB000166"),

  MATHEMATICAL_MODELLING_ONTOLOGY("Mathematical Modelling Ontology",
      "http://bioportal.bioontology.org/ontologies/MAMO",
      new String[]{"urn:miriam:mamo",
          "http://identifiers.org/mamo/",
          "http://identifiers.org/mamo",
          "https://identifiers.org/mamo/",
          "https://identifiers.org/mamo",
      }, new Class<?>[]{}, "MIR:00000517",
      new Class<?>[]{},
      "mamo",
      "MAMO_0000026"),

  MATRIXDB("MatrixDB",
      "http://matrixdb.univ-lyon1.fr/",
      new String[]{"urn:miriam:matrixdb.association",
          "http://identifiers.org/matrixdb.association/",
          "http://identifiers.org/matrixdb.association",
          "https://identifiers.org/matrixdb.association/",
          "https://identifiers.org/matrixdb.association",
      }, new Class<?>[]{}, "MIR:00000068",
      new Class<?>[]{},
      "matrixdb.association",
      "P00747__P07355"),
  MEDLINEPLUS("MedlinePlus",
      "http://www.nlm.nih.gov/medlineplus/",
      new String[]{"urn:miriam:medlineplus",
          "http://identifiers.org/medlineplus/",
          "http://identifiers.org/medlineplus",
          "https://identifiers.org/medlineplus/",
          "https://identifiers.org/medlineplus",
      }, new Class<?>[]{}, "MIR:00000476",
      new Class<?>[]{},
      "medlineplus",
      "002804"),
  MEROPS_FAMILY("MEROPS Family",
      "http://merops.sanger.ac.uk/index.htm",
      new String[]{"urn:miriam:merops.family",
          "http://identifiers.org/merops.family/",
          "http://identifiers.org/merops.family",
          "https://identifiers.org/merops.family/",
          "https://identifiers.org/merops.family",
      }, new Class<?>[]{}, "MIR:00000302",
      new Class<?>[]{},
      "merops.family",
      "S1"),
  MEROPS_INHIBITOR("MEROPS Inhibitor",
      "http://merops.sanger.ac.uk/index.htm",
      new String[]{"urn:miriam:merops.inhibitor",
          "http://identifiers.org/merops.inhibitor/",
          "http://identifiers.org/merops.inhibitor",
          "https://identifiers.org/merops.inhibitor/",
          "https://identifiers.org/merops.inhibitor",
      }, new Class<?>[]{}, "MIR:00000491",
      new Class<?>[]{},
      "merops.inhibitor",
      "I31.952"),

  MEROPS("MEROPS",
      "http://merops.sanger.ac.uk/index.htm",
      new String[]{"urn:miriam:merops",
          "http://identifiers.org/merops/",
          "http://identifiers.org/merops",
          "https://identifiers.org/merops/",
          "https://identifiers.org/merops",
      }, new Class<?>[]{}, "MIR:00000059",
      new Class<?>[]{},
      "merops",
      "S01.001"),

  /**
   * MeSH 2012: http://www.nlm.nih.gov/mesh/.
   */
  MESH_2012("MeSH",
      "http://www.nlm.nih.gov/mesh/",
      new String[]{"urn:miriam:mesh", "urn:miriam:mesh.2012", "urn:miriam:mesh.2013",
          "http://identifiers.org/mesh/",
          "http://identifiers.org/mesh",
          "https://identifiers.org/mesh/",
          "https://identifiers.org/mesh",
      },
      new Class<?>[]{Phenotype.class, Compartment.class, Complex.class}, "MIR:00000560",
      new Class<?>[]{},
      "mesh",
      "D010300"),

  @NoMiriamUrn
  METANETX_CHEMICAL("MetaNetX chemical",
      "https://www.metanetx.org/",
      new String[]{
          "http://identifiers.org/metanetx.chemical/",
          "http://identifiers.org/metanetx.chemical",
          "https://identifiers.org/metanetx.chemical/",
          "https://identifiers.org/metanetx.chemical",
      },
      new Class<?>[]{}, "MIR:00000567",
      new Class<?>[]{},
      "metanetx.chemical",
      "MNXM1723"),

  @NoMiriamUrn
  METANETX_COMPARTMENT("MetaNetX compartment",
      "https://www.metanetx.org/",
      new String[]{
          "http://identifiers.org/metanetx.compartment/",
          "http://identifiers.org/metanetx.compartment",
          "https://identifiers.org/metanetx.compartment/",
          "https://identifiers.org/metanetx.compartment",
      },
      new Class<?>[]{}, "MIR:00000569",
      new Class<?>[]{},
      "metanetx.compartment",
      "MNXC15"),

  @NoMiriamUrn
  METANETX_REACTION("MetaNetX reaction",
      "https://www.metanetx.org/",
      new String[]{
          "http://identifiers.org/metanetx.reaction/",
          "http://identifiers.org/metanetx.reaction",
          "https://identifiers.org/metanetx.reaction/",
          "https://identifiers.org/metanetx.reaction",
      },
      new Class<?>[]{}, "MIR:00000568",
      new Class<?>[]{},
      "metanetx.reaction",
      "MNXR101574"),

  METABOLIGHTS("MetaboLights",
      "https://www.ebi.ac.uk/metabolights/",
      new String[]{"urn:miriam:metabolights",
          "http://identifiers.org/metabolights/",
          "http://identifiers.org/metabolights",
          "https://identifiers.org/metabolights/",
          "https://identifiers.org/metabolights",
      }, new Class<?>[]{}, "MIR:00000380",
      new Class<?>[]{},
      "metabolights",
      "MTBLS1"),
  METLIN("METLIN",
      "http://masspec.scripps.edu/",
      new String[]{"urn:miriam:metlin",
          "http://identifiers.org/metlin/",
          "http://identifiers.org/metlin",
          "https://identifiers.org/metlin/",
          "https://identifiers.org/metlin",
      }, new Class<?>[]{}, "MIR:00000322",
      new Class<?>[]{},
      "metlin",
      "1455"),

  MI("Molecular Interactions Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/mi",
      new String[]{"urn:miriam:psimi", "urn:miriam:obo.mi",
          "http://identifiers.org/MI/",
          "http://identifiers.org/MI",
          "https://identifiers.org/MI/",
          "https://identifiers.org/MI",
      }, new Class<?>[]{}, "MIR:00000109",
      new Class<?>[]{},
      "",
      "MI:0308"),

  MICROSPORIDIADB("MicrosporidiaDB",
      "http://microsporidiadb.org/micro/",
      new String[]{"urn:miriam:microsporidia",
          "http://identifiers.org/microsporidia/",
          "http://identifiers.org/microsporidia",
          "https://identifiers.org/microsporidia/",
          "https://identifiers.org/microsporidia",
      }, new Class<?>[]{}, "MIR:00000152",
      new Class<?>[]{},
      "microsporidia",
      "ECU03_0820i"),

  MICROBIAL_PROTEIN_INTERACTION_DATABASE("Microbial Protein Interaction Database",
      "http://www.jcvi.org/mpidb/about.php",
      new String[]{"urn:miriam:mpid",
          "http://identifiers.org/mpid/",
          "http://identifiers.org/mpid",
          "https://identifiers.org/mpid/",
          "https://identifiers.org/mpid",
      }, new Class<?>[]{}, "MIR:00000124",
      new Class<?>[]{},
      "mpid",
      "172"),

  MIMODB("MimoDB",
      "http://immunet.cn/bdb/",
      new String[]{"urn:miriam:mimodb",
          "http://identifiers.org/mimodb/",
          "http://identifiers.org/mimodb",
          "https://identifiers.org/mimodb/",
          "https://identifiers.org/mimodb",
      }, new Class<?>[]{}, "MIR:00000251",
      new Class<?>[]{},
      "mimodb",
      "2"),

  MIPMODDB("MIPModDB",
      "http://bioinfo.iitk.ac.in/MIPModDB",
      new String[]{"urn:miriam:mipmod",
          "http://identifiers.org/mipmod/",
          "http://identifiers.org/mipmod",
          "https://identifiers.org/mipmod/",
          "https://identifiers.org/mipmod",
      }, new Class<?>[]{}, "MIR:00000256",
      new Class<?>[]{},
      "mipmod",
      "HOSAPI0399"),

  /**
   * miRBase Sequence: http://www.mirbase.org/.
   */
  MI_R_BASE_SEQUENCE("miRBase Sequence Database",
      "http://www.mirbase.org/",
      new String[]{"urn:miriam:mirbase",
          "http://identifiers.org/mirbase/",
          "http://identifiers.org/mirbase",
          "https://identifiers.org/mirbase/",
          "https://identifiers.org/mirbase",
      },
      new Class<?>[]{}, "MIR:00000078",
      new Class<?>[]{},
      "mirbase",
      "MI0000001"),

  /**
   * miRBase Mature Sequence: http://www.mirbase.org/.
   */
  MI_R_BASE_MATURE_SEQUENCE("miRBase Mature Sequence Database",
      "http://www.mirbase.org/",
      new String[]{"urn:miriam:mirbase.mature",
          "http://identifiers.org/mirbase.mature/",
          "http://identifiers.org/mirbase.mature",
          "https://identifiers.org/mirbase.mature/",
          "https://identifiers.org/mirbase.mature",
      },
      new Class<?>[]{}, "MIR:00000235",
      new Class<?>[]{},
      "mirbase.mature",
      "MIMAT0000001"),

  MIREX("mirEX",
      "http://comgen.pl/mirex/?page=home",
      new String[]{"urn:miriam:mirex",
          "http://identifiers.org/mirex/",
          "http://identifiers.org/mirex",
          "https://identifiers.org/mirex/",
          "https://identifiers.org/mirex",
      }, new Class<?>[]{}, "MIR:00000329",
      new Class<?>[]{},
      "mirex",
      "165a"),

  MIRIAM_REGISTRY_COLLECTION("MIRIAM Registry collection",
      "https://www.ebi.ac.uk/miriam/",
      new String[]{"urn:miriam:miriam.collection", "urn:miriam:miriam", "urn:miriam:miriam.datatype",
          "http://identifiers.org/miriam.collection/",
          "http://identifiers.org/miriam.collection",
          "https://identifiers.org/miriam.collection/",
          "https://identifiers.org/miriam.collection",
      }, new Class<?>[]{}, "MIR:00000008",
      new Class<?>[]{},
      "miriam.collection",
      "MIR:00000008"),

  MIRIAM_REGISTRY_RESOURCE("MIRIAM Registry resource",
      "https://www.ebi.ac.uk/miriam/",
      new String[]{"urn:miriam:miriam.resource",
          "http://identifiers.org/miriam.resource/",
          "http://identifiers.org/miriam.resource",
          "https://identifiers.org/miriam.resource/",
          "https://identifiers.org/miriam.resource",
      }, new Class<?>[]{}, "MIR:00000188",
      new Class<?>[]{},
      "miriam.resource",
      "MIR:00100005"),

  MIRNEST("miRNEST",
      "http://rhesus.amu.edu.pl/mirnest/copy/",
      new String[]{"urn:miriam:mirnest",
          "http://identifiers.org/mirnest/",
          "http://identifiers.org/mirnest",
          "https://identifiers.org/mirnest/",
          "https://identifiers.org/mirnest",
      }, new Class<?>[]{}, "MIR:00000246",
      new Class<?>[]{},
      "mirnest",
      "MNEST029358"),

  /**
   * miRTaRBase Mature Sequence: http://mirtarbase.mbc.nctu.edu.tw/.
   */
  MIR_TAR_BASE_MATURE_SEQUENCE("miRTarBase Mature Sequence Database",
      "http://mirtarbase.mbc.nctu.edu.tw/",
      new String[]{"urn:miriam:mirtarbase",
          "http://identifiers.org/mirtarbase/",
          "http://identifiers.org/mirtarbase",
          "https://identifiers.org/mirtarbase/",
          "https://identifiers.org/mirtarbase",
      },
      new Class<?>[]{}, "MIR:00000562",
      new Class<?>[]{},
      "mirtarbase",
      "MIRT000002"),

  /**
   * Mouse Genome Database: http://www.informatics.jax.org/.
   */
  MGD("Mouse Genome Database",
      "http://www.informatics.jax.org/",
      new String[]{"urn:miriam:mgd",
          "http://identifiers.org/mgi/",
          "http://identifiers.org/MGI",
          "https://identifiers.org/mgi/",
          "https://identifiers.org/MGI",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000037",
      new Class<?>[]{},
      "",
      "MGI:2442292"),

  MGED_ONTOLOGY("MGED Ontology",
      "http://bioportal.bioontology.org/",
      new String[]{"urn:miriam:mo",
          "http://identifiers.org/mo/",
          "http://identifiers.org/mo",
          "https://identifiers.org/mo/",
          "https://identifiers.org/mo",
      }, new Class<?>[]{}, "MIR:00000303",
      new Class<?>[]{},
      "mo",
      "ArrayGroup"),

  MGNIFY_PROJECT("MGnify Project",
      "https://www.ebi.ac.uk/metagenomics",
      new String[]{"urn:miriam:ebimetagenomics.proj",
          "http://identifiers.org/mgnify.proj/",
          "http://identifiers.org/mgnify.proj",
          "https://identifiers.org/mgnify.proj/",
          "https://identifiers.org/mgnify.proj",
      }, new Class<?>[]{}, "MIR:00000535",
      new Class<?>[]{},
      "mgnify.proj",
      "ERP004492"),

  MGNIFY_SAMPLE("MGnify Sample",
      "https://www.ebi.ac.uk/metagenomics",
      new String[]{"urn:miriam:ebimetagenomics.samp", "urn:miriam:ebimetagenomics",
          "http://identifiers.org/mgnify.samp/",
          "http://identifiers.org/mgnify.samp",
          "https://identifiers.org/mgnify.samp/",
          "https://identifiers.org/mgnify.samp",
      }, new Class<?>[]{}, "MIR:00000510",
      new Class<?>[]{},
      "mgnify.samp",
      "SRS086444"),

  MINT("MINT",
      "http://mint.bio.uniroma2.it/mint/",
      new String[]{"urn:miriam:mint",
          "http://identifiers.org/mint/",
          "http://identifiers.org/mint",
          "https://identifiers.org/mint/",
          "https://identifiers.org/mint",
      }, new Class<?>[]{}, "MIR:00000042",
      new Class<?>[]{},
      "mint",
      "MINT-10000"),

  MMRRC("MMRRC",
      "http://www.mmrrc.org/",
      new String[]{"urn:miriam:mmrrc",
          "http://identifiers.org/mmrrc/",
          "http://identifiers.org/mmrrc",
          "https://identifiers.org/mmrrc/",
          "https://identifiers.org/mmrrc",
      }, new Class<?>[]{}, "MIR:00000324",
      new Class<?>[]{},
      "mmrrc",
      "71"),

  MOD("Protein Modification Ontology",
      "https://bioportal.bioontology.org/ontologies/PSIMOD",
      new String[]{"urn:miriam:obo.psi-mod", "urn:miriam:psimod", "urn:miriam:mod",
          "http://identifiers.org/mod/",
          "http://identifiers.org/psimod/",
          "http://identifiers.org/obo.psi-mod/",
          "http://identifiers.org/MOD",
          "https://identifiers.org/mod/",
          "https://identifiers.org/obo.psi-mod/",
          "https://identifiers.org/psimod/",
          "https://identifiers.org/MOD",
      },
      new Class<?>[]{}, "MIR:00000056",
      new Class<?>[]{},
      "",
      "MOD:00001"),

  MODELDB("ModelDB",
      "http://senselab.med.yale.edu/ModelDB/",
      new String[]{"urn:miriam:modeldb",
          "http://identifiers.org/modeldb/",
          "http://identifiers.org/modeldb",
          "https://identifiers.org/modeldb/",
          "https://identifiers.org/modeldb",
      }, new Class<?>[]{}, "MIR:00000131",
      new Class<?>[]{},
      "modeldb",
      "45539"),

  MOLBASE("Molbase",
      "http://www.molbase.com/",
      new String[]{"urn:miriam:molbase",
          "http://identifiers.org/molbase/",
          "http://identifiers.org/molbase",
          "https://identifiers.org/molbase/",
          "https://identifiers.org/molbase",
      }, new Class<?>[]{}, "MIR:00000458",
      new Class<?>[]{},
      "molbase",
      "128796-39-4"),

  MOLECULAR_MODELING_DATABASE("Molecular Modeling Database",
      "http://www.ncbi.nlm.nih.gov/sites/entrez?db=structure",
      new String[]{"urn:miriam:mmdb",
          "http://identifiers.org/mmdb/",
          "http://identifiers.org/mmdb",
          "https://identifiers.org/mmdb/",
          "https://identifiers.org/mmdb",
      }, new Class<?>[]{}, "MIR:00000121",
      new Class<?>[]{},
      "mmdb",
      "50885"),

  MOUSE_ADULT_GROSS_ANATOMY("Mouse Adult Gross Anatomy",
      "http://bioportal.bioontology.org/ontologies/MA",
      new String[]{"urn:miriam:ma",
          "http://identifiers.org/ma/",
          "http://identifiers.org/MA",
          "https://identifiers.org/ma/",
          "https://identifiers.org/MA",
      }, new Class<?>[]{}, "MIR:00000445",
      new Class<?>[]{},
      "",
      "MA:0002502"),

  MYCOBANK("MycoBank",
      "http://www.mycobank.org/",
      new String[]{"urn:miriam:mycobank",
          "http://identifiers.org/mycobank/",
          "http://identifiers.org/mycobank",
          "https://identifiers.org/mycobank/",
          "https://identifiers.org/mycobank",
      }, new Class<?>[]{}, "MIR:00000178",
      new Class<?>[]{},
      "mycobank",
      "349124"),
  MYCOBROWSER_LEPRAE("MycoBrowser leprae",
      "http://mycobrowser.epfl.ch/leprosy.html",
      new String[]{"urn:miriam:myco.lepra",
          "http://identifiers.org/myco.lepra/",
          "http://identifiers.org/myco.lepra",
          "https://identifiers.org/myco.lepra/",
          "https://identifiers.org/myco.lepra",
      }, new Class<?>[]{}, "MIR:00000217",
      new Class<?>[]{},
      "myco.lepra",
      "ML0224"),
  MYCOBROWSER_MARINUM("MycoBrowser marinum",
      "http://mycobrowser.epfl.ch/marinolist.html",
      new String[]{"urn:miriam:myco.marinum",
          "http://identifiers.org/myco.marinum/",
          "http://identifiers.org/myco.marinum",
          "https://identifiers.org/myco.marinum/",
          "https://identifiers.org/myco.marinum",
      }, new Class<?>[]{}, "MIR:00000218",
      new Class<?>[]{},
      "myco.marinum",
      "MMAR_2462"),
  MYCOBROWSER_SMEGMATIS("MycoBrowser smegmatis",
      "http://mycobrowser.epfl.ch/smegmalist.html",
      new String[]{"urn:miriam:myco.smeg",
          "http://identifiers.org/myco.smeg/",
          "http://identifiers.org/myco.smeg",
          "https://identifiers.org/myco.smeg/",
          "https://identifiers.org/myco.smeg",
      }, new Class<?>[]{}, "MIR:00000219",
      new Class<?>[]{},
      "myco.smeg",
      "MSMEG_3769"),
  MYCOBROWSER_TUBERCULOSIS("MycoBrowser tuberculosis",
      "http://tuberculist.epfl.ch/",
      new String[]{"urn:miriam:myco.tuber",
          "http://identifiers.org/myco.tuber/",
          "http://identifiers.org/myco.tuber",
          "https://identifiers.org/myco.tuber/",
          "https://identifiers.org/myco.tuber",
      }, new Class<?>[]{}, "MIR:00000216",
      new Class<?>[]{},
      "myco.tuber",
      "Rv1908c"),

  NAPP("NAPP",
      "http://napp.u-psud.fr/",
      new String[]{"urn:miriam:napp",
          "http://identifiers.org/napp/",
          "http://identifiers.org/napp",
          "https://identifiers.org/napp/",
          "https://identifiers.org/napp",
      }, new Class<?>[]{}, "MIR:00000247",
      new Class<?>[]{},
      "napp",
      "351"),
  NARCIS("NARCIS",
      "http://www.narcis.nl/?Language=en",
      new String[]{"urn:miriam:narcis",
          "http://identifiers.org/narcis/",
          "http://identifiers.org/narcis",
          "https://identifiers.org/narcis/",
          "https://identifiers.org/narcis",
      }, new Class<?>[]{}, "MIR:00000240",
      new Class<?>[]{},
      "narcis",
      "oai:cwi.nl:4725"),
  NASC_CODE("NASC code",
      "http://arabidopsis.info/",
      new String[]{"urn:miriam:nasc",
          "http://identifiers.org/nasc/",
          "http://identifiers.org/nasc",
          "https://identifiers.org/nasc/",
          "https://identifiers.org/nasc",
      }, new Class<?>[]{}, "MIR:00000304",
      new Class<?>[]{},
      "nasc",
      "N1899"),

  NATIONAL_BIBLIOGRAPHY_NUMBER("National Bibliography Number",
      "http://nbn-resolving.org/resolve_urn.htm",
      new String[]{"urn:miriam:nbn",
          "http://identifiers.org/nbn/",
          "http://identifiers.org/nbn",
          "https://identifiers.org/nbn/",
          "https://identifiers.org/nbn",
      }, new Class<?>[]{}, "MIR:00000381",
      new Class<?>[]{},
      "nbn",
      "urn:nbn:fi:tkk-004781"),

  NATIONAL_DRUG_CODE("National Drug Code",
      "http://www.accessdata.fda.gov/scripts/cder/ndc/",
      new String[]{"urn:miriam:ndc",
          "http://identifiers.org/ndc/",
          "http://identifiers.org/ndc",
          "https://identifiers.org/ndc/",
          "https://identifiers.org/ndc",
      }, new Class<?>[]{}, "MIR:00000431",
      new Class<?>[]{},
      "ndc",
      "0002-1975-61"),

  NCBI_PROTEIN("NCBI Protein",
      "https://www.ncbi.nlm.nih.gov/protein",
      new String[]{"urn:miriam:ncbiprotein",
          "http://identifiers.org/ncbiprotein/",
          "http://identifiers.org/ncbiprotein",
          "https://identifiers.org/ncbiprotein/",
          "https://identifiers.org/ncbiprotein",
      },
      new Class<?>[]{}, "MIR:00000344",
      new Class<?>[]{},
      "ncbiprotein",
      "CAA71117.1"),

  NCIM("NCIm",
      "http://ncim.nci.nih.gov/",
      new String[]{"urn:miriam:ncim",
          "http://identifiers.org/ncim/",
          "http://identifiers.org/ncim",
          "https://identifiers.org/ncim/",
          "https://identifiers.org/ncim",
      }, new Class<?>[]{}, "MIR:00000353",
      new Class<?>[]{},
      "ncim",
      "C0026339"),

  NCI_PATHWAY_INTERACTION_DATABASE_PATHWAY("NCI Pathway Interaction Database: Pathway",
      "http://pid.nci.nih.gov/",
      new String[]{"urn:miriam:pid.pathway",
          "http://identifiers.org/pid.pathway/",
          "http://identifiers.org/pid.pathway",
          "https://identifiers.org/pid.pathway/",
          "https://identifiers.org/pid.pathway",
      }, new Class<?>[]{}, "MIR:00000133",
      new Class<?>[]{},
      "pid.pathway",
      "pi3kcipathway"),

  NCIT("NCIt",
      "http://ncit.nci.nih.gov/",
      new String[]{"urn:miriam:ncit",
          "http://identifiers.org/ncit/",
          "http://identifiers.org/ncit",
          "https://identifiers.org/ncit/",
          "https://identifiers.org/ncit",
      }, new Class<?>[]{}, "MIR:00000139",
      new Class<?>[]{},
      "ncit",
      "C80519"),

  NEUROLEX("NeuroLex",
      "http://www.neurolex.org/wiki/Main_Page",
      new String[]{"urn:miriam:neurolex",
          "http://identifiers.org/neurolex/",
          "http://identifiers.org/neurolex",
          "https://identifiers.org/neurolex/",
          "https://identifiers.org/neurolex",
      }, new Class<?>[]{}, "MIR:00000126",
      new Class<?>[]{},
      "neurolex",
      "Birnlex_721"),

  NEUROMORPHO("NeuroMorpho",
      "http://neuromorpho.org/index.jsp",
      new String[]{"urn:miriam:neuromorpho",
          "http://identifiers.org/neuromorpho/",
          "http://identifiers.org/neuromorpho",
          "https://identifiers.org/neuromorpho/",
          "https://identifiers.org/neuromorpho",
      }, new Class<?>[]{}, "MIR:00000095",
      new Class<?>[]{},
      "neuromorpho",
      "Rosa2"),

  NEURONDB("NeuronDB",
      "http://senselab.med.yale.edu/NeuronDB/",
      new String[]{"urn:miriam:neurondb",
          "http://identifiers.org/neurondb/",
          "http://identifiers.org/neurondb",
          "https://identifiers.org/neurondb/",
          "https://identifiers.org/neurondb",
      }, new Class<?>[]{}, "MIR:00000094",
      new Class<?>[]{},
      "neurondb",
      "265"),

  NEXTDB("NEXTDB",
      "http://nematode.lab.nig.ac.jp/",
      new String[]{"urn:miriam:nextdb",
          "http://identifiers.org/nextdb/",
          "http://identifiers.org/nextdb",
          "https://identifiers.org/nextdb/",
          "https://identifiers.org/nextdb",
      }, new Class<?>[]{}, "MIR:00000289",
      new Class<?>[]{},
      "nextdb",
      "6b1"),
  NEXTPROT("nextProt",
      "https://www.nextprot.org/",
      new String[]{"urn:miriam:nextprot",
          "http://identifiers.org/nextprot/",
          "http://identifiers.org/nextprot",
          "https://identifiers.org/nextprot/",
          "https://identifiers.org/nextprot",
      }, new Class<?>[]{}, "MIR:00000236",
      new Class<?>[]{},
      "nextprot",
      "NX_O00165"),
  NIAEST("NIAEST",
      "http://lgsun.grc.nia.nih.gov/cDNA/",
      new String[]{"urn:miriam:niaest",
          "http://identifiers.org/niaest/",
          "http://identifiers.org/niaest",
          "https://identifiers.org/niaest/",
          "https://identifiers.org/niaest",
      }, new Class<?>[]{}, "MIR:00000305",
      new Class<?>[]{},
      "niaest",
      "J0705A10"),

  NITE_BIOLOGICAL_RESEARCH_CENTER_CATALOGUE("NITE Biological Research Center Catalogue",
      "http://www.nbrc.nite.go.jp/e/index.html",
      new String[]{"urn:miriam:nbrc",
          "http://identifiers.org/nbrc/",
          "http://identifiers.org/nbrc",
          "https://identifiers.org/nbrc/",
          "https://identifiers.org/nbrc",
      }, new Class<?>[]{}, "MIR:00000179",
      new Class<?>[]{},
      "nbrc",
      "00001234"),

  NONCODE_V3("NONCODE v3",
      "http://www.noncode.org/",
      new String[]{"urn:miriam:noncodev3", "urn:miriam:noncode",
          "http://identifiers.org/noncodev3/",
          "http://identifiers.org/noncodev3",
          "https://identifiers.org/noncodev3/",
          "https://identifiers.org/noncodev3",
      }, new Class<?>[]{}, "MIR:00000248",
      new Class<?>[]{},
      "noncodev3",
      "377550"),
  NONCODE_V4_GENE("NONCODE v4 Gene",
      "http://www.bioinfo.org/NONCODEv4/",
      new String[]{"urn:miriam:noncodev4.gene",
          "http://identifiers.org/noncodev4.gene/",
          "http://identifiers.org/noncodev4.gene",
          "https://identifiers.org/noncodev4.gene/",
          "https://identifiers.org/noncodev4.gene",
      }, new Class<?>[]{}, "MIR:00000480",
      new Class<?>[]{},
      "noncodev4.gene",
      "NONHSAG00001"),
  NONCODE_V4_TRANSCRIPT("NONCODE v4 Transcript",
      "http://www.bioinfo.org/NONCODEv4/",
      new String[]{"urn:miriam:noncodev4.rna",
          "http://identifiers.org/noncodev4.rna/",
          "http://identifiers.org/noncodev4.rna",
          "https://identifiers.org/noncodev4.rna/",
          "https://identifiers.org/noncodev4.rna",
      }, new Class<?>[]{}, "MIR:00000481",
      new Class<?>[]{},
      "noncodev4.rna",
      "NONHSAT000001"),
  NORINE("NORINE",
      "http://bioinfo.lifl.fr/norine/",
      new String[]{"urn:miriam:norine",
          "http://identifiers.org/norine/",
          "http://identifiers.org/norine",
          "https://identifiers.org/norine/",
          "https://identifiers.org/norine",
      }, new Class<?>[]{}, "MIR:00000498",
      new Class<?>[]{},
      "norine",
      "NOR00681"),
  NUCLEARDB("NucleaRDB",
      "http://www.receptors.org/nucleardb/",
      new String[]{"urn:miriam:nuclearbd",
          "http://identifiers.org/nuclearbd/",
          "http://identifiers.org/nuclearbd",
          "https://identifiers.org/nuclearbd/",
          "https://identifiers.org/nuclearbd",
      }, new Class<?>[]{}, "MIR:00000356",
      new Class<?>[]{},
      "nuclearbd",
      "prgr_human"),

  NUCLEOTIDE_SEQUENCE_DATABASE("Nucleotide Sequence Database",
      "https://www.ncbi.nlm.nih.gov/Genbank/",
      new String[]{"urn:miriam:insdc",
          "http://identifiers.org/insdc/",
          "http://identifiers.org/insdc",
          "https://identifiers.org/insdc/",
          "https://identifiers.org/insdc",
      }, new Class<?>[]{}, "MIR:00000029",
      new Class<?>[]{},
      "insdc",
      "X58356"),

  OBI("Ontology for Biomedical Investigations",
      "https://www.ebi.ac.uk/ols/ontologies/obi/",
      new String[]{"urn:miriam:obi", "urn:miriam:obo.obi",
          "http://identifiers.org/obi/",
          "http://identifiers.org/obi",
          "https://identifiers.org/obi/",
          "https://identifiers.org/obi",
      },
      new Class<?>[]{}, "MIR:00000127",
      new Class<?>[]{},
      "obi",
      "OBI_0000070"),

  OMA_GROUP("OMA Group",
      "https://omabrowser.org/cgi-bin/gateway.pl",
      new String[]{"urn:miriam:oma.grp",
          "http://identifiers.org/oma.grp/",
          "http://identifiers.org/oma.grp",
          "https://identifiers.org/oma.grp/",
          "https://identifiers.org/oma.grp",
      }, new Class<?>[]{}, "MIR:00000343",
      new Class<?>[]{},
      "oma.grp",
      "LCSCCPN"),
  OMA_PROTEIN("OMA Protein",
      "https://omabrowser.org/cgi-bin/gateway.pl",
      new String[]{"urn:miriam:oma.protein",
          "http://identifiers.org/oma.protein/",
          "http://identifiers.org/oma.protein",
          "https://identifiers.org/oma.protein/",
          "https://identifiers.org/oma.protein",
      }, new Class<?>[]{}, "MIR:00000342",
      new Class<?>[]{},
      "oma.protein",
      "HUMAN16963"),
  ODOR_MOLECULES_DATABASE("Odor Molecules DataBase",
      "http://senselab.med.yale.edu/OdorDB",
      new String[]{"urn:miriam:odor",
          "http://identifiers.org/odor/",
          "http://identifiers.org/odor",
          "https://identifiers.org/odor/",
          "https://identifiers.org/odor",
      }, new Class<?>[]{}, "MIR:00000500",
      new Class<?>[]{},
      "odor",
      "75"),

  OLFACTORY_RECEPTOR_DATABASE("Olfactory Receptor Database",
      "http://senselab.med.yale.edu/OrDB/",
      new String[]{"urn:miriam:ordb",
          "http://identifiers.org/ordb/",
          "http://identifiers.org/ordb",
          "https://identifiers.org/ordb/",
          "https://identifiers.org/ordb",
      }, new Class<?>[]{}, "MIR:00000499",
      new Class<?>[]{},
      "ordb",
      "8497"),

  OMIA("OMIA",
      "http://omia.angis.org.au/",
      new String[]{"urn:miriam:omia",
          "http://identifiers.org/omia/",
          "http://identifiers.org/omia",
          "https://identifiers.org/omia/",
          "https://identifiers.org/omia",
      }, new Class<?>[]{}, "MIR:00000142",
      new Class<?>[]{},
      "omia",
      "1000"),

  /**
   * Online Mendelian Inheritance in Man: http://omim.org/.
   */
  OMIM("Online Mendelian Inheritance in Man",
      "http://omim.org/",
      new String[]{"urn:miriam:omim",
          "http://identifiers.org/mim/",
          "http://identifiers.org/mim",
          "https://identifiers.org/mim/",
          "https://identifiers.org/mim",
      },
      new Class<?>[]{Phenotype.class}, "MIR:00000016",
      new Class<?>[]{},
      "mim",
      "603903"),

  @NoMiriamUrn
  OMIT("OMIT",
      "https://www.ebi.ac.uk/ols/ontologies/omit/",
      new String[]{
          "http://identifiers.org/omit/",
          "http://identifiers.org/omit",
          "https://identifiers.org/omit/",
          "https://identifiers.org/omit",
      },
      new Class<?>[]{}, "MIR:00000605",
      new Class<?>[]{},
      "omit",
      "0005506"),

  ONTOLOGY_OF_PHYSICS_FOR_BIOLOGY("Ontology of Physics for Biology",
      "http://bioportal.bioontology.org/ontologies/OPB",
      new String[]{"urn:miriam:opb",
          "http://identifiers.org/opb/",
          "http://identifiers.org/opb",
          "https://identifiers.org/opb/",
          "https://identifiers.org/opb",
      }, new Class<?>[]{}, "MIR:00000129",
      new Class<?>[]{},
      "opb",
      "OPB_00573"),

  OPM("OPM",
      "http://opm.phar.umich.edu/",
      new String[]{"urn:miriam:opm",
          "http://identifiers.org/opm/",
          "http://identifiers.org/opm",
          "https://identifiers.org/opm/",
          "https://identifiers.org/opm",
      }, new Class<?>[]{}, "MIR:00000333",
      new Class<?>[]{},
      "opm",
      "1h68"),
  ORCID("ORCID",
      "https://orcid.org",
      new String[]{"urn:miriam:orcid",
          "http://identifiers.org/orcid/",
          "http://identifiers.org/orcid",
          "https://identifiers.org/orcid/",
          "https://identifiers.org/orcid",
      }, new Class<?>[]{}, "MIR:00000382",
      new Class<?>[]{},
      "orcid",
      "0000-0002-5355-2576"),

  ORIDB_SACCHAROMYCES("OriDB Saccharomyces",
      "http://cerevisiae.oridb.org/index.php",
      new String[]{"urn:miriam:oridb.sacch",
          "http://identifiers.org/oridb.sacch/",
          "http://identifiers.org/oridb.sacch",
          "https://identifiers.org/oridb.sacch/",
          "https://identifiers.org/oridb.sacch",
      }, new Class<?>[]{}, "MIR:00000369",
      new Class<?>[]{},
      "oridb.sacch",
      "9"),

  ORIDB_SCHIZOSACCHAROMYCES("OriDB Schizosaccharomyces",
      "http://pombe.oridb.org/index.php",
      new String[]{"urn:miriam:oridb.schizo",
          "http://identifiers.org/oridb.schizo/",
          "http://identifiers.org/oridb.schizo",
          "https://identifiers.org/oridb.schizo/",
          "https://identifiers.org/oridb.schizo",
      }, new Class<?>[]{}, "MIR:00000368",
      new Class<?>[]{},
      "oridb.schizo",
      "28"),

  ORPHANET("Orphanet",
      "http://www.orpha.net/consor/",
      new String[]{"urn:miriam:orphanet",
          "http://identifiers.org/orphanet/",
          "http://identifiers.org/orphanet",
          "https://identifiers.org/orphanet/",
          "https://identifiers.org/orphanet",
      }, new Class<?>[]{}, "MIR:00000220",
      new Class<?>[]{},
      "orphanet",
      "85163"),

  ORPHANET_RARE_DISEASE_ONTOLOGY("Orphanet Rare Disease Ontology",
      "https://www.ebi.ac.uk/ols/ontologies/ordo",
      new String[]{"urn:miriam:orphanet.ordo",
          "http://identifiers.org/orphanet.ordo/",
          "http://identifiers.org/orphanet.ordo",
          "https://identifiers.org/orphanet.ordo/",
          "https://identifiers.org/orphanet.ordo",
      }, new Class<?>[]{}, "MIR:00000532",
      new Class<?>[]{},
      "orphanet.ordo",
      "Orphanet_C023"),

  ORTHODB("OrthoDB",
      "http://cegg.unige.ch/orthodb4",
      new String[]{"urn:miriam:orthodb",
          "http://identifiers.org/orthodb/",
          "http://identifiers.org/orthodb",
          "https://identifiers.org/orthodb/",
          "https://identifiers.org/orthodb",
      }, new Class<?>[]{}, "MIR:00000221",
      new Class<?>[]{},
      "orthodb",
      "Q9P0K8"),

  ORYZABASE_GENE("Oryzabase Gene",
      "http://www.shigen.nig.ac.jp/rice/oryzabaseV4/",
      new String[]{"urn:miriam:oryzabase.gene",
          "http://identifiers.org/oryzabase.gene/",
          "http://identifiers.org/oryzabase.gene",
          "https://identifiers.org/oryzabase.gene/",
          "https://identifiers.org/oryzabase.gene",
      }, new Class<?>[]{}, "MIR:00000482",
      new Class<?>[]{},
      "oryzabase.gene",
      "117"),

  ORYZABASE_MUTANT("Oryzabase Mutant",
      "http://www.shigen.nig.ac.jp/rice/oryzabaseV4/",
      new String[]{"urn:miriam:oryzabase.mutant",
          "http://identifiers.org/oryzabase.mutant/",
          "http://identifiers.org/oryzabase.mutant",
          "https://identifiers.org/oryzabase.mutant/",
          "https://identifiers.org/oryzabase.mutant",
      }, new Class<?>[]{}, "MIR:00000483",
      new Class<?>[]{},
      "oryzabase.mutant",
      "21393"),

  ORYZABASE_STAGE("Oryzabase Stage",
      "http://www.shigen.nig.ac.jp/rice/oryzabaseV4/",
      new String[]{"urn:miriam:oryzabase.stage",
          "http://identifiers.org/oryzabase.stage/",
          "http://identifiers.org/oryzabase.stage",
          "https://identifiers.org/oryzabase.stage/",
          "https://identifiers.org/oryzabase.stage",
      }, new Class<?>[]{}, "MIR:00000485",
      new Class<?>[]{},
      "oryzabase.stage",
      "34"),

  ORYZABASE_STRAIN("Oryzabase Strain",
      "http://www.shigen.nig.ac.jp/rice/oryzabaseV4/",
      new String[]{"urn:miriam:oryzabase.strain",
          "http://identifiers.org/oryzabase.strain/",
          "http://identifiers.org/oryzabase.strain",
          "https://identifiers.org/oryzabase.strain/",
          "https://identifiers.org/oryzabase.strain",
      }, new Class<?>[]{}, "MIR:00000484",
      new Class<?>[]{},
      "oryzabase.strain",
      "109"),

  ORYZA_TAG_LINE("Oryza Tag Line",
      "http://oryzatagline.cirad.fr/",
      new String[]{"urn:miriam:otl",
          "http://identifiers.org/otl/",
          "http://identifiers.org/otl",
          "https://identifiers.org/otl/",
          "https://identifiers.org/otl",
      }, new Class<?>[]{}, "MIR:00000486",
      new Class<?>[]{},
      "otl",
      "AADB12"),

  P3DB_PROTEIN("P3DB Protein",
      "http://www.p3db.org/",
      new String[]{"urn:miriam:p3db.protein",
          "http://identifiers.org/p3db.protein/",
          "http://identifiers.org/p3db.protein",
          "https://identifiers.org/p3db.protein/",
          "https://identifiers.org/p3db.protein",
      }, new Class<?>[]{}, "MIR:00000501",
      new Class<?>[]{},
      "p3db.protein",
      "70"),

  P3DB_SITE("P3DB Site",
      "http://www.p3db.org/",
      new String[]{"urn:miriam:p3db.site",
          "http://identifiers.org/p3db.site/",
          "http://identifiers.org/p3db.site",
          "https://identifiers.org/p3db.site/",
          "https://identifiers.org/p3db.site",
      }, new Class<?>[]{}, "MIR:00000502",
      new Class<?>[]{},
      "p3db.site",
      "65"),

  PALEODB("PaleoDB",
      "http://paleodb.org/",
      new String[]{"urn:miriam:paleodb",
          "http://identifiers.org/paleodb/",
          "http://identifiers.org/paleodb",
          "https://identifiers.org/paleodb/",
          "https://identifiers.org/paleodb",
      }, new Class<?>[]{}, "MIR:00000197",
      new Class<?>[]{},
      "paleodb",
      "83088"),

  /**
   * PANTHER Family: http://www.pantherdb.org/.
   */
  PANTHER("PANTHER Family",
      "http://www.pantherdb.org/",
      new String[]{"urn:miriam:panther.family",
          "urn:miriam:panther",
          "http://identifiers.org/panther.family/",
          "http://identifiers.org/panther.family",
          "https://identifiers.org/panther.family/",
          "https://identifiers.org/panther.family",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000060",
      new Class<?>[]{},
      "panther.family",
      "PTHR12345"),

  PANTHER_NODE("PANTHER Node",
      "http://pantree.org/",
      new String[]{"urn:miriam:panther.node",
          "http://identifiers.org/panther.node/",
          "http://identifiers.org/panther.node",
          "https://identifiers.org/panther.node/",
          "https://identifiers.org/panther.node",
      }, new Class<?>[]{}, "MIR:00000374",
      new Class<?>[]{},
      "panther.node",
      "PTN000000026"),
  PANTHER_PATHWAY("PANTHER Pathway",
      "http://www.pantherdb.org/",
      new String[]{"urn:miriam:panther.pathway",
          "http://identifiers.org/panther.pathway/",
          "http://identifiers.org/panther.pathway",
          "https://identifiers.org/panther.pathway/",
          "https://identifiers.org/panther.pathway",
      }, new Class<?>[]{}, "MIR:00000363",
      new Class<?>[]{},
      "panther.pathway",
      "P00024"),

  PANTHER_PATHWAY_COMPONENT("PANTHER Pathway Component",
      "http://www.pantherdb.org/",
      new String[]{"urn:miriam:panther.pthcmp",
          "http://identifiers.org/panther.pthcmp/",
          "http://identifiers.org/panther.pthcmp",
          "https://identifiers.org/panther.pthcmp/",
          "https://identifiers.org/panther.pthcmp",
      }, new Class<?>[]{}, "MIR:00000422",
      new Class<?>[]{},
      "panther.pthcmp",
      "S00266"),

  PASS2("PASS2",
      "http://caps.ncbs.res.in/pass2/",
      new String[]{"urn:miriam:pass2",
          "http://identifiers.org/pass2/",
          "http://identifiers.org/pass2",
          "https://identifiers.org/pass2/",
          "https://identifiers.org/pass2",
      }, new Class<?>[]{}, "MIR:00000468",
      new Class<?>[]{},
      "pass2",
      "46977"),

  PATHWAY_ONTOLOGY("Pathway Ontology",
      "http://rgd.mcw.edu/rgdweb/ontology/search.html",
      new String[]{"urn:miriam:pw", "urn:miriam:obo.pw",
          "http://identifiers.org/pw/",
          "http://identifiers.org/PW",
          "https://identifiers.org/pw/",
          "https://identifiers.org/PW",
      }, new Class<?>[]{}, "MIR:00000242",
      new Class<?>[]{},
      "",
      "PW:0000208"),

  PATHWAY_COMMONS("Pathway Commons",
      "http://www.pathwaycommons.org/pc/",
      new String[]{"urn:miriam:pathwaycommons",
          "http://identifiers.org/pathwaycommons/",
          "http://identifiers.org/pathwaycommons",
          "https://identifiers.org/pathwaycommons/",
          "https://identifiers.org/pathwaycommons",
      }, new Class<?>[]{}, "MIR:00000073",
      new Class<?>[]{},
      "pathwaycommons",
      "485991"),

  PATO("PATO",
      "https://www.ebi.ac.uk/ols/ontologies/pato",
      new String[]{"urn:miriam:pato",
          "urn:miriam:obo.pato",
          "http://identifiers.org/pato/",
          "http://identifiers.org/PATO",
          "https://identifiers.org/pato/",
          "https://identifiers.org/PATO",
      },
      new Class<?>[]{}, "MIR:00000112",
      new Class<?>[]{},
      "",
      "PATO:0001998"),

  PAXDB_ORGANISM("PaxDb Organism",
      "http://pax-db.org/",
      new String[]{"urn:miriam:paxdb.organism",
          "http://identifiers.org/paxdb.organism/",
          "http://identifiers.org/paxdb.organism",
          "https://identifiers.org/paxdb.organism/",
          "https://identifiers.org/paxdb.organism",
      }, new Class<?>[]{}, "MIR:00000488",
      new Class<?>[]{},
      "paxdb.organism",
      "9606"),
  PAXDB_PROTEIN("PaxDb Protein",
      "http://pax-db.org/",
      new String[]{"urn:miriam:paxdb.protein",
          "http://identifiers.org/paxdb.protein/",
          "http://identifiers.org/paxdb.protein",
          "https://identifiers.org/paxdb.protein/",
          "https://identifiers.org/paxdb.protein",
      }, new Class<?>[]{}, "MIR:00000489",
      new Class<?>[]{},
      "paxdb.protein",
      "977869"),
  PAZAR_TRANSCRIPTION_FACTOR("Pazar Transcription Factor",
      "http://www.pazar.info/",
      new String[]{"urn:miriam:pazar",
          "http://identifiers.org/pazar/",
          "http://identifiers.org/pazar",
          "https://identifiers.org/pazar/",
          "https://identifiers.org/pazar",
      }, new Class<?>[]{}, "MIR:00000306",
      new Class<?>[]{},
      "pazar",
      "TF0001053"),

  /**
   * PDB: http://www.pdbe.org/.
   */
  PDB("Protein Data Bank",
      "http://www.pdbe.org/",
      new String[]{"urn:miriam:pdb",
          "http://identifiers.org/pdb/",
          "http://identifiers.org/pdb",
          "https://identifiers.org/pdb/",
          "https://identifiers.org/pdb",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000020",
      new Class<?>[]{},
      "pdb",
      "2gc4"),

  PDB_CCD("Chemical Component Dictionary",
      "https://www.ebi.ac.uk/pdbe-srv/pdbechem/",
      new String[]{"urn:miriam:pdb-ccd",
          "http://identifiers.org/pdb-ccd/",
          "http://identifiers.org/pdb-ccd",
          "https://identifiers.org/pdb-ccd/",
          "https://identifiers.org/pdb-ccd",
      },
      new Class<?>[]{}, "MIR:00000113",
      new Class<?>[]{},
      "pdb-ccd",
      "AB0"),

  PEPTIDEATLAS("PeptideAtlas",
      "http://www.peptideatlas.org/",
      new String[]{"urn:miriam:peptideatlas",
          "http://identifiers.org/peptideatlas/",
          "http://identifiers.org/peptideatlas",
          "https://identifiers.org/peptideatlas/",
          "https://identifiers.org/peptideatlas",
      }, new Class<?>[]{}, "MIR:00000053",
      new Class<?>[]{},
      "peptideatlas",
      "PAp00000009"),

  PEROXIBASE("Peroxibase",
      "http://peroxibase.toulouse.inra.fr/",
      new String[]{"urn:miriam:peroxibase",
          "http://identifiers.org/peroxibase/",
          "http://identifiers.org/peroxibase",
          "https://identifiers.org/peroxibase/",
          "https://identifiers.org/peroxibase",
      }, new Class<?>[]{}, "MIR:00000222",
      new Class<?>[]{},
      "peroxibase",
      "5282"),

  /**
   * Protein Family Database: http://pfam.xfam.org/.
   */
  PFAM("Protein Family Database",
      "http://pfam.xfam.org/",
      new String[]{"urn:miriam:pfam",
          "http://identifiers.org/pfam/",
          "http://identifiers.org/pfam",
          "https://identifiers.org/pfam/",
          "https://identifiers.org/pfam",
      },
      new Class<?>[]{}, "MIR:00000028",
      new Class<?>[]{},
      "pfam",
      "PF01234"),

  /**
   * PharmGKB Pathways: http://www.pharmgkb.org/.
   */
  PHARM("PharmGKB Pathways",
      "http://www.pharmgkb.org/",
      new String[]{"urn:miriam:pharmgkb.pathways",
          "http://identifiers.org/pharmgkb.pathways/",
          "http://identifiers.org/pharmgkb.pathways",
          "https://identifiers.org/pharmgkb.pathways/",
          "https://identifiers.org/pharmgkb.pathways",
      },
      new Class<?>[]{}, "MIR:00000089",
      new Class<?>[]{},
      "pharmgkb.pathways",
      "PA146123006"),

  PHARMGKB_DISEASE("PharmGKB Disease",
      "http://www.pharmgkb.org/",
      new String[]{"urn:miriam:pharmgkb.disease",
          "http://identifiers.org/pharmgkb.disease/",
          "http://identifiers.org/pharmgkb.disease",
          "https://identifiers.org/pharmgkb.disease/",
          "https://identifiers.org/pharmgkb.disease",
      }, new Class<?>[]{}, "MIR:00000090",
      new Class<?>[]{},
      "pharmgkb.disease",
      "PA447218"),

  PHARMGKB_DRUG("PharmGKB Drug",
      "http://www.pharmgkb.org/",
      new String[]{"urn:miriam:pharmgkb.drug",
          "http://identifiers.org/pharmgkb.drug/",
          "http://identifiers.org/pharmgkb.drug",
          "https://identifiers.org/pharmgkb.drug/",
          "https://identifiers.org/pharmgkb.drug",
      }, new Class<?>[]{}, "MIR:00000091",
      new Class<?>[]{},
      "pharmgkb.drug",
      "PA448710"),

  PHARMGKB_GENE("PharmGKB Gene",
      "http://www.pharmgkb.org/",
      new String[]{"urn:miriam:pharmgkb.gene",
          "http://identifiers.org/pharmgkb.gene/",
          "http://identifiers.org/pharmgkb.gene",
          "https://identifiers.org/pharmgkb.gene/",
          "https://identifiers.org/pharmgkb.gene",
      }, new Class<?>[]{}, "MIR:00000245",
      new Class<?>[]{},
      "pharmgkb.gene",
      "PA131"),

  PHENOL_EXPLORER("Phenol-Explorer",
      "http://www.phenol-explorer.eu/foods/",
      new String[]{"urn:miriam:phenolexplorer",
          "http://identifiers.org/phenolexplorer/",
          "http://identifiers.org/phenolexplorer",
          "https://identifiers.org/phenolexplorer/",
          "https://identifiers.org/phenolexplorer",
      }, new Class<?>[]{}, "MIR:00000268",
      new Class<?>[]{},
      "phenolexplorer",
      "76"),

  PHOSPHOPOINT_KINASE("PhosphoPoint Kinase",
      "http://kinase.bioinformatics.tw/",
      new String[]{"urn:miriam:phosphopoint.kinase",
          "http://identifiers.org/phosphopoint.kinase/",
          "http://identifiers.org/phosphopoint.kinase",
          "https://identifiers.org/phosphopoint.kinase/",
          "https://identifiers.org/phosphopoint.kinase",
      }, new Class<?>[]{}, "MIR:00000385",
      new Class<?>[]{},
      "phosphopoint.kinase",
      "AURKA"),

  PHOSPHOPOINT_PHOSPHOPROTEIN("PhosphoPoint Phosphoprotein",
      "http://kinase.bioinformatics.tw/",
      new String[]{"urn:miriam:phosphopoint.protein",
          "http://identifiers.org/phosphopoint.protein/",
          "http://identifiers.org/phosphopoint.protein",
          "https://identifiers.org/phosphopoint.protein/",
          "https://identifiers.org/phosphopoint.protein",
      }, new Class<?>[]{}, "MIR:00000386",
      new Class<?>[]{},
      "phosphopoint.protein",
      "AURKA"),

  PHOSPHOSITE_PROTEIN("PhosphoSite Protein",
      "http://www.phosphosite.org/homeAction.do",
      new String[]{"urn:miriam:phosphosite.protein", "urn:miriam:phosphosite",
          "http://identifiers.org/phosphosite.protein/",
          "http://identifiers.org/phosphosite.protein",
          "https://identifiers.org/phosphosite.protein/",
          "https://identifiers.org/phosphosite.protein",
      }, new Class<?>[]{}, "MIR:00000105",
      new Class<?>[]{},
      "phosphosite.protein",
      "12300"),

  PHOSPHOSITE_RESIDUE("PhosphoSite Residue",
      "http://www.phosphosite.org/homeAction.do",
      new String[]{"urn:miriam:phosphosite.residue",
          "http://identifiers.org/phosphosite.residue/",
          "http://identifiers.org/phosphosite.residue",
          "https://identifiers.org/phosphosite.residue/",
          "https://identifiers.org/phosphosite.residue",
      }, new Class<?>[]{}, "MIR:00000125",
      new Class<?>[]{},
      "phosphosite.residue",
      "2842"),

  PHYLOMEDB("PhylomeDB",
      "http://phylomedb.org/",
      new String[]{"urn:miriam:phylomedb",
          "http://identifiers.org/phylomedb/",
          "http://identifiers.org/phylomedb",
          "https://identifiers.org/phylomedb/",
          "https://identifiers.org/phylomedb",
      }, new Class<?>[]{}, "MIR:00000223",
      new Class<?>[]{},
      "phylomedb",
      "Phy000CLXM_RAT"),

  PHYTOZOME_LOCUS("Phytozome Locus",
      "http://www.phytozome.net/",
      new String[]{"urn:miriam:phytozome.locus",
          "http://identifiers.org/phytozome.locus/",
          "http://identifiers.org/phytozome.locus",
          "https://identifiers.org/phytozome.locus/",
          "https://identifiers.org/phytozome.locus",
      }, new Class<?>[]{}, "MIR:00000432",
      new Class<?>[]{},
      "phytozome.locus",
      "Glyma0021s00410"),

  PINA("PINA",
      "http://cbg.garvan.unsw.edu.au/pina/",
      new String[]{"urn:miriam:pina",
          "http://identifiers.org/pina/",
          "http://identifiers.org/pina",
          "https://identifiers.org/pina/",
          "https://identifiers.org/pina",
      }, new Class<?>[]{}, "MIR:00000359",
      new Class<?>[]{},
      "pina",
      "Q13485"),
  PIROPLASMADB("PiroplasmaDB",
      "http://piroplasmadb.org/",
      new String[]{"urn:miriam:piroplasma",
          "http://identifiers.org/piroplasma/",
          "http://identifiers.org/piroplasma",
          "https://identifiers.org/piroplasma/",
          "https://identifiers.org/piroplasma",
      }, new Class<?>[]{}, "MIR:00000351",
      new Class<?>[]{},
      "piroplasma",
      "TA14985"),

  PIRSF("PIRSF",
      "https://pir.georgetown.edu/",
      new String[]{"urn:miriam:pirsf",
          "http://identifiers.org/pirsf/",
          "http://identifiers.org/pirsf",
          "https://identifiers.org/pirsf/",
          "https://identifiers.org/pirsf",
      }, new Class<?>[]{}, "MIR:00000017",
      new Class<?>[]{},
      "pirsf",
      "PIRSF000100"),

  PLANT_ONTOLOGY("Plant Ontology",
      "http://www.plantontology.org/",
      new String[]{"urn:miriam:po", "urn:miriam:obo.po",
          "http://identifiers.org/po/",
          "http://identifiers.org/PO",
          "https://identifiers.org/po/",
          "https://identifiers.org/PO",
      }, new Class<?>[]{}, "MIR:00000307",
      new Class<?>[]{},
      "",
      "PO:0009089"),

  PLASMODB("PlasmoDB",
      "http://plasmodb.org/plasmo/",
      new String[]{"urn:miriam:plasmodb",
          "http://identifiers.org/plasmodb/",
          "http://identifiers.org/plasmodb",
          "https://identifiers.org/plasmodb/",
          "https://identifiers.org/plasmodb",
      }, new Class<?>[]{}, "MIR:00000150",
      new Class<?>[]{},
      "plasmodb",
      "PF11_0344"),

  PMC("PMC International",
      "http://europepmc.org/",
      new String[]{"urn:miriam:pmc",
          "http://identifiers.org/pmc/",
          "http://identifiers.org/pmc",
          "https://identifiers.org/pmc/",
          "https://identifiers.org/pmc",
      },
      new Class<?>[]{}, "MIR:00000147",
      new Class<?>[]{},
      "pmc",
      "PMC3084216"),

  POCKETOME("Pocketome",
      "http://www.pocketome.org/sfSearch.cgi?act=browseall",
      new String[]{"urn:miriam:pocketome",
          "http://identifiers.org/pocketome/",
          "http://identifiers.org/pocketome",
          "https://identifiers.org/pocketome/",
          "https://identifiers.org/pocketome",
      }, new Class<?>[]{}, "MIR:00000400",
      new Class<?>[]{},
      "pocketome",
      "1433C_TOBAC_1_252"),
  POLBASE("PolBase",
      "http://polbase.neb.com/",
      new String[]{"urn:miriam:polbase",
          "http://identifiers.org/polbase/",
          "http://identifiers.org/polbase",
          "https://identifiers.org/polbase/",
          "https://identifiers.org/polbase",
      }, new Class<?>[]{}, "MIR:00000355",
      new Class<?>[]{},
      "polbase",
      "19-T4"),
  POMBASE("PomBase",
      "http://www.pombase.org/",
      new String[]{"urn:miriam:pombase",
          "http://identifiers.org/pombase/",
          "http://identifiers.org/pombase",
          "https://identifiers.org/pombase/",
          "https://identifiers.org/pombase",
      }, new Class<?>[]{}, "MIR:00000335",
      new Class<?>[]{},
      "pombase",
      "SPCC13B11.01"),

  PRIDE("PRIDE",
      "https://www.ebi.ac.uk/pride/",
      new String[]{"urn:miriam:pride",
          "http://identifiers.org/pride/",
          "http://identifiers.org/pride",
          "https://identifiers.org/pride/",
          "https://identifiers.org/pride",
      }, new Class<?>[]{}, "MIR:00000065",
      new Class<?>[]{},
      "pride",
      "2"),

  PRIDE_PROJECT("PRIDE Project",
      "https://www.ebi.ac.uk/pride/",
      new String[]{"urn:miriam:pride.project",
          "http://identifiers.org/pride.project/",
          "http://identifiers.org/pride.project",
          "https://identifiers.org/pride.project/",
          "https://identifiers.org/pride.project",
      }, new Class<?>[]{}, "MIR:00000515",
      new Class<?>[]{},
      "pride.project",
      "PXD000440"),

  PRINTS("PRINTS",
      "http://www.bioinf.manchester.ac.uk/dbbrowser/sprint/",
      new String[]{"urn:miriam:sprint",
          "http://identifiers.org/prints/",
          "http://identifiers.org/prints",
          "https://identifiers.org/prints/",
          "https://identifiers.org/prints",
      },
      new Class<?>[]{}, "MIR:00000061",
      new Class<?>[]{},
      "prints",
      "PR00001"),

  PROGLYCPROT("ProGlycProt",
      "http://www.proglycprot.org/",
      new String[]{"urn:miriam:proglyc",
          "http://identifiers.org/proglyc/",
          "http://identifiers.org/proglyc",
          "https://identifiers.org/proglyc/",
          "https://identifiers.org/proglyc",
      }, new Class<?>[]{}, "MIR:00000354",
      new Class<?>[]{},
      "proglyc",
      "AC119"),

  PRODOM("ProDom",
      "http://prodom.prabi.fr/prodom/current/html/home.php",
      new String[]{"urn:miriam:prodom",
          "http://identifiers.org/prodom/",
          "http://identifiers.org/prodom",
          "https://identifiers.org/prodom/",
          "https://identifiers.org/prodom",
      }, new Class<?>[]{}, "MIR:00000117",
      new Class<?>[]{},
      "prodom",
      "PD10000"),

  PROSITE("PROSITE",
      "https://www.expasy.org/prosite/",
      new String[]{"urn:miriam:prosite",
          "http://identifiers.org/prosite/",
          "http://identifiers.org/prosite",
          "https://identifiers.org/prosite/",
          "https://identifiers.org/prosite",
      }, new Class<?>[]{}, "MIR:00000032",
      new Class<?>[]{},
      "prosite",
      "PS00001"),

  PROTCLUSTDB("ProtClustDB",
      "https://www.ncbi.nlm.nih.gov/proteinclusters?db=proteinclusters",
      new String[]{"urn:miriam:protclustdb",
          "http://identifiers.org/protclustdb/",
          "http://identifiers.org/protclustdb",
          "https://identifiers.org/protclustdb/",
          "https://identifiers.org/protclustdb",
      }, new Class<?>[]{}, "MIR:00000226",
      new Class<?>[]{},
      "protclustdb",
      "O80725"),

  PROTEIN_AFFINITY_REAGENTS("Protein Affinity Reagents",
      "https://www.ebi.ac.uk/ontology-lookup/browse.do?ontName=PAR",
      new String[]{"urn:miriam:psipar",
          "http://identifiers.org/psipar/",
          "http://identifiers.org/psipar",
          "https://identifiers.org/psipar/",
          "https://identifiers.org/psipar",
      }, new Class<?>[]{}, "MIR:00000533",
      new Class<?>[]{},
      "psipar",
      "PAR:0116"),

  PROTEIN_DATA_BANK_LIGAND("Protein Data Bank Ligand",
      "http://www.pdb.org/",
      new String[]{"urn:miriam:pdb.ligand",
          "http://identifiers.org/pdb.ligand/",
          "http://identifiers.org/pdb.ligand",
          "https://identifiers.org/pdb.ligand/",
          "https://identifiers.org/pdb.ligand",
      }, new Class<?>[]{}, "MIR:00000490",
      new Class<?>[]{},
      "pdb.ligand",
      "TRQ"),

  PROTEIN_MODEL_DATABASE("Protein Model Database",
      "https://bioinformatics.cineca.it/PMDB/",
      new String[]{"urn:miriam:pmdb",
          "http://identifiers.org/pmdb/",
          "http://identifiers.org/pmdb",
          "https://identifiers.org/pmdb/",
          "https://identifiers.org/pmdb",
      }, new Class<?>[]{}, "MIR:00000190",
      new Class<?>[]{},
      "pmdb",
      "PM0012345"),

  PROTEIN_ONTOLOGY("Protein Ontology",
      "https://proconsortium.org/",
      new String[]{"urn:miriam:pr", "urn:miriam:obo.pr",
          "http://identifiers.org/pr/",
          "http://identifiers.org/PR",
          "https://identifiers.org/pr/",
          "https://identifiers.org/PR",
      }, new Class<?>[]{}, "MIR:00000141",
      new Class<?>[]{},
      "",
      "PR:000000024"),

  PROTEOMICSDB_PEPTIDE("ProteomicsDB Peptide",
      "https://www.proteomicsdb.org/#peptideSearch",
      new String[]{"urn:miriam:proteomicsdb.peptide",
          "http://identifiers.org/proteomicsdb.peptide/",
          "http://identifiers.org/proteomicsdb.peptide",
          "https://identifiers.org/proteomicsdb.peptide/",
          "https://identifiers.org/proteomicsdb.peptide",
      }, new Class<?>[]{}, "MIR:00000525",
      new Class<?>[]{},
      "proteomicsdb.peptide",
      "53504"),
  PROTEOMICSDB_PROTEIN("ProteomicsDB Protein",
      "https://www.proteomicsdb.org/#human",
      new String[]{"urn:miriam:proteomicsdb.protein",
          "http://identifiers.org/proteomicsdb.protein/",
          "http://identifiers.org/proteomicsdb.protein",
          "https://identifiers.org/proteomicsdb.protein/",
          "https://identifiers.org/proteomicsdb.protein",
      }, new Class<?>[]{}, "MIR:00000524",
      new Class<?>[]{},
      "proteomicsdb.protein",
      "53504"),
  PROTONET_CLUSTER("ProtoNet Cluster",
      "http://www.protonet.cs.huji.ac.il/",
      new String[]{"urn:miriam:protonet.cluster",
          "http://identifiers.org/protonet.cluster/",
          "http://identifiers.org/protonet.cluster",
          "https://identifiers.org/protonet.cluster/",
          "https://identifiers.org/protonet.cluster",
      }, new Class<?>[]{}, "MIR:00000229",
      new Class<?>[]{},
      "protonet.cluster",
      "4349895"),
  PROTONET_PROTEINCARD("ProtoNet ProteinCard",
      "http://www.protonet.cs.huji.ac.il/",
      new String[]{"urn:miriam:protonet.proteincard",
          "http://identifiers.org/protonet.proteincard/",
          "http://identifiers.org/protonet.proteincard",
          "https://identifiers.org/protonet.proteincard/",
          "https://identifiers.org/protonet.proteincard",
      }, new Class<?>[]{}, "MIR:00000228",
      new Class<?>[]{},
      "protonet.proteincard",
      "16941567"),
  PSCDB("PSCDB",
      "http://idp1.force.cs.is.nagoya-u.ac.jp/pscdb/index.html",
      new String[]{"urn:miriam:pscdb",
          "http://identifiers.org/pscdb/",
          "http://identifiers.org/pscdb",
          "https://identifiers.org/pscdb/",
          "https://identifiers.org/pscdb",
      }, new Class<?>[]{}, "MIR:00000370",
      new Class<?>[]{},
      "pscdb",
      "051"),
  PSEUDOMONAS_GENOME_DATABASE("Pseudomonas Genome Database",
      "http://www.pseudomonas.com/",
      new String[]{"urn:miriam:pseudomonas",
          "http://identifiers.org/pseudomonas/",
          "http://identifiers.org/pseudomonas",
          "https://identifiers.org/pseudomonas/",
          "https://identifiers.org/pseudomonas",
      }, new Class<?>[]{}, "MIR:00000180",
      new Class<?>[]{},
      "pseudomonas",
      "PSEEN0001"),

  /**
   * PubChem-compound: http://pubchem.ncbi.nlm.nih.gov/.
   */
  PUBCHEM("PubChem-compound",
      "http://pubchem.ncbi.nlm.nih.gov/",
      new String[]{"urn:miriam:pubchem.compound",
          "http://identifiers.org/pubchem.compound/",
          "http://identifiers.org/pubchem.compound",
          "https://identifiers.org/pubchem.compound/",
          "https://identifiers.org/pubchem.compound",
      },
      new Class<?>[]{Chemical.class}, "MIR:00000034",
      new Class<?>[]{Chemical.class},
      "pubchem.compound",
      "100101"),

  PUBCHEM_BIOASSAY("PubChem-bioassay",
      "https://www.ncbi.nlm.nih.gov/sites/entrez?db=pcassay",
      new String[]{"urn:miriam:pubchem.bioassay",
          "http://identifiers.org/pubchem.bioassay/",
          "http://identifiers.org/pubchem.bioassay",
          "https://identifiers.org/pubchem.bioassay/",
          "https://identifiers.org/pubchem.bioassay",
      }, new Class<?>[]{}, "MIR:00000072",
      new Class<?>[]{},
      "pubchem.bioassay",
      "1018"),

  /**
   * PubChem-substance: http://pubchem.ncbi.nlm.nih.gov/.
   */
  PUBCHEM_SUBSTANCE("PubChem-substance",
      "http://pubchem.ncbi.nlm.nih.gov/",
      new String[]{"urn:miriam:pubchem.substance",
          "http://identifiers.org/pubchem.substance/",
          "http://identifiers.org/pubchem.substance",
          "https://identifiers.org/pubchem.substance/",
          "https://identifiers.org/pubchem.substance",
      },
      new Class<?>[]{Chemical.class}, "MIR:00000033",
      new Class<?>[]{Chemical.class},
      "pubchem.substance",
      "100101"),

  /**
   * PubMed: http://www.ncbi.nlm.nih.gov/PubMed/.
   */
  PUBMED("PubMed",
      "http://www.ncbi.nlm.nih.gov/PubMed/",
      new String[]{"urn:miriam:pubmed",
          "http://identifiers.org/pubmed/",
          "http://identifiers.org/pubmed",
          "https://identifiers.org/pubmed/",
          "https://identifiers.org/pubmed",
      },
      new Class<?>[]{BioEntity.class}, "MIR:00000015",
      new Class<?>[]{Reaction.class},
      "pubmed",
      "28725475"),

  RAT_GENOME_DATABASE_QTL("Rat Genome Database qTL",
      "http://rgd.mcw.edu/",
      new String[]{"urn:miriam:rgd.qtl",
          "http://identifiers.org/rgd.qtl/",
          "http://identifiers.org/rgd.qtl",
          "https://identifiers.org/rgd.qtl/",
          "https://identifiers.org/rgd.qtl",
      }, new Class<?>[]{}, "MIR:00000451",
      new Class<?>[]{},
      "rgd.qtl",
      "1354581"),

  RAT_GENOME_DATABASE_STRAIN("Rat Genome Database strain",
      "http://rgd.mcw.edu/",
      new String[]{"urn:miriam:rgd.strain",
          "http://identifiers.org/rgd.strain/",
          "http://identifiers.org/rgd.strain",
          "https://identifiers.org/rgd.strain/",
          "https://identifiers.org/rgd.strain",
      }, new Class<?>[]{}, "MIR:00000452",
      new Class<?>[]{},
      "rgd.strain",
      "5688061"),

  /**
   * Reactome: http://www.reactome.org/.
   */
  REACTOME("Reactome",
      "http://www.reactome.org/",
      new String[]{"urn:miriam:reactome",
          "http://identifiers.org/reactome/",
          "http://identifiers.org/reactome",
          "https://identifiers.org/reactome/",
          "https://identifiers.org/reactome",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000018",
      new Class<?>[]{},
      "reactome",
      "R-HSA-201451"),

  REBASE("REBASE",
      "http://rebase.neb.com/rebase/",
      new String[]{"urn:miriam:rebase",
          "http://identifiers.org/rebase/",
          "http://identifiers.org/rebase",
          "https://identifiers.org/rebase/",
          "https://identifiers.org/rebase",
      }, new Class<?>[]{}, "MIR:00000230",
      new Class<?>[]{},
      "rebase",
      "101"),

  /**
   * RefSeq: http://www.ncbi.nlm.nih.gov/projects/RefSeq/.
   */
  REFSEQ("RefSeq",
      "http://www.ncbi.nlm.nih.gov/projects/RefSeq/",
      new String[]{"urn:miriam:refseq",
          "http://identifiers.org/refseq/",
          "http://identifiers.org/refseq",
          "https://identifiers.org/refseq/",
          "https://identifiers.org/refseq",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000039",
      new Class<?>[]{},
      "refseq",
      "NP_012345"),

  RELATION_ONTOLOGY("Relation Ontology",
      "http://obofoundry.org/ontology/ro.html",
      new String[]{"urn:miriam:ro", "urn:miriam:obo.ro",
          "http://identifiers.org/ro/",
          "http://identifiers.org/ro",
          "https://identifiers.org/ro/",
          "https://identifiers.org/ro",
      }, new Class<?>[]{}, "MIR:00000120",
      new Class<?>[]{},
      "ro",
      "RO_0002533"),

  RESID("RESID",
      "http://pir0.georgetown.edu/resid/",
      new String[]{"urn:miriam:resid",
          "http://identifiers.org/resid/",
          "http://identifiers.org/resid",
          "https://identifiers.org/resid/",
          "https://identifiers.org/resid",
      }, new Class<?>[]{}, "MIR:00000046",
      new Class<?>[]{},
      "resid",
      "AA0001"),

  RFAM("RFAM",
      "https://rfam.xfam.org/",
      new String[]{"urn:miriam:rfam",
          "http://identifiers.org/rfam/",
          "http://identifiers.org/rfam",
          "https://identifiers.org/rfam/",
          "https://identifiers.org/rfam",
      }, new Class<?>[]{}, "MIR:00000409",
      new Class<?>[]{},
      "rfam",
      "RF00230"),

  RGD("Rat Genome Database",
      "http://rgd.mcw.edu/",
      new String[]{"urn:miriam:rgd",
          "http://identifiers.org/rgd/",
          "http://identifiers.org/rgd",
          "https://identifiers.org/rgd/",
          "https://identifiers.org/rgd",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000047",
      new Class<?>[]{},
      "rgd",
      "2018"),

  /**
   * Rhea: http://www.rhea-db.org/.
   */
  RHEA("Rhea",
      "http://www.rhea-db.org/",
      new String[]{"urn:miriam:rhea",
          "http://identifiers.org/rhea/",
          "http://identifiers.org/rhea",
          "https://identifiers.org/rhea/",
          "https://identifiers.org/rhea",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000082",
      new Class<?>[]{},
      "rhea",
      "12345"),

  RICE_GENOME_ANNOTATION_PROJECT("Rice Genome Annotation Project",
      "http://rice.plantbiology.msu.edu/annotation_pseudo_current.shtml",
      new String[]{"urn:miriam:ricegap",
          "http://identifiers.org/ricegap/",
          "http://identifiers.org/ricegap",
          "https://identifiers.org/ricegap/",
          "https://identifiers.org/ricegap",
      }, new Class<?>[]{}, "MIR:00000358",
      new Class<?>[]{},
      "ricegap",
      "LOC_Os02g13300"),
  RNA_MODIFICATION_DATABASE("RNA Modification Database",
      "http://rna-mdb.cas.albany.edu/RNAmods/rnaover.htm",
      new String[]{"urn:miriam:rnamods",
          "http://identifiers.org/rnamods/",
          "http://identifiers.org/rnamods",
          "https://identifiers.org/rnamods/",
          "https://identifiers.org/rnamods",
      }, new Class<?>[]{}, "MIR:00000308",
      new Class<?>[]{},
      "rnamods",
      "101"),
  ROUGE("Rouge",
      "http://www.kazusa.or.jp/rouge/",
      new String[]{"urn:miriam:rouge",
          "http://identifiers.org/rouge/",
          "http://identifiers.org/rouge",
          "https://identifiers.org/rouge/",
          "https://identifiers.org/rouge",
      }, new Class<?>[]{}, "MIR:00000293",
      new Class<?>[]{},
      "rouge",
      "mKIAA4200"),

  SABIO_RK_EC_RECORD("SABIO-RK EC Record",
      "http://sabiork.h-its.org/",
      new String[]{"urn:miriam:sabiork.ec",
          "http://identifiers.org/sabiork.ec/",
          "http://identifiers.org/sabiork.ec",
          "https://identifiers.org/sabiork.ec/",
          "https://identifiers.org/sabiork.ec",
      }, new Class<?>[]{}, "MIR:00000128",
      new Class<?>[]{},
      "sabiork.ec",
      "2.7.1.1"),

  SABIO_RK_KINETIC_RECORD("SABIO-RK Kinetic Record",
      "http://sabiork.h-its.org/",
      new String[]{"urn:miriam:sabiork.kineticrecord",
          "http://identifiers.org/sabiork.kineticrecord/",
          "http://identifiers.org/sabiork.kineticrecord",
          "https://identifiers.org/sabiork.kineticrecord/",
          "https://identifiers.org/sabiork.kineticrecord",
      }, new Class<?>[]{}, "MIR:00000086",
      new Class<?>[]{},
      "sabiork.kineticrecord",
      "5046"),

  SABIO_RK_REACTION("SABIO-RK Reaction",
      "http://sabiork.h-its.org/",
      new String[]{"urn:miriam:sabiork.reaction",
          "http://identifiers.org/sabiork.reaction/",
          "http://identifiers.org/sabiork.reaction",
          "https://identifiers.org/sabiork.reaction/",
          "https://identifiers.org/sabiork.reaction",
      }, new Class<?>[]{}, "MIR:00000038",
      new Class<?>[]{},
      "sabiork.reaction",
      "77"),

  SACCHAROMYCES_GENOME_DATABASE_PATHWAYS("Saccharomyces genome database pathways",
      "http://pathway.yeastgenome.org/",
      new String[]{"urn:miriam:sgd.pathways",
          "http://identifiers.org/sgd.pathways/",
          "http://identifiers.org/sgd.pathways",
          "https://identifiers.org/sgd.pathways/",
          "https://identifiers.org/sgd.pathways",
      }, new Class<?>[]{}, "MIR:00000057",
      new Class<?>[]{},
      "sgd.pathways",
      "PWY3O-214"),

  SBML_RDF_VOCABULARY("SBML RDF Vocabulary",
      "http://biomodels.net/rdf/vocabulary.rdf",
      new String[]{"urn:miriam:biomodels.vocabulary",
          "http://identifiers.org/biomodels.vocabulary/",
          "http://identifiers.org/biomodels.vocabulary",
          "https://identifiers.org/biomodels.vocabulary/",
          "https://identifiers.org/biomodels.vocabulary",
      }, new Class<?>[]{}, "MIR:00000514",
      new Class<?>[]{},
      "biomodels.vocabulary",
      "rateRule"),

  SBO_TERM("Systems Biology Ontology",
      "http://www.ebi.ac.uk/sbo/main/",
      new String[]{"urn:miriam:biomodels.sbo", "urn:miriam:obo.sbo", "urn:miriam:sbo",
          "http://identifiers.org/biomodels.sbo/",
          "http://identifiers.org/sbo/",
          "http://identifiers.org/SBO",
          "https://identifiers.org/biomodels.sbo/",
          "https://identifiers.org/sbo/",
          "https://identifiers.org/SBO",
      },
      new Class<?>[]{}, "MIR:00000024",
      new Class<?>[]{},
      "",
      "SBO:0000262"),

  SCOP("SCOP",
      "http://scop.mrc-lmb.cam.ac.uk/scop/",
      new String[]{"urn:miriam:scop",
          "http://identifiers.org/scop/",
          "http://identifiers.org/scop",
          "https://identifiers.org/scop/",
          "https://identifiers.org/scop",
      }, new Class<?>[]{}, "MIR:00000371",
      new Class<?>[]{},
      "scop",
      "47419"),

  SCERTF("ScerTF",
      "http://stormo.wustl.edu/ScerTF/",
      new String[]{"urn:miriam:scretf",
          "http://identifiers.org/scretf/",
          "http://identifiers.org/scretf",
          "https://identifiers.org/scretf/",
          "https://identifiers.org/scretf",
      }, new Class<?>[]{}, "MIR:00000244",
      new Class<?>[]{},
      "scretf",
      "RSC3"),

  @NoMiriamUrn
  SEED_COMPOUND("SEED Compound",
      "http://modelseed.org/",
      new String[]{
          "http://identifiers.org/seed.compound/",
          "http://identifiers.org/seed.compound",
          "https://identifiers.org/seed.compound/",
          "https://identifiers.org/seed.compound",
      }, new Class<?>[]{}, "MIR:00000553",
      new Class<?>[]{},
      "seed.compound",
      "cpd15380"),

  @NoMiriamUrn
  SEED_REACTIONS("SEED Reactions",
      "http://modelseed.org/biochem/reactions/",
      new String[]{
          "http://identifiers.org/seed.reaction/",
          "http://identifiers.org/seed.reaction",
          "https://identifiers.org/seed.reaction/",
          "https://identifiers.org/seed.reaction",
      }, new Class<?>[]{}, "MIR:00000692",
      new Class<?>[]{},
      "seed.reaction",
      "rxn00001"),

  SEQUENCE_ONTOLOGY("Sequence Ontology",
      "http://www.sequenceontology.org/",
      new String[]{"urn:miriam:so", "urn:miriam:obo.so",
          "http://identifiers.org/so/",
          "http://identifiers.org/SO",
          "https://identifiers.org/so/",
          "https://identifiers.org/SO",
      }, new Class<?>[]{}, "MIR:00000081",
      new Class<?>[]{},
      "",
      "SO:0000704"),

  SEQUENCE_READ_ARCHIVE("Sequence Read Archive",
      "https://www.ncbi.nlm.nih.gov/sra",
      new String[]{"urn:miriam:insdc.sra",
          "http://identifiers.org/insdc.sra/",
          "http://identifiers.org/insdc.sra",
          "https://identifiers.org/insdc.sra/",
          "https://identifiers.org/insdc.sra",
      }, new Class<?>[]{}, "MIR:00000243",
      new Class<?>[]{},
      "insdc.sra",
      "SRX000007"),

  /**
   * SGD: http://www.yeastgenome.org/.
   */
  SGD("Saccharomyces Genome Database",
      "http://www.yeastgenome.org/",
      new String[]{"urn:miriam:sgd",
          "http://identifiers.org/sgd/",
          "http://identifiers.org/sgd",
          "https://identifiers.org/sgd/",
          "https://identifiers.org/sgd",
      },
      new Class<?>[]{}, "MIR:00000023",
      new Class<?>[]{},
      "sgd",
      "S000003909"),

  SIDER_DRUG("SIDER Drug",
      "http://sideeffects.embl.de/",
      new String[]{"urn:miriam:sider.drug",
          "http://identifiers.org/sider.drug/",
          "http://identifiers.org/sider.drug",
          "https://identifiers.org/sider.drug/",
          "https://identifiers.org/sider.drug",
      }, new Class<?>[]{}, "MIR:00000435",
      new Class<?>[]{},
      "sider.drug",
      "2244"),
  SIDER_SIDE_EFFECT("SIDER Side Effect",
      "http://sideeffects.embl.de/",
      new String[]{"urn:miriam:sider.effect",
          "http://identifiers.org/sider.effect/",
          "http://identifiers.org/sider.effect",
          "https://identifiers.org/sider.effect/",
          "https://identifiers.org/sider.effect",
      }, new Class<?>[]{}, "MIR:00000436",
      new Class<?>[]{},
      "sider.effect",
      "C0017565"),

  SIGNALING_GATEWAY("Signaling Gateway",
      "http://www.signaling-gateway.org/molecule",
      new String[]{"urn:miriam:signaling-gateway",
          "http://identifiers.org/signaling-gateway/",
          "http://identifiers.org/signaling-gateway",
          "https://identifiers.org/signaling-gateway/",
          "https://identifiers.org/signaling-gateway",
      }, new Class<?>[]{}, "MIR:00000045",
      new Class<?>[]{},
      "signaling-gateway",
      "A001094"),

  SITEX("SitEx",
      "http://www-bionet.sscc.ru/sitex/",
      new String[]{"urn:miriam:sitex",
          "http://identifiers.org/sitex/",
          "http://identifiers.org/sitex",
          "https://identifiers.org/sitex/",
          "https://identifiers.org/sitex",
      }, new Class<?>[]{}, "MIR:00000252",
      new Class<?>[]{},
      "sitex",
      "1004"),

  SMALL_MOLECULE_PATHWAY_DATABASE("Small Molecule Pathway Database",
      "https://smpdb.ca/",
      new String[]{"urn:miriam:smpdb",
          "http://identifiers.org/smpdb/",
          "http://identifiers.org/smpdb",
          "https://identifiers.org/smpdb/",
          "https://identifiers.org/smpdb",
      }, new Class<?>[]{}, "MIR:00000104",
      new Class<?>[]{},
      "smpdb",
      "SMP00001"),

  SMART("SMART",
      "http://smart.embl-heidelberg.de/",
      new String[]{"urn:miriam:smart",
          "http://identifiers.org/smart/",
          "http://identifiers.org/smart",
          "https://identifiers.org/smart/",
          "https://identifiers.org/smart",
      }, new Class<?>[]{}, "MIR:00000118",
      new Class<?>[]{},
      "smart",
      "SM00015"),

  SNOMED_CT("SNOMED CT",
      "http://www.snomedbrowser.com/",
      new String[]{"urn:miriam:snomedct",
          "http://identifiers.org/snomedct/",
          "http://identifiers.org/snomedct",
          "https://identifiers.org/snomedct/",
          "https://identifiers.org/snomedct",
      }, new Class<?>[]{}, "MIR:00000269",
      new Class<?>[]{},
      "snomedct",
      "284196006"),

  SOL_GENOMICS_NETWORK("Sol Genomics Network",
      "http://solgenomics.net/",
      new String[]{"urn:miriam:sgn",
          "http://identifiers.org/sgn/",
          "http://identifiers.org/sgn",
          "https://identifiers.org/sgn/",
          "https://identifiers.org/sgn",
      }, new Class<?>[]{}, "MIR:00000185",
      new Class<?>[]{},
      "sgn",
      "0001"),

  SOYBASE("SoyBase",
      "http://soybase.org/",
      new String[]{"urn:miriam:soybase",
          "http://identifiers.org/soybase/",
          "http://identifiers.org/soybase",
          "https://identifiers.org/soybase/",
          "https://identifiers.org/soybase",
      }, new Class<?>[]{}, "MIR:00000291",
      new Class<?>[]{},
      "soybase",
      "BARC-013845-01256"),

  SPECTRAL_DATABASE_FOR_ORGANIC_COMPOUNDS("Spectral Database for Organic Compounds",
      "http://riodb01.ibase.aist.go.jp/sdbs/cgi-bin/direct_frame_top.cgi",
      new String[]{"urn:miriam:sdbs",
          "http://identifiers.org/sdbs/",
          "http://identifiers.org/sdbs",
          "https://identifiers.org/sdbs/",
          "https://identifiers.org/sdbs",
      }, new Class<?>[]{}, "MIR:00000319",
      new Class<?>[]{},
      "sdbs",
      "4544"),

  SPIKE("SPIKE Map",
      "http://www.cs.tau.ac.il/~spike/",
      new String[]{"urn:miriam:spike.map",
          "http://identifiers.org/spike.map/",
          "http://identifiers.org/spike.map",
          "https://identifiers.org/spike.map/",
          "https://identifiers.org/spike.map",
      },
      new Class<?>[]{}, "MIR:00000321",
      new Class<?>[]{},
      "spike.map",
      "spike00001"),

  STAP("STAP",
      "http://psb.kobic.re.kr/STAP/refinement/",
      new String[]{"urn:miriam:stap",
          "http://identifiers.org/stap/",
          "http://identifiers.org/stap",
          "https://identifiers.org/stap/",
          "https://identifiers.org/stap",
      }, new Class<?>[]{}, "MIR:00000399",
      new Class<?>[]{},
      "stap",
      "1a24"),

  /**
   * STITCH: http://stitch.embl.de/.
   */
  STITCH("STITCH",
      "http://stitch.embl.de/",
      new String[]{"urn:miriam:stitch",
          "http://identifiers.org/stitch/",
          "http://identifiers.org/stitch",
          "https://identifiers.org/stitch/",
          "https://identifiers.org/stitch",
      },
      new Class<?>[]{}, "MIR:00000266",
      new Class<?>[]{},
      "stitch",
      "BQJCRHHNABKAKU"),

  /**
   * STRING: http://string-db.org/.
   */
  STRING("STRING",
      "http://string-db.org/",
      new String[]{"urn:miriam:string",
          "http://identifiers.org/string/",
          "http://identifiers.org/string",
          "https://identifiers.org/string/",
          "https://identifiers.org/string",
      },
      new Class<?>[]{}, "MIR:00000265",
      new Class<?>[]{},
      "string",
      "P53350"),

  SUBSTRATEDB("SubstrateDB",
      "http://substrate.burnham.org/",
      new String[]{"urn:miriam:pmap.substratedb",
          "http://identifiers.org/pmap.substratedb/",
          "http://identifiers.org/pmap.substratedb",
          "https://identifiers.org/pmap.substratedb/",
          "https://identifiers.org/pmap.substratedb",
      }, new Class<?>[]{}, "MIR:00000224",
      new Class<?>[]{},
      "pmap.substratedb",
      "1915"),

  SUBTILIST("SubtiList",
      "http://genolist.pasteur.fr/SubtiList/",
      new String[]{"urn:miriam:subtilist",
          "http://identifiers.org/subtilist/",
          "http://identifiers.org/subtilist",
          "https://identifiers.org/subtilist/",
          "https://identifiers.org/subtilist",
      }, new Class<?>[]{}, "MIR:00000433",
      new Class<?>[]{},
      "subtilist",
      "BG11523"),

  SUPFAM("SUPFAM",
      "http://supfam.org/SUPERFAMILY/",
      new String[]{"urn:miriam:supfam",
          "http://identifiers.org/supfam/",
          "http://identifiers.org/supfam",
          "https://identifiers.org/supfam/",
          "https://identifiers.org/supfam",
      }, new Class<?>[]{}, "MIR:00000357",
      new Class<?>[]{},
      "supfam",
      "SSF57615"),

  SUBTIWIKI("SubtiWiki",
      "http://www.subtiwiki.uni-goettingen.de/wiki/index.php/Main_Page",
      new String[]{"urn:miriam:subtiwiki",
          "http://identifiers.org/subtiwiki/",
          "http://identifiers.org/subtiwiki",
          "https://identifiers.org/subtiwiki/",
          "https://identifiers.org/subtiwiki",
      }, new Class<?>[]{}, "MIR:00000132",
      new Class<?>[]{},
      "subtiwiki",
      "BSU29180"),

  @NoMiriamUrn
  SWISS_LIPIDS("SwissLipids",
      "http://www.swisslipids.org/#/",
      new String[]{
          "http://identifiers.org/slm/",
          "http://identifiers.org/SLM",
          "https://identifiers.org/slm/",
          "https://identifiers.org/SLM",
      },
      new Class<?>[]{}, "MIR:00000550",
      new Class<?>[]{},
      "",
      "SLM:000048885"),

  SWISS_MODEL("SWISS-MODEL",
      "https://swissmodel.expasy.org",
      new String[]{"urn:miriam:swiss-model",
          "http://identifiers.org/swiss-model/",
          "http://identifiers.org/swiss-model",
          "https://identifiers.org/swiss-model/",
          "https://identifiers.org/swiss-model",
      },
      new Class<?>[]{}, "MIR:00000231",
      new Class<?>[]{},
      "swiss-model",
      "P23298"),

  T3DB("T3DB",
      "http://www.t3db.org/",
      new String[]{"urn:miriam:t3db",
          "http://identifiers.org/t3db/",
          "http://identifiers.org/t3db",
          "https://identifiers.org/t3db/",
          "https://identifiers.org/t3db",
      }, new Class<?>[]{}, "MIR:00000103",
      new Class<?>[]{},
      "t3db",
      "T3D0001"),

  TAIR_GENE("TAIR Gene",
      "http://arabidopsis.org/index.jsp",
      new String[]{"urn:miriam:tair.gene",
          "http://identifiers.org/tair.gene/",
          "http://identifiers.org/tair.gene",
          "https://identifiers.org/tair.gene/",
          "https://identifiers.org/tair.gene",
      }, new Class<?>[]{}, "MIR:00000049",
      new Class<?>[]{},
      "tair.gene",
      "Gene:2200934"),

  TAIR_PROTEIN("TAIR Protein",
      "http://arabidopsis.org/index.jsp",
      new String[]{"urn:miriam:tair.protein",
          "http://identifiers.org/tair.protein/",
          "http://identifiers.org/tair.protein",
          "https://identifiers.org/tair.protein/",
          "https://identifiers.org/tair.protein",
      }, new Class<?>[]{}, "MIR:00000048",
      new Class<?>[]{},
      "tair.protein",
      "AASequence:1009107926"),

  /**
   * The Arabidopsis Information Resource (TAIR) maintains a database of genetic
   * and molecular biology data for the model higher plant Arabidopsis thaliana.
   * The name of a Locus is unique and used by TAIR, TIGR, and MIPS:
   * http://arabidopsis.org/index.jsp.
   */
  TAIR_LOCUS("TAIR Locus",
      "http://arabidopsis.org/index.jsp",
      new String[]{"urn:miriam:tair.locus",
          "http://identifiers.org/tair.locus/",
          "http://identifiers.org/tair.locus",
          "https://identifiers.org/tair.locus/",
          "https://identifiers.org/tair.locus",
      },
      new Class<?>[]{}, "MIR:00000050",
      new Class<?>[]{},
      "tair.locus",
      "2200950"),

  TARBASE("TarBase",
      "http://diana.imis.athena-innovation.gr/DianaTools/index.php?r=tarbase/index",
      new String[]{"urn:miriam:tarbase",
          "http://identifiers.org/tarbase/",
          "http://identifiers.org/tarbase",
          "https://identifiers.org/tarbase/",
          "https://identifiers.org/tarbase",
      }, new Class<?>[]{}, "MIR:00000340",
      new Class<?>[]{},
      "tarbase",
      "hsa-let-7a-2-3p"),

  /**
   * Taxonomy: http://www.ncbi.nlm.nih.gov/taxonomy/.
   */
  TAXONOMY("Taxonomy",
      "http://www.ncbi.nlm.nih.gov/taxonomy/",
      new String[]{"urn:miriam:taxonomy",
          "http://identifiers.org/taxonomy/",
          "http://identifiers.org/taxonomy",
          "https://identifiers.org/taxonomy/",
          "https://identifiers.org/taxonomy",
      },
      new Class<?>[]{}, "MIR:00000006",
      new Class<?>[]{},
      "taxonomy",
      "9606"),

  TEDDY("TEDDY",
      "http://teddyontology.sourceforge.net/",
      new String[]{"urn:miriam:biomodels.teddy", "urn:miriam:teddy",
          "http://identifiers.org/biomodels.teddy/",
          "http://identifiers.org/biomodels.teddy",
          "https://identifiers.org/biomodels.teddy/",
          "https://identifiers.org/biomodels.teddy",
      }, new Class<?>[]{}, "MIR:00000107",
      new Class<?>[]{},
      "biomodels.teddy",
      "TEDDY_0000066"),

  TETRAHYMENA_GENOME_DATABASE("Tetrahymena Genome Database",
      "http://ciliate.org/index.php/",
      new String[]{"urn:miriam:tgd",
          "http://identifiers.org/tgd/",
          "http://identifiers.org/tgd",
          "https://identifiers.org/tgd/",
          "https://identifiers.org/tgd",
      }, new Class<?>[]{}, "MIR:00000313",
      new Class<?>[]{},
      "tgd",
      "TTHERM_00648910"),
  TIGRFAMS("TIGRFAMS",
      "http://www.jcvi.org/cgi-bin/tigrfams/Listing.cgi",
      new String[]{"urn:miriam:tigrfam",
          "http://identifiers.org/tigrfam/",
          "http://identifiers.org/tigrfam",
          "https://identifiers.org/tigrfam/",
          "https://identifiers.org/tigrfam",
      }, new Class<?>[]{}, "MIR:00000315",
      new Class<?>[]{},
      "tigrfam",
      "TIGR00010"),
  TISSUE_LIST("Tissue List",
      "https://www.uniprot.org/docs/tisslist.txt",
      new String[]{"urn:miriam:tissuelist",
          "http://identifiers.org/tissuelist/",
          "http://identifiers.org/tissuelist",
          "https://identifiers.org/tissuelist/",
          "https://identifiers.org/tissuelist",
      }, new Class<?>[]{}, "MIR:00000360",
      new Class<?>[]{},
      "tissuelist",
      "TS-0285"),

  TOPDB("TOPDB",
      "http://topdb.enzim.hu/",
      new String[]{"urn:miriam:topdb",
          "http://identifiers.org/topdb/",
          "http://identifiers.org/topdb",
          "https://identifiers.org/topdb/",
          "https://identifiers.org/topdb",
      }, new Class<?>[]{}, "MIR:00000503",
      new Class<?>[]{},
      "topdb",
      "AP00378"),
  TOPFIND("TopFind",
      "http://clipserve.clip.ubc.ca/topfind",
      new String[]{"urn:miriam:topfind",
          "http://identifiers.org/topfind/",
          "http://identifiers.org/topfind",
          "https://identifiers.org/topfind/",
          "https://identifiers.org/topfind",
      }, new Class<?>[]{}, "MIR:00000255",
      new Class<?>[]{},
      "topfind",
      "Q9UKQ2"),

  /**
   * Toxicogenomic: Chemical: http://ctdbase.org/detail.go.
   * http://ctdbase.org/detail.go
   */
  TOXICOGENOMIC_CHEMICAL("Toxicogenomic Chemical",
      "http://ctdbase.org/",
      new String[]{"urn:miriam:ctd.chemical",
          "http://identifiers.org/ctd.chemical/",
          "http://identifiers.org/ctd.chemical",
          "https://identifiers.org/ctd.chemical/",
          "https://identifiers.org/ctd.chemical",
      },
      new Class<?>[]{}, "MIR:00000098",
      new Class<?>[]{},
      "ctd.chemical",
      "D001151"),

  TOXODB("ToxoDB",
      "http://toxodb.org/toxo/",
      new String[]{"urn:miriam:toxoplasma",
          "http://identifiers.org/toxoplasma/",
          "http://identifiers.org/toxoplasma",
          "https://identifiers.org/toxoplasma/",
          "https://identifiers.org/toxoplasma",
      }, new Class<?>[]{}, "MIR:00000153",
      new Class<?>[]{},
      "toxoplasma",
      "TGME49_053730"),

  TREEBASE("TreeBASE",
      "http://treebase.org/",
      new String[]{"urn:miriam:treebase",
          "http://identifiers.org/treebase/",
          "http://identifiers.org/treebase",
          "https://identifiers.org/treebase/",
          "https://identifiers.org/treebase",
      }, new Class<?>[]{}, "MIR:00000312",
      new Class<?>[]{},
      "treebase",
      "TB2:S1000"),
  TREEFAM("TreeFam",
      "http://www.treefam.org/",
      new String[]{"urn:miriam:treefam",
          "http://identifiers.org/treefam/",
          "http://identifiers.org/treefam",
          "https://identifiers.org/treefam/",
          "https://identifiers.org/treefam",
      }, new Class<?>[]{}, "MIR:00000395",
      new Class<?>[]{},
      "treefam",
      "TF101014"),

  TREE_OF_LIFE("Tree of Life",
      "http://tolweb.org/tree/",
      new String[]{"urn:miriam:tol",
          "http://identifiers.org/tol/",
          "http://identifiers.org/tol",
          "https://identifiers.org/tol/",
          "https://identifiers.org/tol",
      }, new Class<?>[]{}, "MIR:00000405",
      new Class<?>[]{},
      "tol",
      "98034"),

  TRICHDB("TrichDB",
      "http://trichdb.org/trichdb/",
      new String[]{"urn:miriam:trichdb",
          "http://identifiers.org/trichdb/",
          "http://identifiers.org/trichdb",
          "https://identifiers.org/trichdb/",
          "https://identifiers.org/trichdb",
      }, new Class<?>[]{}, "MIR:00000154",
      new Class<?>[]{},
      "trichdb",
      "TVAG_386080"),
  TRITRYPDB("TriTrypDB",
      "http://tritrypdb.org/tritrypdb/",
      new String[]{"urn:miriam:tritrypdb",
          "http://identifiers.org/tritrypdb/",
          "http://identifiers.org/tritrypdb",
          "https://identifiers.org/tritrypdb/",
          "https://identifiers.org/tritrypdb",
      }, new Class<?>[]{}, "MIR:00000155",
      new Class<?>[]{},
      "tritrypdb",
      "Tb927.8.620"),

  TTD_Drug("TTD Drug",
      "http://bidd.nus.edu.sg/group/ttd/ttd.asp",
      new String[]{"urn:miriam:ttd.drug",
          "http://identifiers.org/ttd.drug/",
          "http://identifiers.org/ttd.drug",
          "https://identifiers.org/ttd.drug/",
          "https://identifiers.org/ttd.drug",
      },
      new Class<?>[]{}, "MIR:00000092",
      new Class<?>[]{},
      "ttd.drug",
      "DAP000773"),

  TTD_TARGET("TTD Target",
      "http://bidd.nus.edu.sg/group/ttd/ttd.asp",
      new String[]{"urn:miriam:ttd.target",
          "http://identifiers.org/ttd.target/",
          "http://identifiers.org/ttd.target",
          "https://identifiers.org/ttd.target/",
          "https://identifiers.org/ttd.target",
      }, new Class<?>[]{}, "MIR:00000093",
      new Class<?>[]{},
      "ttd.target",
      "TTDS00056"),

  TRANSPORT_CLASSIFICATION_DATABASE("Transport Classification Database",
      "http://www.tcdb.org/",
      new String[]{"urn:miriam:tcdb",
          "http://identifiers.org/tcdb/",
          "http://identifiers.org/tcdb",
          "https://identifiers.org/tcdb/",
          "https://identifiers.org/tcdb",
      }, new Class<?>[]{}, "MIR:00000040",
      new Class<?>[]{},
      "tcdb",
      "5.A.1.1.1"),

  UBERON("UBERON",
      "http://bioportal.bioontology.org/ontologies/UBERON",
      new String[]{"urn:miriam:uberon",
          "http://identifiers.org/uberon/",
          "http://identifiers.org/UBERON",
          "https://identifiers.org/uberon/",
          "https://identifiers.org/UBERON",
      }, new Class<?>[]{}, "MIR:00000446",
      new Class<?>[]{},
      "",
      "UBERON:0008203"),

  UBIO_NAMEBANK("uBio NameBank",
      "http://www.ubio.org",
      new String[]{"urn:miriam:ubio.namebank",
          "http://identifiers.org/ubio.namebank/",
          "http://identifiers.org/ubio.namebank",
          "https://identifiers.org/ubio.namebank/",
          "https://identifiers.org/ubio.namebank",
      }, new Class<?>[]{}, "MIR:00000338",
      new Class<?>[]{},
      "ubio.namebank",
      "2555646"),
  UM_BBD_COMPOUND("UM-BBD Compound",
      "http://umbbd.ethz.ch/",
      new String[]{"urn:miriam:umbbd.compound",
          "http://identifiers.org/umbbd.compound/",
          "http://identifiers.org/umbbd.compound",
          "https://identifiers.org/umbbd.compound/",
          "https://identifiers.org/umbbd.compound",
      }, new Class<?>[]{}, "MIR:00000276",
      new Class<?>[]{},
      "umbbd.compound",
      "c0001"),
  UM_BBD_ENZYME("UM-BBD Enzyme",
      "http://umbbd.ethz.ch/",
      new String[]{"urn:miriam:umbbd.enzyme",
          "http://identifiers.org/umbbd.enzyme/",
          "http://identifiers.org/umbbd.enzyme",
          "https://identifiers.org/umbbd.enzyme/",
          "https://identifiers.org/umbbd.enzyme",
      }, new Class<?>[]{}, "MIR:00000326",
      new Class<?>[]{},
      "umbbd.enzyme",
      "e0333"),
  UM_BBD_PATHWAY("UM-BBD Pathway",
      "http://umbbd.ethz.ch/",
      new String[]{"urn:miriam:umbbd.pathway",
          "http://identifiers.org/umbbd.pathway/",
          "http://identifiers.org/umbbd.pathway",
          "https://identifiers.org/umbbd.pathway/",
          "https://identifiers.org/umbbd.pathway",
      }, new Class<?>[]{}, "MIR:00000327",
      new Class<?>[]{},
      "umbbd.pathway",
      "ala"),
  UM_BBD_REACTION("UM-BBD Reaction",
      "http://umbbd.ethz.ch/",
      new String[]{"urn:miriam:umbbd.reaction",
          "http://identifiers.org/umbbd.reaction/",
          "http://identifiers.org/umbbd.reaction",
          "https://identifiers.org/umbbd.reaction/",
          "https://identifiers.org/umbbd.reaction",
      }, new Class<?>[]{}, "MIR:00000325",
      new Class<?>[]{},
      "umbbd.reaction",
      "r0001"),
  UM_BBD_BIOTRANSFORMATION_RULE("UM-BBD Biotransformation Rule",
      "http://umbbd.ethz.ch/servlets/pageservlet?ptype=allrules",
      new String[]{"urn:miriam:umbbd.rule",
          "http://identifiers.org/umbbd.rule/",
          "http://identifiers.org/umbbd.rule",
          "https://identifiers.org/umbbd.rule/",
          "https://identifiers.org/umbbd.rule",
      }, new Class<?>[]{}, "MIR:00000328",
      new Class<?>[]{},
      "umbbd.rule",
      "bt0001"),
  /**
   * UniGene: http://www.ncbi.nlm.nih.gov/unigene.
   */
  UNIGENE("UniGene",
      "http://www.ncbi.nlm.nih.gov/unigene",
      new String[]{"urn:miriam:unigene",
          "http://identifiers.org/unigene/",
          "http://identifiers.org/unigene",
          "https://identifiers.org/unigene/",
          "https://identifiers.org/unigene",
      },
      new Class<?>[]{}, "MIR:00000346",
      new Class<?>[]{},
      "unigene",
      "4900"),

  UNII("UNII",
      "http://fdasis.nlm.nih.gov/srs/",
      new String[]{"urn:miriam:unii",
          "http://identifiers.org/unii/",
          "http://identifiers.org/unii",
          "https://identifiers.org/unii/",
          "https://identifiers.org/unii",
      }, new Class<?>[]{}, "MIR:00000531",
      new Class<?>[]{},
      "unii",
      "3G6A5W338E"),

  UNIMOD("Unimod",
      "http://www.unimod.org/",
      new String[]{"urn:miriam:unimod",
          "http://identifiers.org/unimod/",
          "http://identifiers.org/unimod",
          "https://identifiers.org/unimod/",
          "https://identifiers.org/unimod",
      }, new Class<?>[]{}, "MIR:00000447",
      new Class<?>[]{},
      "unimod",
      "1200"),

  UNIPARC("UniParc",
      "https://www.ebi.ac.uk/uniparc/",
      new String[]{"urn:miriam:uniparc",
          "http://identifiers.org/uniparc/",
          "http://identifiers.org/uniparc",
          "https://identifiers.org/uniparc/",
          "https://identifiers.org/uniparc",
      }, new Class<?>[]{}, "MIR:00000041",
      new Class<?>[]{},
      "uniparc",
      "UPI000000000A"),

  @NoMiriamUrn
  @Deprecated
  UNIPATHWAY_REACTION("UniPathway Reaction",
      "http://www.grenoble.prabi.fr/obiwarehouse/unipathway/",
      new String[]{
          "http://identifiers.org/unipathway.reaction/",
          "http://identifiers.org/unipathway.reaction",
          "https://identifiers.org/unipathway.reaction/",
          "https://identifiers.org/unipathway.reaction",
      }, new Class<?>[]{}, "MIR:00000570",
      new Class<?>[]{},
      "unipathway.reaction",
      "UCR00226"),

  /**
   * Uniprot: http://www.uniprot.org/.
   */
  UNIPROT("Uniprot",
      "http://www.uniprot.org/",
      new String[]{"urn:miriam:uniprot", "urn:lsid:uniprot.org",
          "http://identifiers.org/uniprot/",
          "http://identifiers.org/uniprot",
          "https://purl.uniprot.org/uniprot/",
          "https://identifiers.org/uniprot/",
          "https://identifiers.org/uniprot",
      },
      new Class<?>[]{Protein.class, Gene.class, Rna.class}, "MIR:00000005",
      new Class<?>[]{},
      "uniprot",
      "P0DP23"),

  /**
   * UniProt Isoform: http://www.uniprot.org/.
   */
  UNIPROT_ISOFORM("UniProt Isoform",
      "http://www.uniprot.org/",
      new String[]{"urn:miriam:uniprot.isoform",
          "http://identifiers.org/uniprot.isoform/",
          "http://identifiers.org/uniprot.isoform",
          "https://identifiers.org/uniprot.isoform/",
          "https://identifiers.org/uniprot.isoform",
      },
      new Class<?>[]{Protein.class}, "MIR:00000388",
      new Class<?>[]{},
      "uniprot.isoform",
      "Q5BJF6-3"),

  UNISTS("UniSTS",
      "https://www.ncbi.nlm.nih.gov/sites/entrez?db=unists",
      new String[]{"urn:miriam:unists",
          "http://identifiers.org/unists/",
          "http://identifiers.org/unists",
          "https://identifiers.org/unists/",
          "https://identifiers.org/unists",
      }, new Class<?>[]{}, "MIR:00000162",
      new Class<?>[]{},
      "unists",
      "456789"),

  UNIT_ONTOLOGY("Unit Ontology",
      "http://bioportal.bioontology.org/ontologies/UO",
      new String[]{"urn:miriam:unit", "urn:miriam:obo.unit",
          "http://identifiers.org/uo/",
          "http://identifiers.org/UO",
          "https://identifiers.org/uo/",
          "https://identifiers.org/UO",
      }, new Class<?>[]{}, "MIR:00000136",
      new Class<?>[]{},
      "",
      "UO:0000080"),

  UNITE("Unite",
      "http://unite.ut.ee/",
      new String[]{"urn:miriam:unite",
          "http://identifiers.org/unite/",
          "http://identifiers.org/unite",
          "https://identifiers.org/unite/",
          "https://identifiers.org/unite",
      }, new Class<?>[]{}, "MIR:00000352",
      new Class<?>[]{},
      "unite",
      "UDB000691"),

  USPTO("USPTO",
      "http://patft.uspto.gov/netahtml/PTO/index.html",
      new String[]{"urn:miriam:uspto",
          "http://identifiers.org/uspto/",
          "http://identifiers.org/uspto",
          "https://identifiers.org/uspto/",
          "https://identifiers.org/uspto",
      },
      new Class<?>[]{}, "MIR:00000538",
      new Class<?>[]{},
      "uspto",
      "4145692"),

  /**
   * Unknown reference type...
   */
  @Deprecated
  UNKNOWN("Unknown",
      null,
      new String[]{},
      new Class<?>[]{}, null,
      new Class<?>[]{},
      null,
      null),

  VARIO("VariO",
      "http://bioportal.bioontology.org/ontologies/VARIO",
      new String[]{"urn:miriam:vario",
          "http://identifiers.org/vario/",
          "http://identifiers.org/VariO",
          "https://identifiers.org/vario/",
          "https://identifiers.org/VariO",
      }, new Class<?>[]{}, "MIR:00000406",
      new Class<?>[]{},
      "",
      "VariO:0294"),

  VBASE2("Vbase2",
      "http://www.vbase2.org/vbase2.php",
      new String[]{"urn:miriam:vbase2",
          "http://identifiers.org/vbase2/",
          "http://identifiers.org/vbase2",
          "https://identifiers.org/vbase2/",
          "https://identifiers.org/vbase2",
      }, new Class<?>[]{}, "MIR:00000320",
      new Class<?>[]{},
      "vbase2",
      "humIGHV025"),
  VBRC("VBRC",
      "http://vbrc.org/",
      new String[]{"urn:miriam:vbrc",
          "http://identifiers.org/vbrc/",
          "http://identifiers.org/vbrc",
          "https://identifiers.org/vbrc/",
          "https://identifiers.org/vbrc",
      }, new Class<?>[]{}, "MIR:00000448",
      new Class<?>[]{},
      "vbrc",
      "35742"),

  VFDB_GENE("VFDB Gene",
      "http://www.mgc.ac.cn/VFs/",
      new String[]{"urn:miriam:vfdb.gene",
          "http://identifiers.org/vfdb.gene/",
          "http://identifiers.org/vfdb.gene",
          "https://identifiers.org/vfdb.gene/",
          "https://identifiers.org/vfdb.gene",
      }, new Class<?>[]{}, "MIR:00000472",
      new Class<?>[]{},
      "vfdb.gene",
      "VFG2154"),
  VFDB_GENUS("VFDB Genus",
      "http://www.mgc.ac.cn/VFs/",
      new String[]{"urn:miriam:vfdb.genus",
          "http://identifiers.org/vfdb.genus/",
          "http://identifiers.org/vfdb.genus",
          "https://identifiers.org/vfdb.genus/",
          "https://identifiers.org/vfdb.genus",
      }, new Class<?>[]{}, "MIR:00000471",
      new Class<?>[]{},
      "vfdb.genus",
      "Chlamydia"),

  VIRALZONE("ViralZone",
      "http://www.expasy.org/viralzone/",
      new String[]{"urn:miriam:viralzone",
          "http://identifiers.org/viralzone/",
          "http://identifiers.org/viralzone",
          "https://identifiers.org/viralzone/",
          "https://identifiers.org/viralzone",
      }, new Class<?>[]{}, "MIR:00000449",
      new Class<?>[]{},
      "viralzone",
      "992"),
  VIRSIRNA("VIRsiRNA",
      "http://crdd.osdd.net/servers/virsirnadb",
      new String[]{"urn:miriam:virsirna",
          "http://identifiers.org/virsirna/",
          "http://identifiers.org/virsirna",
          "https://identifiers.org/virsirna/",
          "https://identifiers.org/virsirna",
      }, new Class<?>[]{}, "MIR:00000249",
      new Class<?>[]{},
      "virsirna",
      "virsi1909"),

  /**
   * VMH metabolite: https://vmh.uni.lu/.
   */
  VMH_METABOLITE("VMH metabolite",
      "https://vmh.uni.lu/",
      new String[]{"urn:miriam:vmhmetabolite",
          "http://identifiers.org/vmhmetabolite/",
          "http://identifiers.org/vmhmetabolite",
          "https://identifiers.org/vmhmetabolite/",
          "https://identifiers.org/vmhmetabolite",
      },
      new Class<?>[]{Chemical.class}, "MIR:00000636",
      new Class<?>[]{},
      "vmhmetabolite",
      "h2o"),

  /**
   * VMH reaction: https://vmh.uni.lu/.
   */
  VMH_REACTION("VMH reaction",
      "https://vmh.uni.lu/",
      new String[]{"urn:miriam:vmhreaction",
          "http://identifiers.org/vmhreaction/",
          "http://identifiers.org/vmhreaction",
          "https://identifiers.org/vmhreaction/",
          "https://identifiers.org/vmhreaction",
      },
      new Class<?>[]{Reaction.class}, "MIR:00000640",
      new Class<?>[]{},
      "vmhreaction",
      "HEX1"),

  /**
   * Wikidata: https://www.wikidata.org/.
   */
  WIKIDATA("Wikidata",
      "https://www.wikidata.org/",
      new String[]{"urn:miriam:wikidata",
          "http://identifiers.org/wikidata/",
          "http://identifiers.org/wikidata",
          "https://identifiers.org/wikidata/",
          "https://identifiers.org/wikidata",
      },
      new Class<?>[]{}, "MIR:00000549",
      new Class<?>[]{},
      "wikidata",
      "Q2207226"),

  WIKIGENES("WikiGenes",
      "http://www.wikigenes.org/",
      new String[]{"urn:miriam:wikigenes",
          "http://identifiers.org/wikigenes/",
          "http://identifiers.org/wikigenes",
          "https://identifiers.org/wikigenes/",
          "https://identifiers.org/wikigenes",
      }, new Class<?>[]{}, "MIR:00000437",
      new Class<?>[]{},
      "wikigenes",
      "3771877"),

  /**
   * WikiPathways: http://www.wikipathways.org/.
   */
  WIKIPATHWAYS("WikiPathways",
      "http://www.wikipathways.org/",
      new String[]{"urn:miriam:wikipathways",
          "http://identifiers.org/wikipathways/",
          "http://identifiers.org/wikipathways",
          "https://identifiers.org/wikipathways/",
          "https://identifiers.org/wikipathways"
      },
      new Class<?>[]{}, "MIR:00000076",
      new Class<?>[]{},
      "wikipathways",
      "WP100"),

  /**
   * Wikipedia: http://en.wikipedia.org/wiki/Main_Page.
   */
  WIKIPEDIA("Wikipedia (English)",
      "http://en.wikipedia.org/wiki/Main_Page",
      new String[]{"urn:miriam:wikipedia.en",
          "http://identifiers.org/wikipedia.en/",
          "http://identifiers.org/wikipedia.en",
          "https://identifiers.org/wikipedia.en/",
          "https://identifiers.org/wikipedia.en",
      },
      new Class<?>[]{}, "MIR:00000384",
      new Class<?>[]{},
      "wikipedia.en",
      "SM_UB-81"),

  WORFDB("Worfdb",
      "http://worfdb.dfci.harvard.edu/",
      new String[]{"urn:miriam:worfdb",
          "http://identifiers.org/worfdb/",
          "http://identifiers.org/worfdb",
          "https://identifiers.org/worfdb/",
          "https://identifiers.org/worfdb",
      }, new Class<?>[]{}, "MIR:00000288",
      new Class<?>[]{},
      "worfdb",
      "T01B6.1"),

  /**
   * WormBase: http://wormbase.bio2rdf.org/fct.
   */
  WORM_BASE("WormBase",
      "http://wormbase.bio2rdf.org/fct",
      new String[]{"urn:miriam:wormbase",
          "http://identifiers.org/wb/",
          "http://identifiers.org/wb",
          "https://identifiers.org/wb/",
          "https://identifiers.org/wb",
      },
      new Class<?>[]{}, "MIR:00000027",
      new Class<?>[]{},
      "wb",
      "WBGene00000001"),

  WORM_BASE_RNAI("WormBase RNAi",
      "https://www.wormbase.org/",
      new String[]{"urn:miriam:wormbase.rnai",
          "http://identifiers.org/wb.rnai/",
          "http://identifiers.org/wb.rnai",
          "https://identifiers.org/wb.rnai/",
          "https://identifiers.org/wb.rnai",
      },
      new Class<?>[]{}, "MIR:00000466",
      new Class<?>[]{},
      "wb.rnai",
      "WBRNAi00086878"),

  WORMPEP("Wormpep",
      "https://www.wormbase.org/db/seq/protein",
      new String[]{"urn:miriam:wormpep",
          "http://identifiers.org/wormpep/",
          "http://identifiers.org/wormpep",
          "https://identifiers.org/wormpep/",
          "https://identifiers.org/wormpep",
      }, new Class<?>[]{}, "MIR:00000031",
      new Class<?>[]{},
      "wormpep",
      "CE28239"),

  XENBASE("Xenbase",
      "https://www.xenbase.org/",
      new String[]{"urn:miriam:xenbase",
          "http://identifiers.org/xenbase/",
          "http://identifiers.org/xenbase",
          "https://identifiers.org/xenbase/",
          "https://identifiers.org/xenbase",
      }, new Class<?>[]{}, "MIR:00000186",
      new Class<?>[]{},
      "xenbase",
      "XB-GENE-922462"),
  YDPM("YDPM",
      "http://www-deletion.stanford.edu/YDPM/",
      new String[]{"urn:miriam:ydpm",
          "http://identifiers.org/ydpm/",
          "http://identifiers.org/ydpm",
          "https://identifiers.org/ydpm/",
          "https://identifiers.org/ydpm",
      }, new Class<?>[]{}, "MIR:00000465",
      new Class<?>[]{},
      "ydpm",
      "YAL001C"),

  YEAST_INTRON_DATABASE_V4_3("Yeast Intron Database v4.3",
      "http://intron.ucsc.edu/yeast4.3/",
      new String[]{"urn:miriam:yeastintron",
          "http://identifiers.org/yeastintron/",
          "http://identifiers.org/yeastintron",
          "https://identifiers.org/yeastintron/",
          "https://identifiers.org/yeastintron",
      }, new Class<?>[]{}, "MIR:00000521",
      new Class<?>[]{},
      "yeastintron",
      "SNR17A"),

  YETFASCO("YeTFasCo",
      "http://yetfasco.ccbr.utoronto.ca/",
      new String[]{"urn:miriam:yetfasco",
          "http://identifiers.org/yetfasco/",
          "http://identifiers.org/yetfasco",
          "https://identifiers.org/yetfasco/",
          "https://identifiers.org/yetfasco",
      }, new Class<?>[]{}, "MIR:00000339",
      new Class<?>[]{},
      "yetfasco",
      "YOR172W_571.0"),

  YEAST_INTRON_DATABASE_V3("Yeast Intron Database v3",
      "http://compbio.soe.ucsc.edu/yeast_introns.html",
      new String[]{"urn:miriam:yid",
          "http://identifiers.org/yid/",
          "http://identifiers.org/yid",
          "https://identifiers.org/yid/",
          "https://identifiers.org/yid",
      }, new Class<?>[]{}, "MIR:00000460",
      new Class<?>[]{},
      "yid",
      "SNR17A"),
  YRC_PDR("YRC PDR",
      "http://www.yeastrc.org/pdr/",
      new String[]{"urn:miriam:yrcpdr",
          "http://identifiers.org/yrcpdr/",
          "http://identifiers.org/yrcpdr",
          "https://identifiers.org/yrcpdr/",
          "https://identifiers.org/yrcpdr",
      }, new Class<?>[]{}, "MIR:00000459",
      new Class<?>[]{},
      "yrcpdr",
      "2673500"),

  ZFIN("ZFIN Bioentity",
      "http://zfin.org",
      new String[]{"urn:miriam:zfin",
          "http://identifiers.org/zfin/",
          "http://identifiers.org/zfin",
          "https://identifiers.org/zfin/",
          "https://identifiers.org/zfin",
      },
      new Class<?>[]{}, "MIR:00000079",
      new Class<?>[]{},
      "zfin",
      "ZDB-GENE-041118-11"),

  ZINC("ZINC",
      "http://zinc15.docking.org/",
      new String[]{"urn:miriam:zinc",
          "http://identifiers.org/zinc/",
          "http://identifiers.org/zinc",
          "https://identifiers.org/zinc/",
          "https://identifiers.org/zinc",
      }, new Class<?>[]{}, "MIR:00000529",
      new Class<?>[]{},
      "zinc",
      "ZINC1084"),

  ;

  private static final Logger logger = LogManager.getLogger();
  /**
   * User friendly name.
   */
  private final String commonName;

  /**
   * url to home page of given resource type.
   */
  private final String dbHomepage;

  /**
   * Identifier of the database in miriam registry.
   */
  private final String registryIdentifier;

  /**
   * Valid URIs to this resource.
   */
  private final List<String> uris = new ArrayList<>();

  /**
   * Classes that can be annotated by this resource.
   */
  private final List<Class<? extends BioEntity>> validClass = new ArrayList<>();

  /**
   * When class from this list is marked as "require at least one annotation" then
   * annotation of this type is valid.
   */
  private final List<Class<? extends BioEntity>> requiredClass = new ArrayList<>();

  private final String namespace;

  private final String exampleIdentifier;

  /**
   * Constructor that initialize enum object.
   *
   * @param dbHomePage         home page of the resource {@link #dbHomepage}
   * @param commonName         {@link #commonName}
   * @param uris               {@link #uris}
   * @param classes            {@link #validClass}
   * @param registryIdentifier {@link #registryIdentifier}
   * @param requiredClasses    {@link #requiredClass}
   */
  MiriamType(final String commonName, final String dbHomePage, final String[] uris, final Class<?>[] classes, final String registryIdentifier,
             final Class<?>[] requiredClasses, final String namespace, final String exampleIdentifier) {
    this.commonName = commonName;
    this.dbHomepage = dbHomePage;
    Collections.addAll(this.uris, uris);
    for (final Class<?> clazz : classes) {
      this.validClass.add((Class<? extends BioEntity>) clazz);
    }
    for (final Class<?> clazz : requiredClasses) {
      this.requiredClass.add((Class<? extends BioEntity>) clazz);
    }
    this.registryIdentifier = registryIdentifier;
    this.namespace = namespace;
    this.exampleIdentifier = exampleIdentifier;
  }

  /**
   * Constructor that initialize enum object.
   *
   * @param dbHomePage         home page of the resource {@link #dbHomepage}
   * @param commonName         {@link #commonName}
   * @param uri                one of {@link #uris}
   * @param registryIdentifier {@link #registryIdentifier}
   * @param classes            {@link #validClass}
   */
  MiriamType(final String commonName, final String dbHomePage, final String uri, final Class<?>[] classes, final String registryIdentifier,
             final Class<?>[] requiredClasses, final String namespace, final String exampleIdentifier) {
    this(commonName, dbHomePage, new String[]{uri}, classes, registryIdentifier, requiredClasses, namespace,
        exampleIdentifier);
  }

  /**
   * Returns {@link MiriamType} associated with parameter uri address.
   *
   * @param uri uri to check
   * @return {@link MiriamType} for given uri
   */
  public static MiriamType getTypeByUri(final String uri) {
    for (final MiriamType mt : MiriamType.values()) {
      for (final String string : mt.getUris()) {
        if (string.equalsIgnoreCase(uri)) {
          return mt;
        }
      }
    }
    logger.warn("Unknown miriam type: " + uri);
    return null;
  }

  /**
   * Returns {@link MiriamType} associated with {@link #commonName}.
   *
   * @param string {@link #commonName}
   * @return {@link MiriamType} for given name
   */
  public static MiriamType getTypeByCommonName(final String string) {
    for (final MiriamType mt : MiriamType.values()) {
      if (string.equalsIgnoreCase(mt.getCommonName())) {
        return mt;
      }
    }
    return null;
  }

  /**
   * Transforms identifier into {@link MiriamData}.
   *
   * @param generalIdentifier identifier in the format NAME:IDENTIFIER. Where NAME is the name
   *                          from {@link MiriamType#commonName} and IDENTIFIER is resource
   *                          identifier.
   * @return {@link MiriamData} representing generalIdentifier, when identifier is invalid InvalidArgumentException is thrown
   */
  public static MiriamData getMiriamDataFromIdentifier(final String generalIdentifier) {
    int index = generalIdentifier.indexOf(":");
    if (index < 0) {
      throw new InvalidArgumentException("Identifier doesn't contain type");
    }
    String type = generalIdentifier.substring(0, index);
    String id = generalIdentifier.substring(index + 1);
    for (final MiriamType mt : MiriamType.values()) {
      if (mt.getCommonName().equalsIgnoreCase(type)) {
        return new MiriamData(mt, id);
      }
    }
    throw new InvalidArgumentException("Unknown miriam type: " + type + " (id: " + id + ")");
  }

  public static MiriamData getMiriamDataFromPrefixIdentifier(final String generalIdentifier) {
    if (generalIdentifier == null) {
      return null;
    }
    int index = generalIdentifier.indexOf(":");
    if (index < 0) {
      throw new InvalidArgumentException("Identifier doesn't contain type");
    }
    String type = generalIdentifier.substring(0, index);
    String id = generalIdentifier.substring(index + 1);
    for (final MiriamType mt : MiriamType.values()) {
      if (mt.getNamespace().equalsIgnoreCase(type)) {
        return new MiriamData(mt, id);
      }
    }
    throw new InvalidArgumentException("Unknown miriam type: " + type + " (id: " + id + ")");
  }

  /**
   * Creates {@link MiriamData} from miriam uri.
   *
   * @param sourceMiriamUri miriam uri defining {@link MiriamData}
   * @return {@link MiriamData} from miriam uri
   */
  public static MiriamData getMiriamByUri(final String sourceMiriamUri) {
    // this hack is due to CellDesigner issue (final CellDesigner incorrectly
    // handle
    // with identifiers that have ":" inside resource ":" inside resource with
    // "%3A" and also the last ":"
    String miriamUri = sourceMiriamUri.replace("%3A", ":");

    String miriamUriLowerCase = miriamUri.toLowerCase();

    String foundUri = "";
    MiriamType foundType = null;

    for (MiriamType type : MiriamType.values()) {
      for (String uri : type.getUris()) {
        uri = uri.toLowerCase();
        if (miriamUriLowerCase.startsWith(uri + ":")) {
          if (uri.length() > foundUri.length()) {
            foundType = type;
            foundUri = uri;
          }
        } else if (miriamUriLowerCase.startsWith(uri) && uri.endsWith("/")) {
          if (uri.length() > foundUri.length()) {
            foundType = type;
            foundUri = uri;
          }
        } else if (miriamUriLowerCase.startsWith(uri) && uri.endsWith("/")) {
          if (uri.length() > foundUri.length()) {
            foundType = type;
            foundUri = uri;
          }
        }
      }
    }
    if (foundType != null) {
      if (foundUri.length() >= miriamUri.length()) {
        throw new InvalidArgumentException("Invalid miriam uri: " + miriamUri);
      }
      int uriLength = foundUri.length();
      if (foundUri.contains("://identifiers.org/") && !foundUri.replaceAll("://identifiers.org/", "").contains("/")
          && (foundType.getNamespace() == null || foundType.getNamespace().isEmpty())) {
        uriLength = foundUri.lastIndexOf("/") + 1;
      } else if (!foundUri.endsWith("/")) { // add one character for separator
        uriLength++;
      }
      String resource = miriamUri.substring(uriLength);
      try {
        resource = URLDecoder.decode(resource, "UTF-8");
      } catch (final UnsupportedEncodingException e) {
        logger.warn("Problematic uri: " + resource, e);
      }
      return new MiriamData(foundType, resource);
    }
    throw new InvalidArgumentException("Invalid miriam uri: " + miriamUri);
  }

  /**
   * @return {@link #commonName}
   */
  public String getCommonName() {
    return commonName;
  }

  /**
   * @return {@link #uris}
   */
  public List<String> getUris() {
    return uris;
  }

  /**
   * @return {@link #validClass}
   */
  public List<Class<? extends BioEntity>> getValidClass() {
    return validClass;
  }

  /**
   * @return the dbHomepage
   * @see #dbHomepage
   */
  public String getDbHomepage() {
    return dbHomepage;
  }

  /**
   * @return the registryIdentifier
   * @see #registryIdentifier
   */
  public String getRegistryIdentifier() {
    return registryIdentifier;
  }

  /**
   * @return the requiredClass
   * @see #requiredClass
   */
  public List<Class<? extends BioEntity>> getRequiredClass() {
    return requiredClass;
  }

  public String getNamespace() {
    return namespace;
  }

  public String getExampleIdentifier() {
    return exampleIdentifier;
  }

}
